/*
 *	listree.h
 *
 *	The List Tree represents a tree using linked lists. The tree
 *	implements multi-level push operations.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_LISTREE_H_
#define	_OPaC_LISTREE_H_

#ifndef _OPaC_CONTAINER_H_
#include <OPaC/container.h>
#endif

#ifndef _OPaC_ITERATOR_H_
#include <OPaC/iterator.h>
#endif

/****************************************************************/

/*
 *	The List Tree class looks very much like a list, but it
 *	has an additional dimension: the depth.
 */

class OP_PUBLIC OPLisTree : public OPContainer
{
	
public:
	enum Type
	{
		OP_LIST_EMPTY		= 0,				//	no more data
		OP_LIST_DATA		= 1,				//	data
		OP_LIST_NEW_LEVEL	= 2,				//	new level
		OP_LIST_END_LEVEL	= 3,				//	end of level
		OP_LIST_MARK		= 4					//	mark
	};
	
protected:
	OPLisTreeRef		parent;					//	up one level (if any)
	OPListRef			list;					//	list of elements at this level
	
	OPLisTreeRef		level;					//	insertion point (or self)
	
protected:
	OPLisTree ();
	virtual ~ OPLisTree ();
	
	virtual void StoreInstance (OPStorageIn storage);		/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);		/* OVERRIDDEN */
	
public:
	virtual void Delete ();									/* OVERRIDDEN */
	virtual OPIteratorRef NewIterator ();					/* OVERRIDDEN */
	
	virtual Bool Insert (OPObjectIn data);					/* OVERRIDDEN */
	virtual Bool InsertNewLevel ();
	virtual Bool InsertEndLevel ();
	virtual Bool InsertMark ();
	
	virtual Bool Remove (OPObjectIn data, Count nth = 0);	/* OVERRIDDEN */
	virtual Bool Remove (Count index);						/* OVERRIDDEN */
	virtual void RemoveAll ();								/* OVERRIDDEN */
	
	virtual Bool Get (Count index, OPObjectOut object);		/* OVERRIDDEN */
	virtual Bool GetSimilar (OPObjectIn model, OPObjectOut object, Count nth = 0);	/* OVERRIDDEN */
	
	virtual Bool LookFor (OPObjectIn value);				/* OVERRIDDEN */
	virtual Bool Find (OPObjectIn value, CountInOut index);	/* OVERRIDDEN */
	
	virtual Bool Clone (OPLisTreeOut tree, OPLisTreeIn parent = OPLisTreeNull) const;
	virtual Bool Rewind ();
	virtual Type PeekHeadType () const;
	virtual Type PopHead (OPObjectOut);
	
	OP_DEFINE_CLASS (OPLisTree, OPContainer);
};

/****************************************************************/

/*
 *	List tree iterator.
 */

class OP_PUBLIC OPLisTreeIterator : public OPIterator
{
	friend class OPLisTree;
	
protected:
	OPLisTreeRef		level;					//	current level
	Count				index;					//	index into level
	
protected:
	OPLisTreeIterator ();
	virtual ~ OPLisTreeIterator ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	//	The iterator exports following overridden methods :
	
	virtual Bool First ();								/* OVERRIDDEN */
	virtual Bool Last ();								/* OVERRIDDEN */
	
	virtual Bool Next (Count move = 1);					/* OVERRIDDEN */
	virtual Bool Prev (Count move = 1);					/* OVERRIDDEN */
	
	virtual Bool Get (OPObjectOut, Count index = 0);	/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPLisTreeIterator, OPIterator);
};

/****************************************************************/

#endif /* _OPaC_LISTREE_H_ */
