/*
 *	new.h
 *
 *	This file defines the C++ memory allocation operators and the
 *	associated macros.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_NEW_H_
#define	_OPaC_NEW_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

/****************************************************************/

#define	OP_MTYPE_BINARY		((Card8)(0x00))
#define	OP_MTYPE_OBJECT		((Card8)(0x01))
#define	OP_MTYPE_OBJARRAY	((Card8)(0x02))
#define	OP_MTYPE_FREE		((Card8)(0xFF))

#define	MEM_STAT_SMALL		128

/****************************************************************/

/*
 *	The new/delete operators should never be used !
 *	
 *	Use OP_NEW and OP_NEW_FROM_ZONE to allocate new OPaC
 *	compatible instances. These return references rather
 *	than pointers.
 *	
 *	Use OP_ALLOC(_ARRAY)(_FROM_ZONE) to allocate the other
 *	structures and buffers for which you would otherwise
 *	have used malloc.
 *	
 *	Use NEW/DELETE macros to allocate/free instances which
 *	do not belong to the OPaC class hierarchy.
 */

/****************************************************************/

/*
 *	The OPaC Memory Zone class maintains pools of memory,
 *	which can be used to reduce scattering of allocations
 *	through the whole VM space.
 */

class OP_PUBLIC OPMemZone
{
	friend class OPObject;
	
protected:
	
	//	A chunk is a piece of memory which can be used by the programmer. It
	//	is part of a larger memory pool called a buffer. Allocating/freeing
	//	chunks usually won't produce system calls and should be fast.
	
	struct ChunkHeader
	{
		ChunkHeader*	 next;					//	pointer to next memory chunk
		ChunkHeader*	 prev;					//	pointer to previous memory chunk
		Card32			 size;					//	size of memory chunk
		BytePtr			 ptr;					//	pointer to memory chunk
		
		Count			 source_line;			//	source line # in file
		const LatinChar* source_file;			//	source file name
		const LatinChar* source_type;			//	source data type
		Count			 call_id;				//	call ID
		
		Card8			 mem_type;				//	one of OP_MTYPE_xxx
	};
	
	//	A buffer is a contiguous array of bytes which has been allocated by
	//	some system dependent call. The buffer is used as a memory pool from
	//	which smaller memoy chunks are extracted by the OPaC MemZone class.
	
	struct BufferHeader
	{
		BufferHeader*	next;					//	next buffer description
		BufferHeader*	prev;					//	previous buffer description
		
		ChunkHeader*	head;					//	first chunk description
		ChunkHeader*	tail;					//	last chunk description
		
		Card32			size;					//	buffer size (may be large)
		BytePtr			buffer;					//	buffer (large array of bytes)
		Card32			mem_id;					//	memory ID associated to buffer
		
		Card32			free_hole;				//	current biggest free hole
	};
	
	//	A header will hold either a buffer header, a chunk header or the pointer
	//	to the next free header.
	
	union Header
	{
		BufferHeader	buffer;					//	it is a buffer header...
		ChunkHeader		chunk;					//	it is a chunk header...
		Header*			next;					//	next generalized header (for empty headers)
	};
	
	struct HeaderArray
	{
		HeaderArray*	next;					//	next array
		HeaderArray*	prev;					//	previous array
		
		Card32			size;					//	size of this header array
		BytePtr			buffer;					//	pointer to buffer (that is ourself)
		Card32			mem_id;					//	memory ID associated to this header array
		
		Card32			num_free;				//	# of free headers in array
		Header*			free;					//	first free header in array
		Card32			padding;				//	sizeof (HeaderArray) = 128 x 32 = 4096 bytes
		
		Header			array[127];				//	array of generalized headers
	};
	
private:
	static OPMemZone*	zone_head;				//	first memory zone
	static OPMemZone*	zone_tail;				//	last memory zone
	static Card32		vector_offset;			//	offset if delete[] used
	
protected:
	const LatinChar*	zone_name;				//	pointer to zone name given at init.
	
	OPMemZone*			next;					//	next memory zone
	OPMemZone*			prev;					//	previous memory zone
	
	Bool				is_zone_ok;				//	TRUE => zone has been initialised
	Card32				sys_zone;				//	system zone ID associated to this memory zone
	
	BufferHeader*		buf_head;				//	first memory buffer in zone
	BufferHeader*		buf_tail;				//	last memory buffer in zone
	
	HeaderArray*		array_head;				//	first header array
	HeaderArray*		array_tail;				//	last header array
	
	ChunkHeader*		latest_allocation;		//	remember latest allocated chunk
	
	Card32				buf_allocated;			//	total memory allocated by the buffers
	Card32				buf_remaining;			//	total memory remaining unused in the buffers
	
	Bool				fill_on_new;			// TRUE => fill w/data on new
	Bool				fill_on_delete;			// TRUE => fill w/data on delete
	
	Card32				data_on_new;			// filler data for new
	Card32				data_on_delete;			// filler data for delete
	
	Card32				stat_alloc[MEM_STAT_SMALL];
	Card32				stat_alloc_large;
	
private:
	void FreeAndDestroy ();
	
protected:
	virtual Header* NewHeader ();
	virtual void DeleteHeader (Header* header);
	
	virtual void UpdateFreeHole (BufferHeader* buffer);
	virtual void MergeChunks (ChunkHeader* chunk, BufferHeader* buffer);
	
	virtual void KillRefFromZone ();
	virtual void KillRefToZone ();
	virtual void KillRefFromBuffer (BufferHeader* buffer);
	virtual void KillRefToBuffer (BufferHeader* buffer);
	
public:
	OPMemZone ();
	virtual ~ OPMemZone ();
	
	virtual Bool Initialise (const LatinChar* name);
	
	//	Allocate memory from the zone. Record informations about the data type
	//	and the source position where the allocation originates from :
	
	virtual void* Alloc (Card32 size, const LatinChar* type,
						 Count line, const LatinChar* file,
						 Card8 mem_type = OP_MTYPE_BINARY);
	
	//	Free memory from the zone. The caller specifies where the deallocation
	//	originates from :
	
	virtual void Free (void* ptr, Count line, const LatinChar* file);
	virtual void FreeVector (void* ptr, Count line, const LatinChar* file);
	
	//	A memory zone may fill the allocated/freed memory with a specified 32
	//	bit pattern. This is provided for debugging only and may not be fully
	//	implemented on the commercial software release :
	
	virtual void SetAllocDebugMode (Bool on, Card32 fill = 0);
	virtual void SetFreeDebugMode (Bool on, Card32 fill = 0);
	
	//	Given a pointer into a zone, it is possible to find out its base. This
	//	can then be used to retrieve more informations about the memory chunk.
	//	Please be careful to specify base pointers where they are required :
	
	virtual const void* FindBasePointer (const void* ptr);
	virtual Bool AboutBasePointer (const void* base, OPStringRef& type,
								   Count& line, OPStringRef& file);
	
	//	It is possible to find out which zone a pointer belongs to and if a
	//	given pointer points to a known zone :
	
	static Bool FindZone (const void* ptr, OPMemZone*& zone);
	static Bool CheckPointer (const void* ptr);
	
	//	The two following functions are to be used exclusively by the OPaC
	//	macros, not by the programmer :
	
	static void* NullAllocZone (const LatinChar* zone, Count line, const LatinChar* file);
	static void  NullFreeZone (const LatinChar* zone, Count line, const LatinChar* file);
	static void  FreeAnyZone (void* ptr, Count line, const LatinChar* file);
	static void  FreeVectorAnyZone (void* ptr, Count line, const LatinChar* file);
	
	//	Initialise the memory manager :
	
	static void Initialise ();
	static void Kill ();
};

/****************************************************************/

extern OP_PUBLIC OPMemZone* OPaC_data_zone;		//	main data zone used by OPaC library
extern OP_PUBLIC OPMemZone* OPaC_app_zone;		//	main data zone used by the application
extern OP_PUBLIC OPMemZone* OPaC_temp_zone;		//	temporary data zone
extern OP_PUBLIC OPMemZone* OPaC_iface_zone;	//	interface objects
extern OP_PUBLIC OPMemZone* OPaC_sys_zone;		//	system zone (internal use only)

/****************************************************************/


/*
 *	You should not use malloc/free in an OPaC application,
 *	since these functions are not 100% portable. Look into
 *	the Win16 doc if you don't believe this :-( !
 *
 *	Use the following macros :
 *
 *	  - OP_ALLOC (type) returns a "type*" on success or
 *		raises an exception on failure. The zone used for
 *		the allocation will be "OPaC_app_zone".
 *
 *	  - OP_ALLOC_FROM_ZONE (zone, type) returns a "type*"
 *		on success or raises an exception on failure. The
 *		"zone" describes an OPMemZone memory.. zone.
 *
 *	  - OP_ALLOC_ARRAY (type, n) allocates "n" elements
 *		of "type". See OP_ALLOC.
 *
 *	  - OP_ALLOC_ARRAY_FROM_ZONE (zone, type, n) allocates
 *		"n" elements of "type". See OP_ALLOC_FROM_ZONE.
 *
 *	The use of a memory zone allows the logical grouping
 *	of memory requests in the same group of virtual memory
 *	pages and may reduce dramatically access times.
 */

#define	OP_ALLOC(T)												\
	((T*)(OPaC_app_zone->Alloc ((Card32)(sizeof (T)),			\
								#T, __LINE__, __OPaC_src_file)))

#define	OP_ALLOC_FROM_ZONE(z,T)									\
	((T*)(z ? z->Alloc ((Card32)(sizeof (T)), #T "[]",			\
						__LINE__, __OPaC_src_file)				\
		    : OPMemZone::NullAllocZone (#z, __LINE__, __OPaC_src_file)))

#define	OP_ALLOC_ARRAY(T,n)										\
	((T*)(OPaC_app_zone->Alloc ((Card32)(sizeof (T))*(n),		\
								#T, __LINE__, __OPaC_src_file)))

#define	OP_ALLOC_ARRAY_FROM_ZONE(z,T,n)							\
	((T*)(z ? z->Alloc ((Card32)(sizeof (T))*(n), #T "[]",		\
		  				__LINE__, __OPaC_src_file)				\
		    : OPMemZone::NullAllocZone (#z, __LINE__, __OPaC_src_file)))

#define	OP_FREE(p)												\
	OPMemZone::FreeAnyZone (p, __LINE__, __OPaC_src_file)

#define	OP_FREE_VECTOR(p)										\
	OPMemZone::FreeVectorAnyZone (p, __LINE__, __OPaC_src_file)

#define	OP_FREE_FROM_ZONE(z,p)									\
	(z ? z->Free (p, __LINE__, __OPaC_src_file)					\
	   : OPMemZone::NullFreeZone (#z, __LINE__, __OPaC_src_file))

#define	OP_FREE_VECTOR_FROM_ZONE(z,p)							\
	(z ? z->FreeVector (p, __LINE__, __OPaC_src_file)			\
	   : OPMemZone::NullFreeZone (#z, __LINE__, __OPaC_src_file))

/****************************************************************/


/*
 *	You should not use new/delete in an OPaC application,
 *	since these operators are not 100% portable. Once more,
 *	its Win16's fault !
 *	
 *	Use the following macros to instanciate OPaC classes :
 *	
 *	  - OP_NEW (class) creates an instance of "class".
 *		The class must be known at compile time. Returns
 *		a reference to the instance on success or raises
 *		an exception on failure. The zone used for the
 *		allocation will be "OPaC_app_zone".
 *	
 *	  - OP_NEW_FROM_ZONE (zone, class) is very similar
 *		to OP_NEW, but memory allocation is done using
 *		the specified memory zone.
 *	
 *	It is possible to instanciate classes dynamically.
 *	The class name has to be specified as an atomic string
 *	using the following macros :
 *
 *	  - OP_DYNAMIC_NEW (class, model) creates an instance
 *		of class "class", which has a super-class of type
 *		"model". The model must be a class type known at
 *		compile time, whereas the name is represented by
 *		an instance of OPRuntime.
 *
 *	  - OP_DYNAMIC_NEW_FROM_ZONE (zone, model, name).
 *
 *	You may not allocate arrays of instances if you are
 *	building an OPaC compatible application !
 */

#define	OP_NEW(T)												\
	((T##Ref)((T*)(OPClassMan::New (OP_RUNTIME (T),				\
									OPaC_app_zone,				\
									__LINE__, __OPaC_src_file))))

#define	OP_NEW_FROM_ZONE(z,T)									\
	((T##Ref)((T*)(OPClassMan::New (OP_RUNTIME (T),				\
									z,							\
									__LINE__, __OPaC_src_file))))

#define	OP_DYNAMIC_NEW(C,T)										\
	((T##Ref)((T*)(OPClassMan::New (C, OPaC_app_zone,			\
									__LINE__, __OPaC_src_file))))

#define	OP_DYNAMIC_NEW_FROM_ZONE(z,C,T)							\
	((T##Ref)((T*)(OPClassMan::New (C, z,						\
									__LINE__, __OPaC_src_file))))

/****************************************************************/


/*
 *	If a class which does not belong to the OPaC hierarchy
 *	and you still want to instanciate it, replace operator
 *	"new" by "NEW", operator "delete x" by "DELETE (x)".
 *
 *	You should not allocate array of instances, since not
 *	every C++ compiler recognizes the "delete[]" operator.
 *
 *	CAUTION: This will by portable only if the allocated
 *			 instances are smaller than 32 KB.
 *
 *	The NEW operator uses the "OPaC_app_zone" memory zone.
 *	Another zone can be specified using NEW_FROM_ZONE (z)
 *	in place of NEW.
 *
 *	Here is a small example :
 *
 *	  T* t_1 = NEW T ("Hello", 10);
 *	  T* t_2 = NEW_FROM_ZONE (OPaC_temp_zone) T ("Bye", 0);
 *	  ...
 *	  DELETE (t_1);
 *	  DELETE (t_2);
 *
 *	CAUTION: The argument of macro DELETE may be evaluated
 *			 twice !
 */

#define	NEW														\
	new (OPaC_app_zone, "?", __LINE__, __OPaC_src_file)

#define	NEW_FROM_ZONE(z)										\
	new (z, "?", __LINE__, __OPaC_src_file)

#define	NEW_USING(buffer)										\
	new (buffer, __LINE__, __OPaC_src_file)

#define	JUST_DELETE(x)											\
	delete (x)

#define	DELETE(x)												\
	do {														\
		delete (x);												\
		OP_FREE (x);											\
	} while (0)

#define	DELETE_ARRAY(x)											\
	do {														\
		delete[] (x);											\
		OP_FREE_VECTOR (x);										\
	} while (0)

#define	DELETE_FROM_ZONE(z,x)									\
	do {														\
		delete (x);												\
		OP_FREE_FROM_ZONE (z,x);								\
	} while (0)

#define	DELETE_ARRAY_FROM_ZONE(z,x)								\
	do {														\
		delete[] (x);											\
		OP_FREE_VECTOR_FROM_ZONE (z,x);							\
	} while (0)

/****************************************************************/

/*
 *	The Class Manager does not belong to the standard
 *	class hierarchy. It is provided in order to enumerate
 *	available classes and get the associated run-time
 *	informations.
 *
 *	You never instanciate the Class Manager. Just use it
 *	like a library.
 */

class OP_PUBLIC OPClassMan
{
	struct ClassDef
	{
		ClassDef*			next;
		ClassDef*			prev;
		OPRuntimeRef		runtime;
		OPAtomRef			name;
	};
	
private:
	static Count			class_num;			//	total # of run-time classes
	static Bool				is_init_ok;			//	TRUE => end of initialization
	
	static ClassDef*		list_head;			//	first class (used recently)
	static ClassDef*		list_tail;			//	last class (used a long time ago)
	
	static OPRuntime*		real_head;			//	real run-time list head
	static OPRuntime*		real_tail;			//	real run-time list tail
	
public:
	static void Initialise ();
	static void Kill ();
	static Bool IsInitOk ()		{ return OPClassMan::is_init_ok; }
	
	static Count TotalCount ()	{ return class_num; }
	
	static Bool GetClass (Count pos, OPRuntimeOut runtime);
	static Bool GetClass (OPAtomIn name, OPRuntimeOut runtime);
	static Bool GetClass (OPStringIn name, OPRuntimeOut runtime);
	static Bool GetClass (const LatinChar* name, OPRuntimeOut runtime);
	
	//	Functions used by the OP_NEW/OP_DELETE macros :
	
	static OPObject* New (OPAtomIn name, OPMemZone* zone, Count, const LatinChar*);
	static OPObject* New (OPStringIn name, OPMemZone* zone, Count, const LatinChar*);
	static OPObject* New (OPRuntimeIn name, OPMemZone* zone, Count, const LatinChar*);
	static OPObject* New (const LatinChar* name, OPMemZone* zone, Count, const LatinChar*);
};

#endif
