/*
 *	objectbrowser.h
 *
 *	The Object Browser class collects object panes into a single
 *	view which allows horizontal browsing.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_OBJECTBROWSER_H_
#define	_OPaC_OBJECTBROWSER_H_

#ifndef	_OPaC_OBJECTPANE_H_
#include <OPaC/objectpane.h>
#endif

/****************************************************************/

/*
 *	The Object browser collects panes into a single view.
 */

class OP_PUBLIC OPObjectBrowser : public OPView
{

protected:
	OPObjectPaneArray	panes;					//	array of object panes
	OPIconArray			icons;					//	array of large icons
	Count				num_panes;				//	current # of useful panes
	Count				max_panes;				//	maximum # of panes (including hidden)
	Count				vis_panes;				//	visible # of panes
	Count				zero_pane;				//	index of first (leftmost) pane
	OPCoord				dx;						//	horizontal size of a pane
	OPCoord				dy;						//	vertical size of a pane
	Real				visible_offset;			//	offset for SetVisibleValue...
	Bool				icons_enabled;			//	TRUE => icons enabled
	OPObjectPaneMode	pane_mode;				//	pane mode
	
	OPGenericProviderRef provider;				//	provider for the root pane
	OPObjectProviderRef	 empty_provider;		//	empty provider
	OPListRef			 empty_list;			//	empty list
	OPAtomRef			 model;					//	model name for the cell grids
	
protected:
	OPObjectBrowser ();
	virtual ~ OPObjectBrowser ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void Cleanup ();
	
	virtual void ChangeGeometry (OPRectIn box);			/* OVERRIDDEN */
	virtual void UpdateGeometry ();
	virtual void CreateEmptyPane (Count index);
	virtual void MakePaneVisible (Count index, Bool notify_icons, Bool notify_value);
	
	virtual OPCoord ReturnHorizontalPanePos (Count index) const;
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Initialise (Count panes, OPGenericProviderIn provider);
	virtual void Initialise (Count panes, OPGenericProviderIn provider, OPAtomIn model);
	
	virtual Bool EnableIcons (Bool yes);
	virtual Bool SetPaneMode (OPObjectPaneMode mode);
	
	virtual Bool ScrollLeft ();
	virtual Bool ScrollRight ();
	
	virtual Bool SetLevel (Count index, Bool notify = TRUE);
	virtual Bool SetZeroPane (Count index, Bool notify_icons = TRUE, Bool notify_value = TRUE);
	
	virtual Count ReturnLevel () const;
	virtual Count ReturnZeroPane () const;
	virtual Count ReturnNumPanes () const;
	virtual Count ReturnVisiblePanes () const;
	
	virtual Bool SetVisibleValue (Real value);
	virtual Real ReturnVisibleValue () const;
	virtual Real ReturnVisibleRatio () const;
	
	virtual Bool GetIcon (Count nth, OPIconOut icon, OPCoordOut x) const;
	virtual Bool GetPane (Count nth, OPObjectPaneOut pane) const;
	virtual Bool GetProvider (Count nth, OPGenericProviderOut provider) const;
	virtual Bool GetSelectedPath (OPStringOut path) const;
	
	virtual Bool CellActivated (OPObjectIn observer, Card32 id);
	virtual Bool CellDeactivated (OPObjectIn observer, Card32 id);
	
	OP_DEFINE_SLOT (Real, visible_value, SetVisibleValue, ReturnVisibleValue);
	OP_DEFINE_SLOT_READ (Real, visible_ratio, ReturnVisibleRatio);
	OP_DEFINE_CLASS (OPObjectBrowser, OPView);
};

/****************************************************************/

/*
 *	The Browser Provider is to be used with the object
 *	browser. It provides information about an object's
 *	data fields.
 */

class OP_PUBLIC OPObjBProvider : public OPGenericProvider
{
protected:
	OPObjectRef			object;					//	object which is being inspected
	Count				num_fields;				//	total # of real fields in object
	Bool				is_container;			//	TRUE => object is a container
	
protected:
	OPObjBProvider ();
	virtual ~ OPObjBProvider ();
	
	virtual void StoreInstance (OPStorageIn storage);		/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);		/* OVERRIDDEN */
	
	virtual Bool FindDeepObject (OPObjectInOut object);
	
public:
	virtual void Initialise (OPObjectIn object);
	
	virtual Count ReturnNumObjects ();						/* OVERRIDDEN */
	
	virtual Bool GetObject (Count index, OPObjectOut);		/* OVERRIDDEN */
	virtual Bool GetText (Count index, OPStringOut);		/* OVERRIDDEN */
	virtual Bool GetLargeIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	virtual Bool GetSmallIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	virtual Bool GetMarkIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	
	virtual Bool EnterSubtree (Count, OPGenericProviderOut);/* OVERRIDDEN */
	virtual Bool ReleaseSubtree (Count);					/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPObjBProvider, OPGenericProvider);
};

/****************************************************************/

/*
 *	Return informations about an object. This can be
 *	used to build an object browser.
 */

class OP_PUBLIC OPObjRootProvider : public OPGenericProvider
{
protected:
	OPObjectRef			object;					//	root object

protected:
	OPObjRootProvider ();
	virtual ~ OPObjRootProvider ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Initialise (OPObjectIn object);
	
	virtual Count ReturnNumObjects ();						/* OVERRIDDEN */
	
	virtual Bool GetObject (Count index, OPObjectOut);		/* OVERRIDDEN */
	virtual Bool GetText (Count index, OPStringOut);		/* OVERRIDDEN */
	virtual Bool GetLargeIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	virtual Bool GetSmallIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	virtual Bool GetMarkIcon (Count index, OPLookIconOut);	/* OVERRIDDEN */
	
	virtual Bool EnterSubtree (Count, OPGenericProviderOut);/* OVERRIDDEN */
	virtual Bool ReleaseSubtree (Count);					/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPObjRootProvider, OPGenericProvider);
};

/****************************************************************/

#endif /* _OPaC_OBJECTBROWSER_H_ */

