/*
 *	sd_memory.cxx
 *	
 *	Memory allocator for 32-bit memory chunks.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 18.11.97

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <stdlib.h>
#include "sd_interface.h"

#include <OPaC/memory.h>
#include <OPaC/object.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;

/****************************************************************/

static Count mem_zone_use    = 0;
static Count mem_big_alloc   = 0;
static Count mem_small_alloc = 0;

/*
 *	Create a system zone. This will not do anything yet on
 *	this implementation... just return the same zone ID
 *	for every call.
 */

Bool
SysMemOpenZone (const LatinChar* name,
				Card32Out zone_id)
{
	OP_FUNCTION (SysMemOpenZone, 2);
	OP_ARG (1, const LatinChar*, name);
	OP_ARG (2, Card32Out, zone_id);
	
	zone_id = 0;
	mem_zone_use++;
	
	return TRUE;
}



/*
 *	Close a memory zone. Don't do anything... just pretend
 *	we handle the call properly.
 */

Bool
SysMemCloseZone (Card32In zone_id)
{
	OP_FUNCTION (SysMemCloseZone, 1);
	OP_ARG (1, Card32In, zone_id);
	OP_ASSERT (zone_id == 0);
	OP_ASSERT (mem_zone_use != 0);
	
	mem_zone_use--;
	
	return TRUE;
}



/*
 *	The memory allocation function returns TRUE on success
 *	and sets allocated >= size to the real number of bytes
 *	which has been returned by the system.
 *
 */

Bool
SysMemAlloc (Card32In zone_id, Card32In size,
			 BytePtrOut ptr,
			 Card32Out allocated,
			 Card32Out mem_id)
{
	OP_FUNCTION (SysMemAlloc, 5);
	OP_ARG (1, Card32In, zone_id);
	OP_ARG (2, Card32In, size);
	OP_ARG (3, BytePtrOut, ptr);
	OP_ARG (4, Card32Out, allocated);
	OP_ARG (5, Card32Out, mem_id);
	
	OP_ASSERT (zone_id == 0);
	
	// Currently use malloc
		
		void* buffer = malloc ((size_t)(size));
		
		ptr       = (BytePtr)(buffer);
		allocated = (ptr ? size : 0);
		mem_id    = 0;
		
		mem_small_alloc += allocated;
		
		return (ptr ? TRUE : FALSE);
}


/*
 *	The memory free function will have to free the memory
 *	which has been allocated previously by SysMemAlloc.
 */

Bool
SysMemFree (Card32In zone_id,
			BytePtrIn ptr,
			Card32In allocated,
			Card32In mem_id)
{
	OP_FUNCTION (SysMemFree, 4);
	OP_ARG (1, Card32In, zone_id);
	OP_ARG (2, BytePtrIn, ptr);
	OP_ARG (3, Card32In, allocated);
	OP_ARG (4, Card32In, mem_id);
	
	OP_ASSERT (zone_id == 0);
	
	//	Memory had been allocated using "malloc", just use "free" now.
		
	mem_small_alloc -= allocated;
	free (ptr);
		
	return TRUE;
}

