/*
 *	sd_event.cxx
 *
 *	Portable event interface. Not every event will be returned to
 *	the application, since some need to be handled by the WNDPROC
 *	or else, Win32 wouldn't be very glad...
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <string.h>
#include <stdio.h>

#include "sd_interface.h"

#include <OPaC/app.h>
#include <OPaC/container.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/values.h>
#include <OPaC/window.h>

/****************************************************************/

#define	OP_TIMER_TICK	(5000)			//	timer tick interval [ms]

OP_IMPLEMENTATION_FILE;

/****************************************************************/

static UINT   timer_id      = 0;
static HANDLE pipe_handle   = 0;
static int    capture_count = 0;
static int    capture_grab  = 0;
static HWND   capture_hwnd  = NULL;

OPEventRef OPEvent::current = 0;


/****************************************************************/

static void FindSourceWindow (MessagePtr, OPWindowOut);
static void UpdateButtonState (Card16, Card32&);
static void HandleButtonEvent (Card32, HWND);
static Bool HandleKeyUpDown (MessagePtr, Card32&, OPKeyOut, Card32&, LatinChar*);

/****************************************************************/

/*
 *	Try to identify the event source window and return it
 *	to the caller. If it does not belong to OPaC, then an
 *	invalid window will be returned.
 */

static void
FindSourceWindow (HWND hwnd, OPWindowOut window)
{
	OP_FUNCTION (FindSourceWindow, 2);
	OP_ARG (1, HWND, hwnd);
	OP_ARG (2, OPWindowOut, window);
	
	LONG ext_wdo = 0;
	LONG magic   = 0;
	char buffer[200];

	if (hwnd == 0) {
		window = 0;
		return;
	}
	
	ext_wdo = GetWindowLong (hwnd, 0);
	magic   = GetWindowLong (hwnd, 4);
	
	GetClassName (hwnd, buffer, 200);
	
	if ( (ext_wdo != 0)
	  && (magic == GRAIL_MAGIC)
	  && (strncmp (buffer, OPaC_GrWindow::globname, strlen (OPaC_GrWindow::globname)) == 0) ) {
		
		//	Ok, we got an event coming from some window belonging to the OPaC
		//	class library. The only way we can know for sure is to compare the
		//	window's class name with the magic key.
		
		OPGrailWindow* grail = (OPGrailWindow*)(ext_wdo);
		
		grail->GetWindow (window);
		
	} else {
		
		//	The window from where the event comes from does not belong to the
		//	OPaC class library. So we don't know about the window !
		
		window = 0;
	}
}

/****************************************************************/

/*
 *	Update the button state according to the parameter from
 *	Window's event message. This may be used to be sure not
 *	to miss events in case Win16 would forget to send one.
 */

static void
UpdateButtonState (Card16 param, Card32& buttons)
{
	OP_FUNCTION (UpdateButtonState, 2);
	OP_ARG (1, Card16, param);
	OP_ARG (2, Card32&, buttons);
	
	buttons &= ~ ( OP_EV_BUTTON_LEFT
	             | OP_EV_BUTTON_MIDDLE
	             | OP_EV_BUTTON_RIGHT );
	
	if (param & MK_LBUTTON) buttons |= OP_EV_BUTTON_LEFT;
	if (param & MK_MBUTTON) buttons |= OP_EV_BUTTON_MIDDLE;
	if (param & MK_RBUTTON) buttons |= OP_EV_BUTTON_RIGHT;
}

/****************************************************************/

/*
 *	Hand the event capture from the current window to the
 *	specified one. This may be useful if some dragging op
 *	creates a new window and needs to give it the focus.
 */

void
OPEvent::HandOverToWindow (OPWindowIn window)
{
	OP_MEMBER (HandOverToWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	OP_ASSERT (capture_count);
	OP_ASSERT (window.IsValid ());
	
	OPGrailWindowRef gwdo = window->ReturnGrailPort ();
	OPaC_GrWindow*   iwdo = gwdo->ReturnPrivateWindowInfo ();

	if (iwdo->window != capture_hwnd) {
		ReleaseCapture ();
		SetCapture (iwdo->window);
		capture_hwnd = iwdo->window;
	}
}

void
OPEvent::GrabWindowEvents (OPWindowIn window, Bool mode)
{
	OP_MEMBER (GrabWindowEvents, 2);
	OP_ARG (1, OPWindowIn, window);
	OP_ARG (2, Bool, mode);
	
	capture_grab += mode ? 1 : -1;
	
	if (capture_grab < 0) {
		capture_grab = 0;
	}
}


/*
 *	When a mouse button gets pressed, we have to grab it
 *	so we will get the release event. This is called an
 *	event "capture" by the Win16 system.
 */

static void
HandleButtonEvent (Card32 event, HWND hwnd)
{
	OP_FUNCTION (HandleButtonEvent, 2);
	OP_ARG (1, Card32, event);
	OP_ARG (2, HWND, hwnd);
	
	switch (event) {
		case OP_EVENT_PRESS:
			if (capture_count++ == 0) {
				SetCapture (hwnd);
				capture_hwnd = hwnd;
			}
			break;
		
		case OP_EVENT_RELEASE:
			if (--capture_count == 0) {
				ReleaseCapture ();
				capture_hwnd = NULL;
			}
			break;
		
		default:
			OP_ABORT ("Bad event recieved");
			break;
	}
	
	//	This is for debugging purposes only. This should raise problems only
	//	if we don't get the right events :
	
	OP_ASSERT (capture_count >= 0);
	OP_ASSERT (capture_count <= 3);
}

/****************************************************************/

/*
 *	Handle a key event (up, down or repeat). This might be
 *	produced due to some automatic repetition or to some
 *	newly pressed key, or even a SHIFT/CTRL key pressure.
 */

static Bool
HandleKeyUpDown (MessagePtr message,
				 Card32& event_type,
				 OPKeyOut key,
				 Card32& key_oem,
				 LatinChar* key_name)
{
	OP_FUNCTION (handle_key_updown, 5);
	OP_ARG (1, MessagePtr, message);
	OP_ARG (2, Card32&, event_type);
	OP_ARG (3, OPKeyOut, key);
	OP_ARG (4, Card32&, key_oem);
	OP_ARG (5, LatinChar*, key_name);
	
	//	Extract the pertinent data from the Win16 event message.
	
	Card32 info     = message->lParam;					//	flags & bits
	Card16 vir_code = message->wParam;					//	virtual key code
	Card16 oem_code = (Card16)((info >> 16) & 0x00FF);	//	OEM key code
	Card32 mod_code = 0;								//	modifier code(s) 
	
	Bool is_key_rep = (info & OP_BIT (30)) ? TRUE : FALSE;
	Bool is_key_rel = (info & OP_BIT (31)) ? TRUE : FALSE;
	Bool is_key_ext = (info & OP_BIT (24)) ? TRUE : FALSE;
	Bool is_key_alt = (info & OP_BIT (29)) ? TRUE : FALSE;
	
	if (GetKeyState (VK_SHIFT)   & 0x8000) mod_code |= OP_EV_SHIFT;
	if (GetKeyState (VK_CONTROL) & 0x8000) mod_code |= OP_EV_CONTROL;
	if (GetKeyState (VK_MENU)    & 0x8000) mod_code |= OP_EV_META_1;
	if (GetKeyState (VK_CAPITAL) & 0x0001) mod_code |= OP_EV_CAPS_LOCK;
	
	//	Define default key information (only update the modifiers). If other
	//	data has to be returned, it will get initialised later on.
	
	event_type = OP_EVENT_NONE;
	
	key.modifiers = mod_code;
	key.type      = 0;
	key.value     = 0;
	key.code      = 0;
	
	key_name[0] = 0;
	key_oem     = 0;
	
	//	Is the key down event referring to a shift or modifier key change ?
	
	if ( (vir_code == VK_SHIFT)
	  || (vir_code == VK_CONTROL)
	  || (vir_code == VK_MENU)
	  || (vir_code == VK_CAPITAL) ) {
		
		//	The OP_EVENT_KEY_CHANGE event will be the only one to be produced,
		//	if the caller detects some change in the modifier state...
		
		event_type = OP_EVENT_KEY_CHANGE;
		return FALSE;
	}
	
	//	Since the pressed/released key doesn't seem to be virtual, it must have
	//	been some real key. Investigate...
	
	LatinChar  buffer[16];
	LatinChar* name    = "";
	Card16     unicode = 0;
	Card16     keycode = 0;
	
	{
		//	Here we do some real 100% Win16 job. If you only knew how much I
		//	love Micr*s*ft for this ! By the way, did you know that Bill Gates'
		//	wife said that he was both "Micro and Soft" ;-) ?
		
		BYTE  k_state[256];
		DWORD win_key = 0;
		
		GetKeyboardState (k_state);
		
#ifdef WIN32
#define	TMPPTR1	PBYTE
#define	TMPPTR2	LPWORD
#else
#define	TMPPTR1	BYTE FAR*
#define	TMPPTR2	DWORD FAR*
#endif
		if ( (ToAscii ( (UINT)(vir_code), (UINT)(info >> 16),
						(TMPPTR1)(k_state), (TMPPTR2)(& win_key),
						(UINT)(0) )) == 1) {
			
			unicode = (Card16)(win_key & 0x0000FFFF);
			keycode = unicode;
			
			if ( (keycode >= 'a') && (keycode <= 'z') ) {
				keycode = keycode + 'A' - 'a';
			}

			if ( (keycode >= ('A'-'@')) && (keycode <= ('Z'-'@')) ) {
				keycode = keycode + '@';
			}
		}
	}
	
#undef	TMPPTR1
#undef	TMPPTR2
	
	const Card32 event_type_table[4] = {	/*  rel  |  rep  */
		OP_EVENT_KEY_DOWN,					/*   0   |   0   */
		OP_EVENT_KEY_UP,					/*   1   |   0   */
		OP_EVENT_KEY_REP,					/*   0   |   1   */
		OP_EVENT_KEY_UP						/*   1   |   1   */
	};
	
	//	Look for event type and assume that the key will be alphanumerical. If
	//	not, the key type will just have to be changed afterwards.
	
	event_type = event_type_table[is_key_rel + 2*is_key_rep];
	key.type   = OP_KEY_ALPHANUM;
	
	//	Fill in the keyboard event description, depending on Win16' virtual key
	//	code.
	
	switch (vir_code & 0x00FF) {
		
		//	Special keys with no values or with implicit values :
		
		case VK_BACK:	   name = "BACKSPACE"; key.type  = OP_KEY_BACKSPACE; break;
		case VK_CLEAR:	   name = "CLEAR";     key.type  = OP_KEY_SPECIAL;   break;
		case VK_DELETE:    name = "DELETE";    key.type  = OP_KEY_DELETE;    break;
		case VK_ESCAPE:	   name = "ESCAPE";    key.type  = OP_KEY_SPECIAL;   break;
		case VK_EXECUTE:   name = "EXECUTE";   key.type  = OP_KEY_SPECIAL;   break;
		case VK_HELP:      name = "HELP";      key.type  = OP_KEY_SPECIAL;   break;
		case VK_INSERT:    name = "INSERT";    key.type  = OP_KEY_SPECIAL;   break;
		case VK_NUMLOCK:   name = "NUM LOCK";  key.type  = OP_KEY_SPECIAL;   break;
		case VK_PAUSE:	   name = "PAUSE";     key.type  = OP_KEY_SPECIAL;   break;
		case VK_PRINT:	   name = "PRINT";     key.type  = OP_KEY_SPECIAL;   break;
		case VK_RETURN:	   name = "RETURN";    key.type  = OP_KEY_RETURN;    break;
		case VK_SCROLL:    name = "SCRL LOCK"; key.type  = OP_KEY_SPECIAL;   break;
		case VK_SELECT:	   name = "SELECT";    key.type  = OP_KEY_SPECIAL;   break;
		case VK_SEPARATOR: name = "SEPARATOR"; key.type  = OP_KEY_SPECIAL;   break;
		case VK_SNAPSHOT:  name = "SNAPSHOT";  key.type  = OP_KEY_SPECIAL;   break;
		case VK_SPACE:	   name = "SPACE";     key.value = keycode = ' ';    break;
		case VK_TAB:	   name = "TAB";       key.type  = OP_KEY_TAB;       break;
		
		//	Cursor keys, with value specifying the kind of cursor move which has
		//	been asked for :
		
		case VK_PRIOR: name = "PAGE UP";   key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_PAGE_UP;   break;
		case VK_NEXT:  name = "PAGE DOWN"; key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_PAGE_DOWN; break;
		case VK_END:   name = "END";       key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_END;       break;
		case VK_HOME:  name = "HOME";      key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_HOME;      break;
		case VK_LEFT:  name = "LEFT";      key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_LEFT;      break;
		case VK_UP:    name = "UP";        key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_UP;        break;
		case VK_RIGHT: name = "RIGHT";     key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_RIGHT;     break;
		case VK_DOWN:  name = "DOWN";      key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_DOWN;      break;
		
		//	Numeric pad key, which are considered to be normal keys by OPaC, but
		//	not by the Win16 API.
		
		case VK_NUMPAD0: case VK_NUMPAD1: case VK_NUMPAD2: case VK_NUMPAD3: case VK_NUMPAD4:
		case VK_NUMPAD5: case VK_NUMPAD6: case VK_NUMPAD7: case VK_NUMPAD8: case VK_NUMPAD9:
			key.value = keycode = vir_code - VK_NUMPAD0 + '0';
			break;
		
		case VK_MULTIPLY: key.value = keycode = '*'; break;
		case VK_ADD:      key.value = keycode = '+'; break;
		case VK_SUBTRACT: key.value = keycode = '-'; break;
		case VK_DECIMAL:  key.value = keycode = '.'; break;
		case VK_DIVIDE:   key.value = keycode = '/'; break;
		
		case VK_F1:  case VK_F2:  case VK_F3:  case VK_F4:
		case VK_F5:  case VK_F6:  case VK_F7:  case VK_F8:
		case VK_F9:  case VK_F10: case VK_F11: case VK_F12:
		case VK_F13: case VK_F14: case VK_F15: case VK_F16:
		case VK_F17: case VK_F18: case VK_F19: case VK_F20:
		case VK_F21: case VK_F22: case VK_F23: case VK_F24:
			
			if ( (vir_code - VK_F1 + 1) > 9) {
				buffer[0] = 'F';
				buffer[1] = (LatinChar)((vir_code - VK_F1 + 1) / 10) + '0';
				buffer[2] = (LatinChar)((vir_code - VK_F1 + 1) % 10) + '0';
				buffer[3] = 0;
			} else {
				buffer[0] = 'F';
				buffer[1] = (LatinChar)(vir_code - VK_F1 + 1) + '0';
				buffer[2] = 0;
			}
			
			name      = buffer;
			key.type  = OP_KEY_FUNCTION;
			key.code  = vir_code - VK_F1 + 1;
			break;
		
		default:
			key.value = unicode;
			break;
	}
	
	key_oem = oem_code | ((is_key_ext) ? 0x0100 : 0x0000);
	
	if ( (key.value)
	  && (key.code == 0) ) {
		key.code  = (Card8)(keycode);
	}
	
	while (*key_name++ = *name++) {
		/* nothing */
	}
	
	return TRUE;
}


/****************************************************************/

extern OP_PUBLIC HANDLE g_wait_handles[MAXIMUM_WAIT_OBJECTS];
extern OP_PUBLIC Count  g_num_handles;

HANDLE  g_wait_handles[MAXIMUM_WAIT_OBJECTS] = { 0 };
Count	g_num_handles  = 0;

/****************************************************************/


/*
 *	Wait for an event and fill in the event record. This
 *	should only be called by the OPApp class.
 *
 *	Returns FALSE when no more event can be delivered by
 *	the system (the application has been killed).
 */

Bool
OPEvent::WaitForEvent ()
{
	OP_METHOD (WaitForEvent, 0);
	
	static Bool        move_event_ok  = FALSE;
	static OPWindowRef move_event_wdo = OPWindowNull;
	
	static Card32 chg_buttons   = 0;
	static Card32 chg_modifiers = 0;
	static HWND   last_hwnd     = 0;
	
	static Card32    pending_event_type = OP_EVENT_NONE;
	static OPKey     pending_key;
	static Card32    pending_oem        = 0;
	static LatinChar pending_name[16]   = { 0 };
	
	//	In case we were dead, just resurrect the application !
	
	this->is_dead = FALSE;
	this->type    = OP_EVENT_NONE;
	
	while ( (this->is_dead == FALSE)
	     && (this->type == OP_EVENT_NONE) ) {
		
		//	The event record does not contain valid information other than the
		//	window, the mouse state & position and the keyboard modifiers.
		
		this->id++;
		
		this->change = 0;
		this->data   = 0;
		this->name   = "";
		this->code   = 0;
		
		if (chg_modifiers) {
			
			//	At least one modifier key has changed. Notify about the change
			//	but it is up to the user to find out whether it was a press or
			//	a release event.
			//
			//	Only one modifier change is being signaled at a given time.
			
			Card32 mask = 0x00000001;
			
			while ((mask & chg_modifiers) == 0) {
				chg_modifiers *= 2;
			}
			
			chg_modifiers ^= mask;
			this->type = OP_EVENT_KEY_CHANGE;
			this->change = mask;
			this->key.modifiers ^= mask;
			break;
		}
		
		if (chg_buttons) {
			
			//	At least one button has changed. Notify about the change and
			//	find out if it was a press or a release event.
			//
			//	Only one button change can be signaled at a given time.
			
			Card32 mask = 0x00000001;
			
			while ((mask & chg_buttons) == 0) {
				mask *= 2;
			}
			
			chg_buttons   ^= mask;
			this->type     = (this->buttons & mask) ? OP_EVENT_RELEASE : OP_EVENT_PRESS;
			this->change   = mask;
			this->buttons ^= mask;
			
			HandleButtonEvent (this->type, last_hwnd);
#if 0
			move_event_ok  = TRUE;
#endif
			move_event_wdo = this->window;
			break;
		}
		
		if (pending_event_type != OP_EVENT_NONE) {
			
			//	Return the pending keyboard event. This could not be returned by
			//	a previous call to this method, since it was part of a multiple
			//	event message...
			
			this->type   = pending_event_type;
			this->key    = pending_key;
			this->code   = pending_oem;
			this->name   = pending_name;
			
			pending_event_type = OP_EVENT_NONE;
			
			break;
		}
		
		
		//	Wait for some real event.
		
		MSG    message   = { 0, 0, 0, 0, 0, 0 };
		Bool   dispatch  = FALSE;
		Card32 buttons   = this->buttons;
		Card32 modifiers = this->key.modifiers;
		
		while (PeekMessage (&message, (HWND)(NULL), 0, 0, PM_REMOVE | PM_NOYIELD) == FALSE) {
			
			//	No message available in the queue. Wait for the next message (or for
			//	some external synchronisation event).
			
			DWORD timeout = INFINITE;
			DWORD result  = MsgWaitForMultipleObjects (g_num_handles, g_wait_handles,
													   FALSE, timeout, QS_ALLEVENTS);
			
			if ( (result > WAIT_OBJECT_0)
			  && (result < (WAIT_OBJECT_0 + g_num_handles)) ) {
				
				//	Got a synchronisation event...
				
				OP_TRACE ("Wait satisfied in event loop (%d)\n", result - WAIT_OBJECT_0);
				
			} else if (result == WAIT_OBJECT_0 + g_num_handles) {
				
				//	Got a user event
				
				OP_TRACE ("User event: ");
				
			} else if (result == WAIT_TIMEOUT) {
				
				OP_TRACE ("Timeout in event loop\n");
				
			} else if (result == WAIT_OBJECT_0) {
				
				OP_TRACE ("Change in global wait table\n");
				
			} else {
				
				OP_TRACE ("Bad Msg caught in event loop\n");
			}
		}

		OP_TRACE ("%d\n", message.message);
		
		FindSourceWindow (message.hwnd, this->window);
		
		last_hwnd = message.hwnd;
		
		//	Translate the Win16 event into some standardized OPaC compatible event
		//	description. Some single events will produce more than one OPaC event,
		//	and vice versa.
		
		switch (message.message) {
			
			case WM_QUIT:
				this->is_dead = TRUE;
				break;
			
			case WM_MOUSEMOVE:
			case WM_LBUTTONDOWN:
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDOWN:
			case WM_LBUTTONUP:
			case WM_RBUTTONUP:
			case WM_MBUTTONUP:
				
				this->type = OP_EVENT_MOTION;
				this->x    = (OPRelCoord)((Int16)(LOWORD (message.lParam)));
				this->y    = (OPRelCoord)((Int16)(HIWORD (message.lParam)));
#if 0
				//	If we had captured the mouse input, we will get messages for the
				//	wrong window. Handle this special case here...
				//
				//	If event grabbing has been explicitely requested, don't do any
				//	magical capture changes.
				
				if ( (capture_count)
				  && (capture_grab == 0) ) {
					
					POINT point;
					RECT  rect;
					
					point.x = LOWORD (message.lParam);
					point.y = HIWORD (message.lParam);
					
					GetWindowRect (message.hwnd, &rect);
					
					point.x += rect.left;
					point.y += rect.top;
					
					last_hwnd = WindowFromPoint (point);
					FindSourceWindow (last_hwnd, this->window);
					
					if (last_hwnd != message.hwnd) {
						
						//	The mouse moved to another window. Find out in which window it will
						//	be and notify the previous window that the mouse just exited.
						
						if ( (move_event_ok)
						  && (move_event_wdo.IsValid ()) ) {
							
							move_event_wdo->LeaveNotify (this);
							
							//	The windows might have changed since the last test. Try again
							//	to see which window will become active.
							
							last_hwnd      = WindowFromPoint (point);
							move_event_wdo = 0;
						}
						
						FindSourceWindow (last_hwnd, this->window);
						
						if ( (this->window.IsValid ())
						  && (this->window->IsVisible ()) ) {
							this->HandOverToWindow (this->window);
							move_event_wdo = this->window;
						}
					}
				}
#endif
				if ( (move_event_ok)
				  && (move_event_wdo != this->window) ) {
					if (move_event_wdo.IsValid ()) {
						move_event_wdo->LeaveNotify (this);
					}
				}
				
				move_event_ok  = TRUE;
				move_event_wdo = this->window;
				
				//	In case Win16 has forgotten to tell us about some button change,
				//	we can resynchronize the button events here.
				
				UpdateButtonState (message.wParam, buttons);
				break;
			
			case WM_ACTIVATE:
				//	msg.wParam != WA_INACTIVE
				break;
			
			case WM_SIZE:
				//	msg.wParam == SIZE_MAXIMIZED
				//	msg.wParam == SIZE_MINIMIZED
				break;
			
			case WM_KEYDOWN:
			case WM_KEYUP:
			case WM_SYSKEYDOWN:
			case WM_SYSKEYUP:
				
				//	Usually, the event will be reported, but it might be swallowed if it
				//	is of no interest to OPaC (shift key repetition is some quite silly
				//	thing implemented by these PCs).
				
				HandleKeyUpDown (&message, pending_event_type,
								 pending_key, pending_oem, pending_name);
				break;
			
			case WM_TIMER:
				
				OPApp::CurrentApp ()->HandleTick (OP_TIMER_TICK);
				
				if ( (move_event_ok)
				  && (move_event_wdo.IsValid ()) ) {
					
					POINT            point;
					OPGrailWindowRef grail;
					
					if (GetCursorPos (&point)) {
						if (OPGrailWindow::FindOwnWindow (point.x, point.y, grail)) {
							
							//	Still in an OPaC window. No problem... we know in which window
							//	the mouse cursor is.
							
						} else {
							
							//	Left the latest used OPaC window. Notify the mouse exit.
							
							move_event_wdo->LeaveNotify (this);
							
							move_event_ok  = FALSE;
							move_event_wdo = OPWindowNull;
						}
					}
				}
				
				break;
			
			case WM_CLOSE:
				this->window->SysChangedOpen (FALSE);
				
				//	It is important to let the windows procedure know about the CLOSE
				//	event, since it might have been generated by that very procedure,
				//	which has to track the real number of closes which have occurred.

				dispatch = TRUE;
				break;
			
			case WM_SYSCOMMAND:					//	system menu has issued a command
				
				switch (message.wParam & 0xFFF0) {
					
					case SC_CLOSE:				//	close a window ?
						OPEvent::PostQuitEvent ();
						break;
					
					default:
						dispatch = TRUE;
						break;
				}
				break;
			
			case WM_USER:
				this->type   = OP_EVENT_USER;
				this->code   = message.lParam;
				this->change = message.wParam;
				this->name   = "USER";
				break;
				
			default:
				if ( (message.message >= 0xC000)
				  && (message.message < 0xFFFF)
				  && (OPEvent::user_events.IsValid ()) ) {
					
					OPAtomRef  key;
					OPValueRef value;
					Bool       ok = FALSE;
					Count index = 0;
					
					while (OPEvent::user_events->Get (index++, key, value)) {
						if (value->ReturnCard32 () == message.message) {
							this->type   = OP_EVENT_USER;
							this->code   = message.lParam;
							this->change = message.wParam;
							this->name   = key;
							ok = TRUE;
							break;
						}
					}
					
					if (ok) break;
				}
				
				dispatch = TRUE;
				break;
		}
		
		//	Compute the button and modifier changes.
		
		chg_buttons   = buttons ^ this->buttons;
		chg_modifiers = modifiers ^ this->key.modifiers;
		
		//	After having successfully translated an event, it has to
		//	be communicated to Windows, otherwise it would freeze :-(
		
		if (dispatch) {
			TranslateMessage (& message);
			DispatchMessage (& message);
		}
	}
	
	return this->is_dead ? FALSE : TRUE;
}

/****************************************************************/

/*
 *	Send a quit event to ourself, so that we can exit that
 *	bloody event loop :-).
 */

void
OPEvent::PostQuitEvent ()
{
	OP_MEMBER (PostQuitEvent, 0);
	::PostQuitMessage (0);
}



/*
 *	Initialise the event-loop related resources (eg. the
 *	timer).
 */

void
OPEvent::Initialise ()
{
	OP_MEMBER (Initialise, 0);
	OP_ASSERT (timer_id == 0);
	
	if (OPaC_GrWindow::globname[0] == 0) {
		snprintf (OPaC_GrWindow::globname, sizeof (OPaC_GrWindow::globname),
				  GLOBNAME_ID "%lx:%lx",
				  (long)(GetCurrentThreadId ()),
				  (long)(GetCurrentProcessId ()));
	}

	OPEvent::current = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPEvent);
	
	char pipe_name[256];
	
	strcpy (pipe_name, "\\\\.\\pipe\\");
	strcat (pipe_name, OPaC_GrWindow::globname);
	
	timer_id    = SetTimer (NULL, 0, OP_TIMER_TICK, NULL);
	pipe_handle = CreateNamedPipe (pipe_name,
								   PIPE_ACCESS_DUPLEX,
								   PIPE_TYPE_BYTE | PIPE_READMODE_BYTE,
								   1, 0, 0, 0, NULL);
	
	OP_ASSERT (timer_id != 0);

	g_wait_handles[0] = CreateEvent (NULL, FALSE, FALSE, NULL);
	g_num_handles     = 1;
}



/*
 *	Clean-up before exiting the application. This should
 *	undo what the "OPaC_InitEvent" method set up.
 */

void
OPEvent::Kill ()
{
	OP_MEMBER (Kill, 0);
	OP_ASSERT (timer_id != 0);
	
	OPEvent::current = 0;
	CloseHandle (pipe_handle);
	
	KillTimer (NULL, timer_id);
}

/****************************************************************/

OPDictRef OPEvent::user_events;


Bool
OPEvent::InsertUserEvent (OPAtomIn name)
{
	OP_MEMBER (InsertUserEvent, 1);
	OP_ARG (1, OPAtomIn, name);
	
	if (OPEvent::user_events.IsInvalid ()) {
		OPEvent::user_events = OP_NEW (OPDict);
	}
	if (OPEvent::FindUserEvent (name)) {
		return FALSE;
	}
	
	const LatinChar* string = name->ReturnQuickLatin ();
	OPValueRef value = RegisterWindowMessage (string);
	
	return OPEvent::user_events->Insert (name, value);
}

Bool
OPEvent::RemoveUserEvent (OPAtomIn name)
{
	if (OPEvent::user_events.IsInvalid ()) {
		return FALSE;
	}
	
	return OPEvent::user_events->Remove (name);
}

Bool
OPEvent::FindUserEvent (OPAtomIn name)
{
	if (OPEvent::user_events.IsInvalid ()) {
		return FALSE;
	}
	
	return OPEvent::user_events->LookFor (name);
}

