/*
 *	sd_pipeserver.cxx
 *
 *	The pipe server is the server part of a remote connection;
 *	this is machine dependent.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#include <OPaC/pipestream.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPPipeServer);

/****************************************************************/

#define	DEFAULT_TIMEOUT	60.0		//	just a minute

/****************************************************************/

struct OPaC_SPipe
{
	HANDLE				handle;
	Bool				is_created;
	Bool				is_connected;
};

/****************************************************************/

OPPipeServer::OPPipeServer ()
{
	this->pipe = NEW (OPaC_SPipe);

	this->pipe->handle       = 0;
	this->pipe->is_created   = FALSE;
	this->pipe->is_connected = FALSE;
	
	this->server_name = OP_NEW (OPString);
	this->pipe_name   = OP_NEW (OPString);
	this->timeout     = DEFAULT_TIMEOUT;
}

OPPipeServer::~OPPipeServer ()
{
	if (this->pipe->is_connected) {
		this->Close ();
	}
	
	if (this->pipe->is_created) {
		this->Kill ();
	}
	
	DELETE (this->pipe);
}

void
OPPipeServer::StoreInstance (OPStorageIn storage)
{
	OP_ABORT ("Trying to save Pipe Server");
}

void
OPPipeServer::RestoreInstance (OPStorageIn storage)
{
	OP_ABORT ("Trying to restore Pipe Server");
}


Bool
OPPipeServer::SetServerName (OPStringIn name)
{
	return this->server_name->SetString (name);
}


Bool
OPPipeServer::SetPipeName (OPStringIn name)
{
	return this->pipe_name->SetString (name);
}


Bool
OPPipeServer::SetTimeOut (Real time)
{
	this->timeout = time;
	return TRUE;
}


void
OPPipeServer::Create ()
{
	if (this->pipe->is_connected) {
		OP_ASSERT (this->pipe->is_connected == FALSE);
	}
	if (this->pipe->is_created) {
		OP_ASSERT (this->pipe->is_created == FALSE);
	}
	
	OPStringRef name = OP_NEW (OPString);
	DWORD       time = (DWORD)(timeout*1000);
	
	name->AppendString ("\\\\.\\pipe\\");
	
	if ( (this->server_name.IsValid ())
	  && (this->server_name->CountCharacters ()) ) {
		name->AppendString (this->server_name);
		name->AppendString (":");
	}
	
	name->AppendString (this->pipe_name);
	name->ConvertToLatin ();
	
	this->pipe->handle = CreateNamedPipe (name->AccessLatin (),
										  PIPE_ACCESS_DUPLEX,
										  PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE | PIPE_WAIT,
										  PIPE_UNLIMITED_INSTANCES,
										  0, 0, time, NULL);

	if (this->pipe->handle == INVALID_HANDLE_VALUE) {
		OP_ASSERT (this->pipe->handle != INVALID_HANDLE_VALUE);
	}
	
	this->pipe->is_created = TRUE;
}


void
OPPipeServer::Open ()
{
	if (this->pipe->is_created == FALSE) {
		OP_ASSERT (this->pipe->is_created);
	}
	if (this->pipe->is_connected == TRUE) {
		OP_ASSERT (this->pipe->is_connected == FALSE);
	}
	
	Bool connect = FALSE;
	
	while (connect == FALSE) {
		connect = ConnectNamedPipe (this->pipe->handle, NULL);
		Sleep (50);
	}
	
	this->pipe->is_connected = TRUE;
}

Count
OPPipeServer::Read (void* buffer, Count size)
{
	if (this->pipe->is_connected == FALSE) {
		return 0;
	}
	
	Bool ok;
	DWORD rdsize = (DWORD)(size);

	ok = ReadFile (this->pipe->handle, buffer, rdsize, &rdsize, NULL);
	
	if (ok == FALSE) {
		return 0;
	}
	
	return (Count)(rdsize);
}


Count
OPPipeServer::Write (const void* buffer, Count size)
{
	if (this->pipe->is_connected == FALSE) {
		return 0;
	}
	
	Bool ok;
	DWORD wrsize = (DWORD)(size);

	ok = WriteFile (this->pipe->handle, buffer, wrsize, &wrsize, NULL);
	
	if (ok == FALSE) {
		return 0;
	}
	
	return (Count)(wrsize);
}


Bool
OPPipeServer::Sync ()
{
	if (this->pipe->is_connected == TRUE) {
		FlushFileBuffers (this->pipe->handle);
	}
	
	return TRUE;
}


void
OPPipeServer::Close ()
{
	if (this->pipe->is_connected == TRUE) {
		FlushFileBuffers (this->pipe->handle);
		DisconnectNamedPipe (this->pipe->handle);
		this->pipe->is_connected = FALSE;
	}
}

void
OPPipeServer::Kill ()
{
	if (this->pipe->is_connected) {
		this->Close ();
	}

	if (this->pipe->is_created) {
		CloseHandle (this->pipe->handle);
		this->pipe->is_created = FALSE;
	}
}

