/*
 *	array.cxx
 *
 *	Array class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/arrayit.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPArray);

/****************************************************************/

/*
 *	A subclass which needs to implement the way the data is stored
 *	in the array will have to override the following methods :
 *
 *	- void InternalGetGapSize (Count pos, CountOut gap, CountOut num)
 *	- void InternalSetGapSize (Count pos, Count gap, Count num)
 *	- Bool SetSize (Count size)
 *	- OPObjectRef& Access (Count pos) const
 *
 *	This should be enough. You may want to override other methods
 *	in order to increase efficiency...
 */

/****************************************************************/

/*
 *	An array consists of lots of references (unless this
 *	is overridden by some subclass).
 */

OPArray::OPArray ()
{
	OP_METHOD (OPArray, 0);
	
	this->references = 0;
}


OPArray::~OPArray ()
{
	OP_METHOD (~OPArray, 0);
	this->SetSize (0);
}


/*
 *	Store an array. If there are gaps in the array (information
 *	provided by the subclass), they won't be stored -- this will
 *	reduce the amount of useless data stored.
 */

void
OPArray::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPArray"
							   " { < gap_size data_size { data } > } >");
	
	storage << info;
	storage << OPStorageSync::begin;
	
	Count index = 0;
	Count count = this->num_element;
	
	storage << OPStorageSync::repeat;
	
	while (index < count) {
		
		Count gap_size;
		Count data_size;
		
		//	If there is a gap in the array, just move ahead and save the
		//	data after it. This helps reducing the amount of useless data
		//	being stored.
		
		this->InternalGetGapSize (index, gap_size, data_size);
		
		storage << OPStorageSync::begin;
		storage << gap_size << data_size;
		
		index += gap_size;
		
		storage << OPStorageSync::repeat;
		
		while (data_size--) {
			storage << this->Access (index++);
		}
		
		storage << OPStorageSync::end_repeat;
		storage << OPStorageSync::end;
	}
	
	storage << OPStorageSync::end_repeat;
	storage << OPStorageSync::end;
}


/*
 *	Restore an array.
 */

void
OPArray::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	this->RemoveAll ();
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::repeat;
	
	Count index = 0;
	Count count = this->num_element;
	
	this->num_element = 0;
	
	while (index < count) {
		
		Count gap_size  = 0;
		Count data_size = 0;
		
		storage >> OPStorageSync::begin;
		storage >> gap_size >> data_size;
		
		this->InternalSetGapSize (index, gap_size, data_size);
		
		index += gap_size;
		
		storage	>> OPStorageSync::repeat;
		
		while (data_size--) {
			OPObjectRef object;
			storage >> object;
			this->Access (index++) = object;
		}
		
		storage >> OPStorageSync::end;
		storage >> OPStorageSync::end;
	}
	
	storage	>> OPStorageSync::end;
	storage	>> OPStorageSync::end;
}


/*
 *	This should be overridden by the subclasses if a special
 *	handling of gaps has to be done.
 *
 *	OVERRIDE if subclass handles the reference access itself.
 */

void
OPArray::InternalGetGapSize (Count pos, CountOut gap, CountOut num )
{
	OP_METHOD (InternalGetGapSize, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, CountOut, gap);
	OP_ARG (3, CountOut, num);
	
	gap = 0;
	num = 0;
	
	//	What is the size of the gap ?
	
	while ( (pos < this->num_element)
		 && (this->references[pos].IsInvalid ()) ) {
		pos++;
		gap++;
	}
	
	//	And what is the size of the chunk following the gap ?
	
	while ( (pos < this->num_element)
		 && (this->references[pos].IsValid () == TRUE) ) {
		pos++;
		num++;
	}
}

void
OPArray::InternalSetGapSize (Count pos, Count gap, Count num)
{
	OP_METHOD (InternalSetGapSize, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, Count, gap);
	OP_ARG (3, Count, num);
	
	//	Creating a gap will simply grow the size of the (uninitialised)
	//	array.
	
	if (pos+gap > this->num_element) {
		this->SetSize (pos+gap);
	}
}


/*
 *	Delete the array of data.
 */

void
OPArray::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->SetSize (0);
	super::Delete ();
}


/*
 *	Allocate an array iterator.
 */

OPIteratorRef
OPArray::NewIterator ()
{
	OP_METHOD (NewIterator, 0);
	
	OPArrayIteratorRef iterator = OP_NEW (OPArrayIterator);
	this->InsertIterator (iterator);
	return iterator;
}


/*
 *	Insert an element at the end of the array.
 */

Bool
OPArray::Insert (OPObjectIn data)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPObjectIn, data);
	
	Count pos = this->num_element;
	
	this->SetSize (pos+1);
	this->Access (pos) = data;
	
	return TRUE;
}


/*
 *	Remove every element in the array. This is a synonym for
 *	SetSize (0) provided for compatibility with the Container
 *	classes.
 */

void
OPArray::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	this->SetSize (0);
}


/*
 *	Specifying an extraction has no effect. Don't accept any
 *	kind of specification, unless it is empty.
 */

Bool
OPArray::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	if ( (x.IsInvalid ())
	  || (x->TotalCount () == 0) ) {
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Return the "nth" element of the array.
 */

Bool
OPArray::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	if (index < this->num_element) {
		object = this->Access (index);
		return TRUE;
	}
	
	object = 0;
	return FALSE;
}


/*
 *	Return the object which is considered to be similar to the
 *	one provided as model. The comparison is based on the result
 *	of method "CompareWith".
 */

Bool
OPArray::GetSimilar (OPObjectIn model, OPObjectOut object, Count nth)
{
	OP_METHOD (GetSimilar, 3);
	OP_ARG (1, OPObjectIn, model);
	OP_ARG (2, OPObjectOut, object);
	OP_ARG (3, Count, nth);
	
	Count pos = 0;
	
	while (pos < this->num_element) {
		OPObjectRef& element = this->Access (pos++);
		if ( (element.IsValid ())
		  && (element->CompareWith (model) == OP_COMP_THIS_SAME)
		  && (nth-- == 0) ) {
			object = element;
			return TRUE;
		}
	}
	
	return FALSE;
}


/*
 *	Compare with another array. This compares the contents of
 *	the arrays. If both arrays hold the same objects (as defined
 *	by CompareWith), they are considered to be the same.
 */

int
OPArray::CompareWith (OPObjectIn object) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, OPObjectIn, object);
	
	if (object.IsInvalid ()) {
		return OP_COMP_THIS_AFTER;
	}
	
	if (object->IsKindOf (OP_RUNTIME (OPArray))) {
		
		OPArrayRef array = object;
		Count      index = 0;
		Count      min   = (this->num_element < array->num_element) ? 
							this->num_element : array->num_element;
		
		//	Compare the contents of both arrays. A null object is considered
		//	to come before a non-null object.
		
		while (index < min) {
			
			const OPObjectRef& a = this->Access (index);
			const OPObjectRef& b = array->Access (index);
			
			index++;
			
			if (a == b) continue;
			
			if (a.IsValid ()) {
				int result = a->CompareWith (b);
				if (result != OP_COMP_THIS_SAME) {
					return result;
				}
			} else {
				return OP_COMP_THIS_BEFORE;
			}
		}
		
		return OP_COMP_THIS_SAME;
	}
	
	return super::CompareWith (object);
}


/*
 *	This looks for the specified object in the array. If an
 *	exact match occurs (same references), it will return TRUE.
 */

Bool
OPArray::LookFor (OPObjectIn model)
{
	OP_METHOD (LookFor, 1);
	OP_ARG (1, OPObjectIn, model);
	
	Count pos = 0;
	return this->Find (model, pos);
}


/*
 *	This looks for the specified object in the array. If an
 *	exact match occurs from the starting position, it will
 *	return its index and TRUE.
 */

Bool
OPArray::Find (OPObjectIn model, CountInOut pos)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, OPObjectIn, model);
	OP_ARG (2, CountInOut, pos);
	
	while (pos < this->num_element) {
		if (this->Access (pos) == model) {
			return TRUE;
		}
		pos++;
	}
	
	return FALSE;
}

/*
 *	Due to some brain-damaged compilers, we cannot use the delete[]
 *	operator to clean up things : we therefore clean up every element
 *	in the array and then free the memory associted with it.
 */

static void
cleanup_and_free_references (OPObjectRef* ref, Count num)
{
	if (ref) {
		for (Count index = 0; index < num; index++) {
			ref[index] = 0;
			JUST_DELETE (&ref[index]);
		}
		OP_FREE (ref);
	}
}


/*
 *	Define the new size of the array. This will simply change
 *	the size of the current array of references.
 *
 *	We should perhaps optimise this code some day.
 *
 *	OVERRIDE if subclass handles the reference access itself.
 */

Bool
OPArray::SetSize (Count size)
{
	OP_METHOD (SetSize, 1);
	OP_ARG (1, Count, size);
	
	if (size == this->num_element) {
		return TRUE;
	}
	
	if (size == 0) {
		
		cleanup_and_free_references (this->references, this->num_element);

		this->references  = 0;
		this->num_element = 0;
		
		return TRUE;
	}
	
	//	Well, I would have loved to use NEW with an array of objects, but then some
	//	compilers add a size in front of it, which will be a problem on deletion !
	//	We should provide a DELETE[] operator equivalent. Some day :-( when ANSI C++
	//	will be supported everywhere.
	
	OPObjectRef* table = 0;
	Count        index = 0;
	
	table = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_data_zone, OPObjectRef, size);
	
	for (index = 0; index < size; index++) {
		NEW_USING (&table[index]) OPObjectRef;
		table[index] = 0;
	}
	
	if (this->references) {
		Count min = (this->num_element < size) ? this->num_element : size;
		for (Count index = 0; index < min; index++) {
			table[index] = this->references[index];
		}
	}
	
	cleanup_and_free_references (this->references, this->num_element);
	
	this->references  = table;
	this->num_element = size;
	
	return TRUE;
}



/*
 *	Define the memory zone to use for the allocations.
 */

Bool
OPArray::SetMemZone (OPMemZone* zone)
{
	OP_METHOD (SetMemZone, 1);
	OP_ARG (1, OPMemZone*, zone);
	
	//	NIY
	
	return FALSE;
}


/*
 *	Insert a range of elements into the array. This will shift
 *	the following elements further down the array by the specified
 *	amount.
 */

Bool
OPArray::Insert (Count pos, Count num)
{
	OP_METHOD (Insert, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, Count, num);
	
	Count end  = pos + num;
	Count old  = this->num_element;
	Count size = old + num;
	
	if (num == 0) {
		return TRUE;
	}
	
	if (end > size) {
		size = end;
	}
	
	this->SetSize (size);
	
	//	Move the references which come after the insertion point.
	
	if (pos < old) {
		Count count = old - pos;
		while (count--) {
			OPObjectRef& o1 = this->Access (pos+count+num);
			OPObjectRef& o2 = this->Access (pos+count);
			o1 = o2;
			o2 = OPObjectNull;
		}
	}
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPArrayIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyInsert (pos, num);
			}
		}
	}
	
	return TRUE;
}


/*
 *	Remove an object from the array. This will shrink the
 *	array and shift the following elements up the array.
 */

Bool
OPArray::Remove (OPObjectIn data, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, data);
	OP_ARG (2, Count, nth);
	
	Count pos = 0;
	
	while (this->Find (data, pos)) {
		if (nth-- == 0) {
			this->Remove (pos);
			return TRUE;
		}
		pos++;
	}
	
	return FALSE;
}

Bool
OPArray::Remove (Count index)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, index);
	
	return this->Remove (index, 1);
}


/*
 *	Remove a range of elements. This will shrink the array
 *	and shift the following elements up the array.
 */

Bool
OPArray::Remove (Count pos, Count num)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, Count, num);
	
	Count old  = this->num_element;
	
	if (num == 0) {
		return TRUE;
	}
	
	if (pos >= old) {
		return FALSE;
	}
	
	if (pos+num > old) {
		num = old - pos;
	}
	
	Count count = old - pos - num;
	Count index = pos;
	
	//	Move the references in order to fill the gap.
	
	while (count--) {
		OPObjectRef& o1 = this->references[index];
		OPObjectRef& o2 = this->references[index+num];
		o1 = o2;
		index++;
	}
	
	this->SetSize (old-num);
	
	//	Notify the iterators that a range of elements has vanished.
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPArrayIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyRemove (pos, num);
			}
		}
	}
	
	return TRUE;
}


/*
 *	General access method : this returns a reference to the
 *	reference (yes, that's right !) so that it can be handled
 *	as a l-value by the caller.
 *
 *	OVERRIDE if subclass handles the reference access itself.
 */

OPObjectRef&
OPArray::Access (Count pos) const
{
	OP_METHOD (Access, 1);
	OP_ARG (1, Count, pos);
	
	static OPObjectRef null;
	
	if (pos < this->num_element) {
		return this->references[pos];
	}
	
	//	If the previous caller modified the "l-value", null might no longer
	//	be null, so make sure this is the case !
	
	null = OPObjectNull;
	return null;
}


/*
 *	The array accessor is based on the "Access" method. No
 *	magic here...
 */

OPObjectRef&
OPArrayRef::operator [] (Count pos) const
{
	const OPArray* array = (OPArray*)(this->object);
	return array->Access (pos);
}
