/*
 *	callframe.cxx
 *
 *	Call frame information.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>

#include <stdlib.h>
#include <stdio.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;

/****************************************************************/

OPaC_CallFrame*				OPaC_CallFrame::head  = 0;
Count						OPaC_CallFrame::total = 0;
OPaC_CallFrame::StaticInfo	OPaC_CallFrame::empty;
OPaC_DebugHook*				OPaC_CallFrame::debug_hook = 0;

#ifdef _DEBUG
static Bool  do_trace    = FALSE;
static Bool  do_profile  = FALSE;
static FILE* trace_file  = 0;
static FILE* prof_file   = 0;
static Count trace_level = 0;
#endif

/****************************************************************/

OPaC_CallFrame::StaticInfo::StaticInfo ()
{
	this->type        = OPaC_CallFrame::NONE;
	this->func_name   = "?";
	this->file_name   = "?";
	this->file_line   = 0;
	this->total_calls = 0;
}

OPaC_CallFrame::StaticInfo::StaticInfo (OPaC_CallFrame::Type t, const LatinChar* func,
										const LatinChar* file, Count line)
	: type (t), func_name (func), file_name (file), file_line (line), total_calls (0)
{
}

OPaC_CallFrame::StaticInfo::~StaticInfo ()
{
#ifdef _DEBUG
	if ( (do_profile)
	  || (this->total_calls) ) {
		
		static tried = FALSE;
		
		if ( (tried == FALSE)
		  && (prof_file == 0) ) {
			prof_file = fopen ("trace-profile.log", "wt");
			tried     = TRUE;
		}
		
		if (prof_file) {
			fprintf (prof_file,
					 "%08ld %s:%ld: %s '%s'\n",
					 this->total_calls,
					 this->file_name,
					 this->file_line,
					 OPaC_CallFrame::ReturnTypeName (this->type),
					 this->func_name);
			fflush (prof_file);
		}
	}
#endif
}

/****************************************************************/

OPaC_CallFrame::OPaC_CallFrame ()
{
	this->original    = 0;
	this->next        = 0;
	this->static_info = & OPaC_CallFrame::empty;
	this->method_this = 0;
	this->call_id     = 0;
}

OPaC_CallFrame::OPaC_CallFrame (OPaC_CallFrame::StaticInfo* info, const void* that)
	: static_info (info), method_this (that)
{
	this->original = this;
	this->next     = OPaC_CallFrame::head;
	this->call_id  = OPaC_CallFrame::total++;
	
	OPaC_CallFrame::head = this;
	
	info->total_calls++;
	
#ifdef _DEBUG
	if (do_trace) {
		
		static tried = FALSE;
		
		if ( (tried == FALSE)
		  && (trace_file == 0) ) {
			trace_file = fopen ("trace-calls.log", "wt");
			tried      = TRUE;
		}
		
		if (trace_file) {
			fprintf (trace_file,
					 "%ld>%s:%ld: %s '%s' - %ld\n",	 
					 trace_level++,
					 this->static_info->file_name,
					 this->static_info->file_line,
					 OPaC_CallFrame::ReturnTypeName (this->static_info->type),
					 this->static_info->func_name,
					 OPaC_CallFrame::total);
		}
	}
#endif
}

OPaC_CallFrame::~OPaC_CallFrame ()
{
	if (this->original == this) {
#ifdef _DEBUG
		if (do_trace) {
			if (trace_file) {
				fprintf (trace_file,
						 "%ld<%s:%ld: %s '%s'\n",
						 --trace_level,
						 this->static_info->file_name,
						 this->static_info->file_line,
						 OPaC_CallFrame::ReturnTypeName (this->static_info->type),
						 this->static_info->func_name);
			}
		}
#endif
		OPaC_CallFrame::head = this->next;
	}
}

/****************************************************************/

const LatinChar*
OPaC_CallFrame::ReturnTypeName (Type type)
{
	switch (type) {
		case METHOD:	return "method";
		case MEMBER:	return "member";
		case OPERATOR:	return "operator";
		case FUNCTION:	return "function";
		case NONE:		break;
	}
	return "?";
}
