/*
 *	carbon.cxx
 *
 *	The carbon copy class can be used to replicated easily any OPaC data type
 *	by serializing it once and then deserializing it as many times as needed.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/carbon.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPCarbonCopy);

/****************************************************************/

#define	MIN_SIZE	1000


/*
 *	Initialise a new instance.
 */

OPCarbonCopy::OPCarbonCopy ()
{
	OP_METHOD (OPCarbonCopy, 0);

	this->block_head     = 0;
	this->block_tail     = 0;
	this->current_block  = 0;
	this->current_offset = 0;
}



/*
 *	Delete an instance. Free the data buffers.
 */

OPCarbonCopy::~OPCarbonCopy ()
{
	OP_METHOD (~OPCarbonCopy, 0);
	this->StartStoring ();
}



/*
 *	Store a carbon copy.
 */

void
OPCarbonCopy::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPCarbonCopy { <block len data> } >");
	
	storage << info;
	storage << OPStorageSync::begin;
	
	DataBlock* block = this->block_head;
	Card32     zero  = 0;
	
	storage << OPStorageSync::repeat;
	
	while (block && block->len) {
		
		storage << OPStorageSync::begin;
		storage << block->len;
		storage->StoreBinData (block->data, block->len);
		storage << OPStorageSync::end;
		
		block = block->next;
	}
	
	storage << OPStorageSync::end_repeat;
	storage << OPStorageSync::end;
}



/*
 *	Restore a carbon copy.
 */

void
OPCarbonCopy::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	DataBlock* block = this->block_head;
	Card32     size  = 0;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::repeat;
	
	while (! storage->CheckForMark ()) {
		
		storage >> OPStorageSync::begin;
		storage >> size;
		this->AddBlock (size);
		block      = this->block_tail;
		block->len = size;
		storage->RestoreBinData (block->data, block->len);
		storage >> OPStorageSync::end;
	}
	
	storage >> OPStorageSync::end_repeat;
	storage >> OPStorageSync::end;
}



/*
 *	Allocate some memory for the block. Pick the memory in
 *	the same data zone as the carbon copy itself.
 */

void
OPCarbonCopy::AddBlock (Card32 size)
{
	OP_METHOD (AddBlock, 1);
	OP_ARG (1, Card32, size);
	
	OPMemZone* zone  = this->ReturnMemZone ();
	DataBlock* block = OP_ALLOC_FROM_ZONE (zone, DataBlock);
	Card8*     data  = OP_ALLOC_ARRAY_FROM_ZONE (zone, Card8, size);

	block->next = 0;
	block->data = data;
	block->len  = 0;
	block->max  = size;
	
	(this->block_tail ? this->block_tail->next : this->block_head) = block;
	(this->block_tail                                            ) = block;
}



/*
 *	Prepare the carbon copy so that it is clean and empty,
 *	ready to store any data...
 */

void
OPCarbonCopy::StartStoring ()
{
	OP_METHOD (StartStoring, 0);
	
	OPMemZone* zone  = this->ReturnMemZone ();
	DataBlock* block = this->block_head;
	DataBlock* next  = 0;

	while (block) {
		next  = block->next;
		OP_FREE_FROM_ZONE (zone, block->data);
		OP_FREE_FROM_ZONE (zone, block);
		block = next;
	}

	this->block_head     = 0;
	this->block_tail     = 0;
	this->current_block  = 0;
	this->current_offset = 0;
	
	this->SetStorageMode (STORE_DEEP);
}



/*
 *	Rewind the carbon copy... This will also clear the
 *	internal data base which accumulate data about every
 *	(re)stored object.
 */

void
OPCarbonCopy::StartRestoring ()
{
	OP_METHOD (StartRestoring, 0);

	this->current_block  = this->block_head;
	this->current_offset = 0;
	
	this->SetStorageMode (STORE_DEEP);
}



/*
 *	Store data into the carbon copy. This will grow the
 *	current buffer if needed.
 */

void
OPCarbonCopy::StoreData (const void* data, Card32 size)
{
	OP_METHOD (StoreData, 2);
	OP_ARG (1, const void*, data);
	OP_ARG (2, Card32, size);
	
	if ( (this->block_tail == 0)
	  || (this->block_tail->max == this->block_tail->len) ) {
		this->AddBlock ((size < MIN_SIZE) ? MIN_SIZE : size);
	}
	
	Card32       free = this->block_tail->max - this->block_tail->len;
	const Card8* src  = (const Card8*)(data);
	Card8*       dst  = this->block_tail->data + this->block_tail->len;

	if (free < size) {
		
		this->StoreData (src, free);
		this->StoreData (src+free, size-free);
		
	} else {
		
		this->block_tail->len += size;
		
		//	Copy the data byte by byte. Could be optimized somewhat...
		
		while (size--) {
			*dst++ = *src++;
		}
	}
}



/*
 *	Restore data from the carbon copy. This is only possible
 *	if the user called StartRestoring before.
 */

void
OPCarbonCopy::RestoreData (void* data, Card32 size)
{
	OP_METHOD (StoreData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);
	
	if (size == 0) {
		return;
	}
	
	Card32       len = 0;
	Card8*       dst = (Card8*)(data);
	const Card8* src = 0;
	
	while ( (this->current_block)
	     && (size != 0) ) {
		
		len = this->current_block->len  - this->current_offset;
		src = this->current_block->data + this->current_offset;
		
		if (len >= size) {
			
			//	This block contains enough data in order to satisfy the current
			//	request.
			
			this->current_offset += size;

			//	Copy the data byte by byte. Could be optimized somewhat...

			while (size--) {
				*dst++ = *src++;
			}

			return;
		}
		
		this->RestoreData (dst, len);
		
		dst  += len;
		size -= len;
		
		this->current_block  = this->current_block->next;
		this->current_offset = 0;
	}
}

