/*
 *	container.cxx
 *
 *	Base OPaC Containter class. This is just an abstract class implementation.
 *	Hash tables and sorted lists inherit its properties.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/container.h>
#include <OPaC/array.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_ABSTRACT_CLASS (OPContainer);

/****************************************************************/



/*
 *	The container is an abstract class. It does just count
 *	how many objects have been recorded in it.
 */

OPContainer::OPContainer ()
{
	OP_METHOD (OPContainer, 0);
	
	this->num_element = 0;
}



/*
 *	Destroying a non-empty container will be considered as
 *	fatal. The subclass destructor will have to clean up the
 *	mess before returning !
 */

OPContainer::~OPContainer ()
{
	OP_METHOD (~OPContainer, 0);
	OP_ASSERT (this->num_element == 0);
}



/*
 *	Save the number of elements currently stored in the
 *	container. This method could also save the elements,
 *	but I think it's better if every subclass handles
 *	each case separately.
 */

void
OPContainer::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPContainer num_element>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->num_element;
	storage << OPStorageSync::end;
}



/*
 *	Restore the number of elements which were stored
 *	in the container. This can be used by subclasses
 *	to determine how many elements they will have to
 *	read.
 */

void
OPContainer::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->num_element;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the container for its destruction.
 */

void
OPContainer::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	super::Delete ();
}



/*
 *	Count the elements found in the container.
 */

Count
OPContainer::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	return this->num_element;
}


/*
 *	Specify a generic extraction. Well, a container does in
 *	fact not need any specification for extraction. Simply
 *	never call this entry.
 */

Bool
OPContainer::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	if (x->TotalCount () == 0) {
		return TRUE;
	}

	return FALSE;
}

