/*
 *	datarep.cxx
 *
 *	This file defines a class which reports data modifications to another
 *	object when it gets a change notification.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/convert.h>
#include <OPaC/datarep.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPDataReport);

/****************************************************************/

OP_TARGET_OBSERVER (OPDataReport, Changed);

/****************************************************************/

/*
 *	Initialise an empty data reporter.
 */

OPDataReport::OPDataReport ()
{
	OP_METHOD (OPDataReport, 0);
	
	this->query_name   = OPAtomNull;
	this->query_path   = OPStringNull;
	this->query_format = OPStringNull;
	this->def_object   = 0;
	this->def_name     = OPAtomNull;
	this->def_path     = OPStringNull;
	this->action_name  = OPAtomNull;
	this->action_path  = OPStringNull;
	this->is_visible   = TRUE;
}



/*
 *	Clean-up the data reporter.
 */

OPDataReport::~OPDataReport ()
{
	OP_METHOD (~OPDataReport, 0);
}



/*
 *	Store the data reporter.
 */

void
OPDataReport::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPDataReport query_name query_path query_format"
							   " def_object def_name def_path"
							   " action_name action_path"
							   " is_visible>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->query_name << this->query_path << this->query_format;
	storage << this->def_object << this->def_name << this->def_path;
	storage << this->action_name << this->action_path;
	storage << this->is_visible;
	storage << OPStorageSync::end;
}



/*
 *	Restore the data reporter.
 */

void
OPDataReport::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->query_name >> this->query_path >> this->query_format;
	storage >> this->def_object >> this->def_name >> this->def_path;
	storage >> this->action_name >> this->action_path;
	storage >> this->is_visible;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the data reporter for deletion.
 */

void
OPDataReport::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->def_object.IsValid ()) {
		this->def_object->UnregisterAsTargetOf (this);
	}
	
	this->def_object = 0;
	
	super::Delete ();
}



/*
 *	Initialise a very simple data reporter. Take a value
 *	from the source and set it to the destination. Perform
 *	an action if needed.
 */

void
OPDataReport::Initialise (OPObjectIn source, OPAtomIn event, OPAtomIn query,
						  OPObjectIn destination, OPAtomIn target,
						  OPAtomIn action,
						  Bool is_visible)
{
	OP_METHOD (Initialise, 7);
	OP_ARG (1, OPObjectIn, source);
	OP_ARG (2, OPAtomIn, event);
	OP_ARG (3, OPAtomIn, query);
	OP_ARG (4, OPObjectIn, destination);
	OP_ARG (5, OPAtomIn, target);
	OP_ARG (6, OPAtomIn, action);
	OP_ARG (7, Bool, is_visible);
	
	static OPAtomRef changed = "Changed";
	static OPAtomRef vol_obs = "OPaC.volatile.observer";
	
	if (source->GetAttribute (vol_obs)) {
		this->SetAttribute (vol_obs);
	}
	
	this->DefineQuery (query);
	this->DefineTarget (destination, target);
	this->DefineAction (action);
	this->SetVisible (is_visible);
	
	OPObject::AttachObserver (source, this, event, changed, 0);
}


/*
 *	Define the reporter's visibility. If set, this means that
 *	the data link viewer/browser will show this reporter...
 */

void
OPDataReport::SetVisible (Bool yes)
{
	OP_METHOD (SetVisible, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_visible = yes;
}

Bool
OPDataReport::IsVisible () const
{
	OP_METHOD (IsVisible, 0);
	
	return this->is_visible;
}




void
OPDataReport::DefineQuery (OPAtomIn name, OPStringIn format)
{
	OP_METHOD (DefineQuery, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPStringIn, format);
	
	this->query_name   = name;
	this->query_path   = OPStringNull;
	this->query_format = format;
}


void
OPDataReport::DefineQueryPath (OPStringIn path, OPStringIn format)
{
	OP_METHOD (DefineQueryPath, 2);
	OP_ARG (1, OPStringIn, path);
	OP_ARG (2, OPStringIn, format);
	
	this->query_name   = OPAtomNull;
	this->query_path   = path;
	this->query_format = format;
}


void
OPDataReport::DefineTarget (OPObjectIn obj, OPAtomIn name)
{
	OP_METHOD (DefineTarget, 2);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPAtomIn, name);
	
	static OPAtomRef vol_obs = "OPaC.volatile.observer";
	
	if (this->def_object.IsValid ()) {
		this->def_object->UnregisterAsTargetOf (this);
	}
	
	if (obj->GetAttribute (vol_obs)) {
		if (!this->GetAttribute (vol_obs)) {
			this->SetAttribute (vol_obs);
		}
	}
	
	this->def_object = obj;
	this->def_name   = name;
	this->def_path   = OPStringNull;
	
	if (this->def_object.IsValid ()) {
		this->def_object->RegisterAsTargetOf (this);
	}
}


void
OPDataReport::DefineTargetPath (OPObjectIn obj, OPStringIn path)
{
	OP_METHOD (DefineTargetPath, 2);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPStringIn, path);
	
	this->def_object = obj;
	this->def_name   = OPAtomNull;
	this->def_path   = path;
	
	static OPAtomRef vol_obs = "OPaC.volatile.observer";
	
	if (obj->GetAttribute (vol_obs)) {
		if (!this->GetAttribute (vol_obs)) {
			this->SetAttribute (vol_obs);
		}
	}
}


void
OPDataReport::DefineAction (OPAtomIn action)
{
	OP_METHOD (DefineAction, 1);
	OP_ARG (1, OPAtomIn, action);
	
	this->action_name = action;
	this->action_path = OPStringNull;
}


void
OPDataReport::DefineActionPath (OPStringIn path)
{
	OP_METHOD (DefineActionPath, 1);
	OP_ARG (1, OPStringIn, path);
	
	this->action_name = OPAtomNull;
	this->action_path = path;
}


void
OPDataReport::Changed (OPObjectIn obj, Card32 id)
{
	OP_METHOD (Changed, 2);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, Card32, id);
	
	if ( (obj.IsValid ())
	  && (this->def_object.IsValid ())
	  && (this->def_object->GetCurrentReportID () != OPObject::global_report_id) ) {
		
		OPValueRef value;
		
		if (this->query_name.IsValid ()) {
			if (obj->QueryData (this->query_name, value) == FALSE) {
				return;
			}
		} else if (this->query_path.IsValid ()) {
			if (OPObject::QueryData (obj, this->query_path, value) == FALSE) {
				return;
			}
		}
		
		if (this->def_name.IsValid ()) {
			value->SetAutoFormat (this->query_format);
			this->def_object->DefineData (this->def_name, value);
		} else if (this->def_path.IsValid ()){
			value->SetAutoFormat (this->query_format);
			OPObject::DefineData (this->def_object, this->def_path, value);
		}
		
		if (this->action_name.IsValid ()) {
			this->def_object->Action (this->action_name);
		} else if (this->action_path.IsValid ()) {
			OPObject::Action (this->def_object, this->action_path);
		}
	}
}

