/*
 *	dbentity.cxx
 *
 *	This file implements the OPaC Data Base Entity class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/database.h>
#include <OPaC/except.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPEntity);

/****************************************************************/

/* 
 *	Create an empty entity. This will need initialization in
 *	order to be usable.
 */

OPEntity::OPEntity ()
{
	OP_METHOD (OPEntity, 0);
	this->data_num   = 0;
}



/*
 *	Destroy the entity. This does not free the data which have
 *	been stored in the miscellanous attributes.
 */

OPEntity::~OPEntity ()
{
	OP_METHOD (~OPEntity, 0);
	
	if (this->model.IsValid ()) {
		this->model->RemoveUser (this);
	}
}



/*
 *	Save the entity and its attributes.
 */

void
OPEntity::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPEntity model data_num data_array>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->model;
	storage << this->data_num;
	storage << this->data_array;
	storage << OPStorageSync::end;
}



/*
 *	Restore the entity and its attributes (this may also
 *	have to restore automatically the model).
 */

void
OPEntity::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->model;
	storage >> this->data_num;
	storage >> this->data_array;
	storage >> OPStorageSync::end;
	
	if (this->model.IsValid ()) {
		this->model->AddUser (this);
	}
}



/*
 *	Initialise an empty entity : every attribute is set to a
 *	NULL pointer, but the memory has been allocated for the
 *	attributes, depending on the specified model.
 */

void
OPEntity::Initialise (OPEntityModelIn model)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPEntityModelIn, model);
	OP_ASSERT (this->model.IsInvalid ());
	OP_ASSERT (model.IsValid ());
	
	this->model = model;
	this->model->AddUser (this);
	
	Count num = model->CountAttributes ();
	
	this->data_array = OP_NEW (OPArray);
	this->data_array->SetMemZone (this->ReturnMemZone ());
	this->data_array->SetSize (num);
	this->data_num = num;
}



/*
 *	Return the number of attributes available in the
 *	entity.
 */

Count
OPEntity::CountAttributes () const
{
	OP_METHOD (CountAttributes, 0);
	return this->data_num;
}



/*
 *	Remove the attributes stored in the entity.
 */

void
OPEntity::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	if (this->data_array->TotalCount () == 0) {
		return;
	}
	
	Count index = 0;
	
	while (index < this->data_num) {
		this->data_array[index] = 0;
		index++;
	}
}



/*
 *	Return the attribute's data, specified by an index
 *	(range 1 to n). Return FALSE if no data has been
 *	defined yet or if out of range.
 */

Bool
OPEntity::GetAttribute (Count index, OPObjectOut data) const
{
	OP_METHOD (GetAttribute, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, data);
	
	if ( (index == 0)
	  || (index > this->data_num)
	  || (this->data_array->TotalCount () == 0)
	  || (this->model.IsInvalid ()) ) {
		
		data = 0;
		return FALSE;
	}
	
	data = this->data_array[index-1];
	return TRUE;
}



/*
 *	Return the attribute's data, specified by its name.
 *	Return FALSE if no data has been defined yet or if
 *	the specified attribute does not exist in the model.
 */

Bool
OPEntity::GetAttribute (OPAtomIn name, OPObjectOut data) const
{
	OP_METHOD (GetAttribute, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, data);
	
	Count index = 0;
	
	if ( (this->data_array->TotalCount () == 0)
	  || (this->model.IsInvalid ())
	  || ((index = this->model->GetAttributeIndex (name)) == 0) ) {
		
		data = 0;
		return FALSE;
	}
	
	data = this->data_array[index-1];
	return TRUE;
}



/*
 *	Define the attribute's data, specified by an index
 *	(range 1 to n). Return TRUE on success.
 */

Bool
OPEntity::SetAttribute (Count index, OPObjectIn data)
{
	OP_METHOD (SetAttribute, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectIn, data);
	
	if ( (index == 0)
	  || (index > this->data_num)
	  || (this->data_array->TotalCount () == 0)
	  || (this->model.IsInvalid ()) ) {
		
		return FALSE;
	}
	
	this->data_array[index-1] = data;
	return TRUE;
}



/*
 *	Define the attribute's data, specified by the attribute's
 *	name. Return TRUE on success.
 */

Bool
OPEntity::SetAttribute (OPAtomIn name, OPObjectIn data)
{
	OP_METHOD (SetAttribute, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectIn, data);
	
	Count index = 0;
	
	if ( (this->data_array->TotalCount () == 0)
	  || (this->model.IsInvalid ())
	  || ((index = this->model->GetAttributeIndex (name)) == 0) ) {
		
		return FALSE;
	}
	
	this->data_array[index-1] = data;
	return TRUE;
}



/*
 *	Return a reference to the entity's model. This may
 *	be useful if you have to access an unknown entity's
 *	attributes.
 */

const OPEntityModelRef&
OPEntity::GetEntityModel () const
{
	OP_METHOD (GetEntityModel, 0);
	
	return this->model;
}



