/*
 *	dbmodel.cxx
 *
 *	This file implements the OPaC Data Base Entity Model class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/database.h>
#include <OPaC/except.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPEntityModel);

/****************************************************************/


/*
 *	Create an empty model. This will have to be initialised
 *	in ordered to be used by an entity.
 */

OPEntityModel::OPEntityModel ()
{
	OP_METHOD (OPEntityModel, 0);
	this->user_num   = 0;
	this->name_num   = 0;
}



/*
 *	Destroy a model. If there are still entities using it,
 *	this will lead to a fatal error.
 */

OPEntityModel::~OPEntityModel ()
{
	OP_METHOD (~OPEntityModel, 0);
	OP_ASSERT (this->user_num == 0);
}



/*
 *	Save the entity model. This does NOT remember how
 *	many entities had subscribed to this model...
 */

void
OPEntityModel::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPEntityModel containing_db name name_num name_array>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->containing_db;
	storage << this->name;
	storage << this->name_num;
	storage << this->name_array;
	storage << OPStorageSync::end;
}



/*
 *	Restore the entity model. This will NOT restore how
 *	many entities had subscribed to this model. The
 *	restored entities will have to use "AddUser" again !
 */

void
OPEntityModel::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->containing_db;
	storage >> this->name;
	storage >> this->name_num;
	storage >> this->name_array;
	storage >> OPStorageSync::end;
}



/*
 *	Initialise an entity model : give it its name and
 *	specify how many attributes will be used.
 */

Bool
OPEntityModel::Initialise (OPDataBaseIn db, OPStringIn name, Count max)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPDataBaseIn, db);
	OP_ARG (2, OPStringIn, name);
	OP_ARG (3, Count, max);
	
	if ( this->user_num
	  || this->name_num
	  || (max == 0) ) {
		
		return FALSE;
	}
	
	this->containing_db = db;
	this->name_array = OP_NEW (OPArray);
	this->name_array->SetMemZone (this->ReturnMemZone ());
	this->name_array->SetSize (max);
	this->name_num = max;
	this->name     = name;
	
	return TRUE;
}



/*
 *	While initializing, adds a new attribute to the
 *	entity model. Return FALSE if the attribute could
 *	not be added to the model.
 */

Bool
OPEntityModel::AddAttribute (OPAtomIn name)
{
	OP_METHOD (AddAttribute, 1);
	OP_ARG (1, OPAtomIn, name);
	
	Count i;
	
	for (i = 0; i < this->name_num; i++) {
		
		//	We may not use twice the same attribute name in the model; this
		//	would lead to conflicts...
		
		if (this->name_array[i].IsValid ()) {
			if (this->name_array[i] == name) {
				break;
			}
		} else {
			this->name_array[i] = name;
			return TRUE;
		}
	}
	
	return FALSE;
}



/*
 *	Return the number of attributes defined in the model.
 *	This will be 0 if the model hasn't been initialised
 *	yet.
 */

Count
OPEntityModel::CountAttributes () const
{
	OP_METHOD (CountAttributes, 0);
	return this->name_num;
}



/*
 *	Return the index corresponding to the named attribute.
 *	If no matching attribute can be found, return 0.
 */

Count
OPEntityModel::GetAttributeIndex (OPAtomIn name) const
{
	OP_METHOD (GetAttributeIndex, 1);
	OP_ARG (1, OPAtomIn, name);
	
	Count i;
	
	for (i = 0; i < this->name_num; i++) {
		if ( (this->name_array[i].IsValid ())
		  && (this->name_array[i] == name) ) {
			return i+1;
		}
	}
	
	return 0;
}



/*
 *	Find the index corresponding to the named attribute.
 *	If no matching attribute can be found, return FALSE.
 */

Bool
OPEntityModel::GetAttributeIndex (OPAtomIn name, CountOut index) const
{
	OP_METHOD (GetAttributeIndex, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, CountOut, index);
	
	index = this->GetAttributeIndex (name);
	return (index == 0) ? FALSE : TRUE;
}



/*
 *	Return the nth attribute's name. Return FALSE if the
 *	index was out of bounds (accept an index in the range
 *	1..CountAttributes).
 */

Bool
OPEntityModel::GetAttributeName (Count index, OPAtomOut name) const
{
	OP_METHOD (GetAttributeName, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	
	if (index > this->name_num) {
		name = OPAtomNull;
		return FALSE;
	}
	
	name = this->name_array[index-1];
	return TRUE;
}



/*
 *	Add an entity to this model. This will just increment
 *	a counter, for book-keeping; this is only possible if
 *	the model has been properly initialised before.
 */

void
OPEntityModel::AddUser (OPEntity* entity)
{
	OP_METHOD (AddUser, 1);
	OP_ARG (1, OPEntity*, entity);
	OP_ASSERT (this->name_num);
	OP_ASSERT (this->name_array->TotalCount () > 0);
	this->user_num++;
}



/*
 *	Remove an entity from this model. This will yield a
 *	fatal error if un underflow occurs.
 */

void
OPEntityModel::RemoveUser (OPEntity* entity)
{
	OP_METHOD (RemoveUser, 1);
	OP_ARG (1, OPEntity*, entity);
	OP_ASSERT (this->user_num);
	this->user_num--;
}

