/*
 *	filewr.cxx
 *
 *	The FileWrite class is a specialized stream used to write data to a file.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/string.h>
#include <OPaC/stream.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPFileWrite);

/****************************************************************/

/*
 *	The File Write Stream provides an interface for writing
 *	data sequentially to a file.
 *
 *	The file can either be created from scratch or just be
 *	expanded from its previous state.
 */

OPFileWrite::OPFileWrite ()
{
	OP_METHOD (OPFileWrite, 0);
	
	this->is_append_mode = FALSE;
}



/*
 *	Close the file if still open at destruction time.
 */

OPFileWrite::~OPFileWrite ()
{
	OP_METHOD (~OPFileWrite, 0);
	
	this->file.Close ();
}



/*
 *	Store additional information.
 */
	
void
OPFileWrite::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPFileWrite is_append_mode>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->is_append_mode;
	storage << OPStorageSync::end;
}



/*
 *	Restore additional information.
 */

void
OPFileWrite::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->is_append_mode;
	storage >> OPStorageSync::end;
	
	if (this->is_append_mode) {
		this->index = 0;			//	prevent moving past the end on "Awake"
	}
}



/*
 *	If append is FALSE :
 *
 *	- Open a file from scratch. This will erase its previous
 *	  contents.
 *
 *	If append is TRUE :
 *
 *	- Open a file to append data to it. This will expand the
 *	  file from its current end.
 *
 *	If another file was under use before, close it. The file
 *	name has to be a physical, OS dependent file name !
 *
 *	Throws :
 *
 *	- OPExcStream (CANNOT_OPEN) if an error occurred while
 *	  opening the file.
 */

void
OPFileWrite::Open (OPStringIn name, Bool append)
{
	OP_METHOD (Open, 2);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, Bool, append);
	
	this->SetAppendMode (append);
	this->SetName (name);
	this->Open ();
}
	
void
OPFileWrite::Open ()
{
	OP_METHOD (Open, 0);
	
	int mode = (this->is_append_mode)
			 ? OPaC_File::OPEN_WRITE | OPaC_File::OPEN_APPEND
			 : OPaC_File::OPEN_WRITE | OPaC_File::OPEN_CREATE | OPaC_File::OPEN_TRUNC;
	
	this->file.Close ();
	this->is_open = this->file.Open (this->name, (OPaC_File::Mode)(mode));
	this->index   = 0;
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::CANNOT_OPEN);
	}
	
	if (this->is_append_mode) {
		this->index = this->file.TellPos ();
	}
}


/*
 *	Close the current file. This will flush the data still
 *	remaining in the internal buffers.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  closing the file.
 */

void
OPFileWrite::Close ()
{
	OP_METHOD (Close, 0);
	
	this->file.Close ();
	this->is_open = FALSE;
	this->index   = 0;
}



/*
 *	Write data to the file.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  writing to the file.
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

Count
OPFileWrite::Write (const void* buffer, Count size)
{
	OP_METHOD (Write, 2);
	OP_ARG (1, const void*, buffer);
	OP_ARG (2, Count, size);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_WRITE);
	}
	
	if (this->file.Write (buffer, size) != size) {
		throw OPExcStream (OPExcStream::WRITE_ERROR);
	}

	this->index += size;
	
	return size;
}



/*
 *	Synchronize the file with the stream : this flushes the
 *	internal buffers.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  synchronizing the file.
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

Bool
OPFileWrite::Sync ()
{
	OP_METHOD (Sync, 0);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_WRITE);
	}
	
	if (this->file.Flush () == FALSE) {
		throw OPExcStream (OPExcStream::WRITE_ERROR);
	}
	
	return TRUE;
}



/*
 *	Truncate file to the current position.
 *
 *	Throws :
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

void
OPFileWrite::Trunc ()
{
	OP_METHOD (Trunc, 0);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_WRITE);
	}
	
	this->file.Truncate ();
}

Int64
OPFileWrite::Skip (Int64 skip)
{
	OP_METHOD (Skip, 1);
	OP_ARG (1, Int64, skip);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_SKIP);
	}
	
	if (this->file.SeekRel (skip) == FALSE) {
		throw OPExcStream (OPExcStream::WRITE_ERROR);
	}
	
	this->index += skip;
	
	return skip;
}

/****************************************************************/

Bool
OPFileWrite::SetAppendMode (Bool on)
{
	OP_METHOD (SetAppendMode, 1);
	OP_ARG (1, Bool, on);
	
	this->is_append_mode = on;
	return TRUE;
}

Bool
OPFileWrite::ReturnAppendMode () const
{
	OP_METHOD (ReturnAppendMode, 0);
	
	return this->is_append_mode;
}

/****************************************************************/
