/*
 *	menu.cxx
 *
 *	The Menu class can be used to build independent menus, pop up
 *	menus and pull down menus.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/menu.h>
#include <OPaC/new.h>
#include <OPaC/cellgrid.h>
#include <OPaC/control.h>
#include <OPaC/event.h>
#include <OPaC/notmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPMenu);

/****************************************************************/

OP_NOTIFY_EVENT (OPMenu, ButtonReleased);

OP_TARGET_OBSERVER (OPMenu, PopupAsPopup);
OP_TARGET_OBSERVER (OPMenu, PopupAsSubmenu);
OP_TARGET_OBSERVER (OPMenu, PopupAsPulldown);
OP_TARGET_OBSERVER (OPMenu, HandleEnterNotification);
OP_TARGET_OBSERVER (OPMenu, HandleLeaveNotification);

/****************************************************************/


/*
 *	Construct the menu.
 */

OPMenu::OPMenu ()
{
	OP_METHOD (OPMenu, 0);
	
	this->grid         = 0;
	this->last_choice  = 0;
	this->is_automatic = FALSE;
	this->end_buttons  = OP_EV_BUTTON_LEFT | OP_EV_BUTTON_RIGHT
					   | OP_EV_BUTTON_MIDDLE | OP_EV_BUTTON_EXTRA;
}



/*
 *	Destroy the menu.
 */

OPMenu::~OPMenu ()
{
	OP_METHOD (~OPMenu, 0);
}



/*
 *	Save the menu to the storage.
 */

void
OPMenu::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPMenu grid last_choice end_buttons>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->grid;
	storage << this->last_choice;
	storage << this->end_buttons;
	storage << OPStorageSync::end;
}



/*
 *	Restore the menu from the storage.
 */

void
OPMenu::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->grid;
	storage >> this->last_choice;
	storage >> this->end_buttons;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the menu for destruction. Give up both
 *	arrow buttons.
 */

void
OPMenu::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->grid.IsValid ()) {
		this->grid->Delete ();
		this->grid = 0;
	}
	
	super::Delete ();
}



/*
 *	Awake the menu.
 */

void
OPMenu::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
}

void
OPMenu::AwakeFromApp (OPAppIn app)
{
	OP_METHOD (AwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);
	
	super::AwakeFromApp (app);
}

void
OPMenu::AwakeFromObject (OPObjectIn object)
{
	OP_METHOD (AwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	super::AwakeFromObject (object);
}



/*
 *	Define the cell grid to be used with the menu. This is
 *	only possible once...
 */

void
OPMenu::SetCellGrid (OPCellGridIn grid)
{
	OP_METHOD (SetCellGrid, 1);
	OP_ARG (1, OPCellGridIn, grid);
	OP_ASSERT (this->grid.IsValid () == FALSE);
	OP_ASSERT (grid.IsValid ());
	
	this->grid = grid;
	this->grid->SetParent (OPMenuRef (this));
	this->grid->UpdateCellGeometry ();
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	this->grid->GetSize (dx, dy);
	this->SetSize (dx, dy);
	
	Count width  = 0;
	Count height = 0;
	Count pos    = 0;
	
	this->grid->GetArraySize (width, height);
	
	for (Count h = 0; h < width; h++) {
		for (Count v = 0; v < height; v++) {
			OPControlRef cell;
			if ( (this->grid->GetCell (h, v, cell))
			  && (cell.IsValid ())) {
				cell->AttachObserver (this, "EnterNotify", "HandleEnterNotification", pos);
				cell->AttachObserver (this, "LeaveNotify", "HandleLeaveNotification", pos);
			}
			pos++;
		}
	}
}

const OPCellGridRef&
OPMenu::ReturnCellGrid () const
{
	OP_METHOD (ReturnCellGrid, 0);
	return this->grid;
}


/*
 *	Make the menu visible at the specified position. The
 *	menu will be centered about the [x] coordinate.
 */

void
OPMenu::ShowMenu (OPCoord x, OPCoord y)
{
	OP_METHOD (ShowMenu, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ASSERT (this->grid.IsValid ());
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	this->grid->GetSize (dx, dy);
	
	//	Center horizontally the menu. If not possible, simply ask to put at the
	//	left side...
	
	x = ((x * 2) > dx) ? x - dx / 2 : 0;
	
	this->SetStyle (OP_WIN_NAKED);
	this->SetOrigin (x, y);
	this->SetSize (dx, dy);
	this->SetTitle ("-- MENU --");
	this->Open ();
	this->Display ();
	this->ShowNormal ();
}



/*
 *	Hide the menu. This will simply close the associated
 *	window, and thus free some memory.
 */

void
OPMenu::HideMenu ()
{
	OP_METHOD (HideMenu, 0);
	
	this->is_automatic = FALSE;
	
	this->Hide ();
	this->Close ();
}


/*
 *	Look for a button release event... This will trigger a
 *	special action.
 */

void
OPMenu::DispatchEvent (OPEventIn event)
{
	OP_METHOD (DispatchEvent, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::DispatchEvent (event);
	
	if ( (this->end_buttons)
	  && (event->ReturnType () == OP_EVENT_RELEASE)
	  && (event->ReturnChange () & this->end_buttons) ) {
		this->QuickReport (notify_ButtonReleased, OP_REPORT_NEW);
	}
}



void
OPMenu::SetEndButtons (Card32 buttons)
{
	OP_METHOD (SetEndButtons, 1);
	OP_ARG (1, Card32, buttons);
	
	this->end_buttons = buttons;
}

Card32
OPMenu::ReturnEndButtons () const
{
	OP_METHOD (ReturnEndButtons, 0);
	return this->end_buttons;
}


void
OPMenu::PopupAsPopup (OPObjectIn button, Card32 id)
{
	OP_METHOD (PopupAsPopup, 2);
	OP_ARG (1, OPObjectIn, button);
	OP_ARG (2, Card32, id);
	
	OPViewRef   view   = button;
	OPWindowRef window;
	view->GetWindow (window);
	
	OPCoord ox, oy;
	OPCoord dx, dy;
	OPCoord wx = 0;
	OPCoord wy = 0;
	
	view->GetOrigin (ox, oy);
	view->GetSize (dx, dy);
	if (view != window) window->GetOrigin (wx, wy);
	
	this->is_automatic = TRUE;
	this->ShowMenu (ox + wx + dx / 2, oy + wy);
}


void
OPMenu::PopupAsSubmenu (OPObjectIn button, Card32 id)
{
	OP_METHOD (PopupAsSubmenu, 2);
	OP_ARG (1, OPObjectIn, button);
	OP_ARG (2, Card32, id);
	
	OPViewRef   view   = button;
	OPWindowRef window;
	view->GetWindow (window);
	
	OPCoord ox, oy;
	OPCoord dx, dy;
	OPCoord sx, sy;
	OPCoord wx = 0;
	OPCoord wy = 0;
	
	view->GetOrigin (ox, oy);
	view->GetSize (dx, dy);
	this->GetSize (sx, sy);
	if (view != window) window->GetOrigin (wx, wy);
	
	this->is_automatic = TRUE;
	this->ShowMenu (ox + dx + wx + sx / 2, oy + wy);
}


void
OPMenu::PopupAsPulldown (OPObjectIn button, Card32 id)
{
	OP_METHOD (PopupAsPulldown, 2);
	OP_ARG (1, OPObjectIn, button);
	OP_ARG (2, Card32, id);
	
	OPViewRef   view   = button;
	OPWindowRef window;
	view->GetWindow (window);
	
	OPCoord ox, oy;
	OPCoord dx, dy;
	OPCoord wx = 0;
	OPCoord wy = 0;
	
	view->GetOrigin (ox, oy);
	view->GetSize (dx, dy);
	if (view != window) window->GetOrigin (wx, wy);
	
	this->is_automatic = TRUE;
	this->ShowMenu (ox + wx + dx / 2, oy + wy + dy);
}

/*
 *	When a cell of the menu gets the mouse focus (it is being
 *	entered), we might need to open an associated submenu or
 *	pull-down menu.
 */

void
OPMenu::HandleEnterNotification (OPObjectIn cell, Card32 id)
{
	OP_METHOD (HandleEnterNotification, 2);
	OP_ARG (1, OPObjectIn, cell);
	OP_ARG (2, Card32, id);
	
	OPMenuRef menu = 0;
	
	if (cell->GetAttribute ("Submenu", menu)) {
		menu->PopupAsSubmenu (cell, id);
	} else if (cell->GetAttribute ("Pulldown", menu)) {
		menu->PopupAsPulldown (cell, id);
	}
}

/*
 *	When the mouse leaves a cell, it can either move to another
 *	cell or enter the associated submenu or pull-down menu...
 */

void
OPMenu::HandleLeaveNotification (OPObjectIn cell, Card32 id)
{
	OP_METHOD (HandleLeaveNotification, 2);
	OP_ARG (1, OPObjectIn, cell);
	OP_ARG (2, Card32, id);
	
	OPMenuRef menu = 0;
	OPViewRef view = OPEvent::CurrentEvent ()->ReturnWindow ();
	
	if (cell->GetAttribute ("Submenu", menu)) {
		if (menu == view) {
			return;
		}
	} else if (cell->GetAttribute ("Pulldown", menu)) {
		if (menu == view) {
			return;
		}
	}
	
	if (menu.IsValid ()) {
		menu->HideMenu ();
	}
}

void
OPMenu::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->is_automatic)
	  && (event->ReturnWindow () != this) ) {
		this->HideMenu ();
	}
	
	super::LeaveNotify (event);
}
