/*
 *	message.cxx
 *
 *	Message class implementation. This is used when messages are being
 *	sent to dynamically defined methods and/or targets. This used by
 *	the special messaging construct used by OPaC :
 *
 *		"x <- (msg-method) << arg << arg"
 *	or  "x <- msgMethod << arg << arg" (the method must be set before)
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/message.h>
#include <OPaC/opacirt.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPMessage);

/****************************************************************/


/*
 *	Create an empty message. The destination will be defined by the Send,
 *	SetMethod and/or SetTarget methods.
 */

OPMessage::OPMessage ()
{
	OP_METHOD (OPMessage, 0);

	this->data_ptr = this->data_heap;
	this->data_len = sizeof (this->data_heap);
	
	this->SetStorageMode (STORE_SHALLOW);
	
	this->target_ok = FALSE;
	this->method_ok = FALSE;
}

OPMessage::~OPMessage ()
{
	OP_METHOD (~OPMessage, 0);
}

void
OPMessage::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store a message");
}

void
OPMessage::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore a message");
}


/*
 *	Remember about the method which will have to be invoked
 *	when sending the data.
 */

void
OPMessage::SetMethod (OPAtomIn name)
{
	OP_METHOD (SetMethod, 1);
	OP_ARG (1, OPAtomIn, name);
	this->method    = name;
	this->target    = (OPTarget*)(0);
	this->method_ok = TRUE;
	this->target_ok = FALSE;
}

void
OPMessage::SetMethod (OPStringIn name)
{
	OP_METHOD (SetMethod, 1);
	OP_ARG (1, OPStringIn, name);
	OPAtomRef atom = name;
	this->SetMethod (atom);
}

void
OPMessage::SetMethod (const LatinChar* name)
{
	OP_METHOD (SetMethod, 1);
	OP_ARG (1, const LatinChar*, name);
	OPAtomRef atom = name;
	this->SetMethod (atom);
}


/*
 *	Remember about the target which will have to be invoked
 *	when sending the data.
 */

void
OPMessage::SetTarget (OPTargetIn target)
{
	OP_METHOD (SetTarget, 1);
	OP_ARG (1, OPTargetIn, target);

	this->method    = OPAtomNull;
	this->target    = target;
	this->method_ok = FALSE;
	this->target_ok = target.IsValid () && target->IsValid ();
}


/*
 *	Send the message, which has been defined previously, to
 *	the specified object. Return TRUE if the message could be
 *	sent properly.
 */

Bool
OPMessage::Send (OPObjectIn dest)
{
	OP_METHOD (Send, 1);
	OP_ARG (1, OPObjectIn, dest);
	
	//	Make sure that the storage behind the scenes does not remember anything
	//	about the past stores...
	
	this->SetStorageMode (STORE_SHALLOW);
	
	if (this->method_ok) {
		
		//	The method has been defined using an atomic name. Lookup that
		//	name and call the method.

		const OPaC_Target* method = dest->ReturnRuntime ()->FindMethod (this->method);
		OPObject*          object = dest.object;
		
		if (method) {
			OPMessageRef self = this;
			Card8* ptr = this->data_ptr;
			Card32 len = this->data_len;
			this->data_ptr = this->data_heap;
			this->data_len = sizeof (this->data_heap) - this->data_len;
			(method->target_method->m_trampo) (self, object, 0);
			this->data_ptr = ptr;
			this->data_len = len;
			return TRUE;
		}
		
		method = dest->ReturnRuntime ()->FindActor (this->method);

		if (method) {
			(method->target_method->a_trampo) (object);
			return TRUE;
		}

		//	NIY
		
	} else if ( (this->target_ok)
	         && (this->target->AttachToObject (dest)) ) {
		
		//	This is somewhat more efficient if the target doesn't change from one
		//	"send" to the other. The AttachToObject method won't have to find the
		//	real target function in these cases.

		switch (this->target->GetType ()) {
			
			case OP_TARGET_IS_METHOD:
				{
					OPMessageRef self = this;
					Card8*       ptr  = this->data_ptr;
					Card32       len  = this->data_len;
					
					this->data_ptr = this->data_heap;
					this->data_len = sizeof (this->data_heap) - this->data_len;
					
					this->target->DoMessage (self);
					
					this->data_ptr = ptr;
					this->data_len = len;
					return TRUE;
				}

			case OP_TARGET_IS_ACTOR:
				{
					this->target->DoAction ();
					return TRUE;
				}

			default:
				break;
		}
		
		//	NIY
		
	}

	return FALSE;
}



/*
 *	Called at the beginning of the very special constuct used to
 *	send messages dynamically, which is :
 *
 *	  target <- (message - method) << a << b << c;
 *
 *	Remember that the method has already been defined, but that
 *	no parameter has been stored yet.
 */

void
OPMessage::Clear ()
{
	OP_METHOD (Clear, 0);
	
	this->SetStorageMode (STORE_SHALLOW);
	
	this->data_ptr  = this->data_heap;
	this->data_len  = sizeof (this->data_heap);
}


/*
 *	Just store the data into the heap, if possible. The real type
 *	checking is peformed by OPStorage.
 */

void
OPMessage::StoreData (const void* data, Card32 size)
{
	OP_METHOD (StoreData, 2);
	OP_ARG (1, const void*, data);
	OP_ARG (2, Card32, size);

	if (this->data_len > size) {
		
		//	Copy the data into the buffer. There is enough room. Could
		//	be optimized somewhat.
		
		Card8*       dst = (Card8*)(this->data_ptr);
		const Card8* src = (const Card8*)(data);
		
		this->data_ptr += size;
		this->data_len -= size;
		
		while (size--) {
			*dst++ = *src++;
		}
		
	} else {
		
		//	NIY
		
	}
}


/*
 *	Restore data from the heap. It will be rearranged in order to
 *	be usable by OPStorage, which will also do the type checking.
 */

void
OPMessage::RestoreData (void* data, Card32 size)
{
	OP_METHOD (RestoreData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);

	if (this->data_len >= size) {
		
		//	Copy the data into the buffer. There is enough data. Could
		//	be optimized somewhat.
		
		Card8*       dst = (Card8*)(data);
		const Card8* src = (const Card8*)(this->data_ptr);
		
		this->data_ptr += size;
		this->data_len -= size;
		
		while (size--) {
			*dst++ = *src++;
		}
		
	} else {
		
		//	NIY
		
	}
}


