/*
 *	objectbrowser.cxx
 *
 *	The Object Browser class collects object panes into a single
 *	view which allows horizontal browsing.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/button.h>
#include <OPaC/cellgrid.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/objectbrowser.h>
#include <OPaC/selgrid.h>
#include <OPaC/string.h>

#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPObjectBrowser);

OP_TARGET_ACTOR (OPObjectBrowser, ScrollLeft);
OP_TARGET_ACTOR (OPObjectBrowser, ScrollRight);

OP_TARGET_OBSERVER (OPObjectBrowser, CellActivated);
OP_TARGET_OBSERVER (OPObjectBrowser, CellDeactivated);

OP_NOTIFY_EVENT (OPObjectBrowser, IconsChanged);
OP_NOTIFY_EVENT (OPObjectBrowser, VisibleValueChanged);
OP_NOTIFY_EVENT (OPObjectBrowser, VisibleRatioChanged);

OP_NOTIFY_CHANGE (OPObjectBrowser, visible_value, VisibleValueChanged);
OP_NOTIFY_CHANGE (OPObjectBrowser, visible_ratio, VisibleRatioChanged);

OP_PUBLISH_SLOT (OPObjectBrowser, visible_value, ReturnVisibleValue, SetVisibleValue, "visible value", Real);
OP_PUBLISH_SLOT_READ (OPObjectBrowser, visible_ratio, ReturnVisibleRatio, "visible ratio", Real);

/****************************************************************/

OP_TARGET_CLASS (OPObjectBrowser);
OP_TARGET_ARG (Count, panes, "Number of panes");
OP_TARGET_ARG (OPGenericProviderRef, provider, "Data provider");
OP_TARGET_ARG (OPAtomRef, model, "Grid model");
OP_TARGET_CALL (Initialise (panes, provider, model));
OP_TARGET_END ();

/****************************************************************/

OPObjectBrowser::OPObjectBrowser ()
{
	OP_METHOD (OPObjectBrowser, 0);
	
	this->pane_mode = OP_PANE_MODE_NAME_ONLY;
	
	this->num_panes = 0;
	this->max_panes = 0;
	this->vis_panes = 0;
	this->zero_pane = 0;
	
	this->visible_offset = 0.0;
	this->icons_enabled  = TRUE;
	
	this->empty_provider = OP_NEW (OPObjectProvider);
	this->empty_list     = OP_NEW (OPList);
	
	this->empty_provider->Initialise (this->empty_list);
}

OPObjectBrowser::~OPObjectBrowser ()
{
	OP_METHOD (~OPObjectBrowser, 0);
}

void
OPObjectBrowser::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
}

void
OPObjectBrowser::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPObjectBrowser num_panes max_panes vis_panes zero_pane"
							   " visible_offset panes icons provider model icons_enabled>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->num_panes << this->max_panes;
	storage << this->vis_panes << this->zero_pane;
	storage << this->visible_offset;
	storage << this->panes << this->icons;
	storage << this->provider << this->model;
	storage << this->icons_enabled;
	storage << OPStorageSync::end;
}

void
OPObjectBrowser::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->num_panes >> this->max_panes;
	storage >> this->vis_panes >> this->zero_pane;
	storage >> this->visible_offset;
	storage >> this->panes >> this->icons;
	storage >> this->provider >> this->model;
	storage >> this->icons_enabled;
	storage >> OPStorageSync::end;
}


void
OPObjectBrowser::Initialise (Count panes, OPGenericProviderIn provider)
{
	OP_METHOD (Initialise, 2);
	OP_ARG (1, Count, panes);
	OP_ARG (2, OPGenericProviderIn, provider);
	
	this->Initialise (panes, provider, "OPSelGrid");
}


/*
 *	Create a browser object with the number of visible panes
 *	specified by the caller.
 */

void
OPObjectBrowser::Initialise (Count panes, OPGenericProviderIn provider, OPAtomIn model)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, Count, panes);
	OP_ARG (2, OPGenericProviderIn, provider);
	OP_ARG (3, OPAtomIn, model);
	
	OP_ASSERT (this->num_panes == 0);
	OP_ASSERT (panes > 0);
	
	this->num_panes = 0;
	this->max_panes = panes;			//	might grow if needed (add new panes)
	this->vis_panes = panes;			//	won't change any more
	this->provider  = provider;
	this->model     = model;
	
	this->panes = OP_NEW (OPArray);
	this->icons = OP_NEW (OPArray);
	
	this->panes->SetSize (panes);
	this->icons->SetSize (panes);
	this->UpdateGeometry ();
	
	for (Count index = 0; index < this->max_panes; index++) {
		this->CreateEmptyPane (index);
	}
	
	this->UpdateGeometry ();
	this->panes[0]->SetProvider (this->provider);
	this->panes[0]->Update ();
	
	this->num_panes = 1;
	
	this->QuickReport (notify_VisibleRatioChanged, OP_REPORT_NEW);
}



/*
 *	Called whenever the geometry changes.
 */

void
OPObjectBrowser::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 3);
	OP_ARG (1, OPRectIn, box);
	
	OPCoord dx = this->box.DX ();
	OPCoord dy = this->box.DY ();
	
	super::ChangeGeometry (box);
	
	if ( (dx != this->box.DX ())
	  || (dy != this->box.DY ()) ) {
		this->UpdateGeometry ();
	}
}

void
OPObjectBrowser::UpdateGeometry ()
{
	OP_METHOD (UpdateGeometry, 0);
	
	OPCoord x  = 0;
	OPCoord y  = 0;
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	this->GetSize (dx, dy);
	this->GetOrigin (x, y);
	
	dx -= (this->max_panes-1) * 5;
	dx /= this->max_panes;
	
	this->dx = dx;
	this->dy = dy;
	
	for (Count index = 0; index < this->max_panes; index++) {
		
		if (this->panes[index].IsValid ()) {
			this->panes[index]->SetOrigin (x, y);
			this->panes[index]->SetSize (dx, dy);
		}
		
		x += dx + 5;
	}
}




Bool
OPObjectBrowser::SetPaneMode (OPObjectPaneMode mode)
{
	OP_METHOD (SetPaneMode, 1);
	OP_ARG (1, OPObjectPaneMode, mode);
	
	this->pane_mode = mode;
	
	return TRUE;
}

/*
 *	
 */

void
OPObjectBrowser::CreateEmptyPane (Count index)
{
	OP_METHOD (CreateEmptyPane, 1);
	OP_ARG (1, Count, index);
	
	if (index >= this->max_panes) {
		this->max_panes = index+1;
		this->panes->SetSize (this->max_panes);
		this->icons->SetSize (this->max_panes);
	}
	
	for (Count from = 0; from < this->max_panes; from++) {
		if (this->panes[from].IsInvalid ()) {
			
			OPRect box;
			
			box.SetOrigin (0, 0);
			box.SetSize (this->dx, this->dy);
			
			this->panes[from] = OP_NEW (OPObjectPane);
			this->panes[from]->SetParent (this);
			this->panes[from]->Initialise (box, this->empty_provider, this->model);
			this->panes[from]->SetMode (this->pane_mode | OP_PANE_LAYOUT_LINES);
			this->panes[from]->SetNameDX (this->dx-2);
			
			this->icons[from] = OP_NEW (OPIcon);
			
			if (this->icons_enabled) {
				this->icons[from]->SetParent (this);
			}
			
			OPSelGridRef grid;
			
			this->panes[from]->GetCellGrid (grid);
			
			OP_ASSERT (grid->IsKindOf (OP_RUNTIME (OPSelGrid)));
			
			grid->AttachObserver (this, "CellActivated", "CellActivated", from);
			grid->AttachObserver (this, "CellDeactivated", "CellDeactivated", from);
			grid->SetSelectOnClick (TRUE);
			grid->SetSelectMultipleOk (FALSE);
			grid->SetSelectAtLeastOne (FALSE);
		}
	}
	
	this->panes[index]->SetProvider (this->empty_provider);
	this->panes[index]->Update ();
	this->icons[index]->Initialise (OP_ICON_NONE);
}



Bool
OPObjectBrowser::ScrollLeft ()
{
	OP_METHOD (ScrollLeft, 0);
	
	if (this->zero_pane) {
		this->SetZeroPane (this->zero_pane-1);
	}
	
	return TRUE;
}

Bool
OPObjectBrowser::ScrollRight ()
{
	OP_METHOD (ScrollRight, 0);
	
	if (this->zero_pane+this->vis_panes < this->num_panes) {
		this->SetZeroPane (this->zero_pane+1);
	}
	
	return TRUE;
}


/*
 *	Define the level up to which the panes are open.
 *	If the level is set to 0, this means that only the
 *	leftmost pane is shown.
 */

Bool
OPObjectBrowser::SetLevel (Count index, Bool notify)
{
	OP_METHOD (SetLevel, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, Bool, notify);
	
	for (Count id = index+1; id < this->max_panes; id++) {
		this->panes[id]->SetProvider (this->empty_provider);
		this->panes[id]->Update ();
		this->icons[id]->Initialise (OP_ICON_NONE);
	}
	
	this->num_panes = index+1;
	this->MakePaneVisible (index, notify, FALSE);
	this->QuickReport (notify_VisibleRatioChanged, OP_REPORT_NEW);
	this->QuickReport (notify_VisibleValueChanged, OP_REPORT_NEW);	
	
	return TRUE;
}

Count
OPObjectBrowser::ReturnLevel () const
{
	OP_METHOD (ReturnLevel, 0);
	
	return this->num_panes-1;
}


/*
 *	Return the horizontal position of the specified pane. The
 *	co-ordinates might be out of bounds if the pane is not
 *	currently visible.
 */

OPCoord
OPObjectBrowser::ReturnHorizontalPanePos (Count index) const
{
	OP_METHOD (ReturnHorizontalPanePos, 1);
	OP_ARG (1, Count, index);
	
	OPCoord x, y;
	OPCoord delta  = this->dx+5;
	int     offset = index - this->zero_pane;
	
	this->GetOrigin (x, y);
	
	return x + delta*offset;
}


/*
 *	Define which will be the leftmost visible pane.
 */

Bool
OPObjectBrowser::SetZeroPane (Count index, Bool notify_icons, Bool notify_value)
{
	OP_METHOD (SetZeroPane, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Bool, notify_icons);
	OP_ARG (3, Bool, notify_value);
	
	this->visible_offset = 0.0;
	
	if (index >= this->num_panes) {
		return FALSE;
	}
	
	if (index == this->zero_pane) {
		return TRUE;
	}
	
	this->zero_pane = index;
	
	for (Count id = 0; id < this->max_panes; id++) {
		if ( (id >= this->zero_pane)
		  && (id < this->zero_pane+this->vis_panes) ) {
			
			this->panes[id]->SetParent (this);
			this->panes[id]->SetOrigin (this->ReturnHorizontalPanePos (id), this->box.Y1 ());
			
		} else {
			
			this->panes[id]->SetParent (OPViewNull);
			this->icons[id]->SetParent (OPViewNull);
		}
	}
	
	if (notify_icons) {
		this->QuickReport (notify_IconsChanged);
	}
	
	if (notify_value) {
		this->QuickReport (notify_VisibleValueChanged, OP_REPORT_NEW);
	}
	
	return TRUE;
}

Count
OPObjectBrowser::ReturnZeroPane () const
{
	OP_METHOD (ReturnZeroPane, 0);
	
	return this->zero_pane;
}

Count
OPObjectBrowser::ReturnNumPanes () const
{
	OP_METHOD (ReturnNumPanes, 0);
	
	return this->num_panes;
}

Count
OPObjectBrowser::ReturnVisiblePanes () const
{
	OP_METHOD (ReturnVisiblePanes, 0);
	
	return this->vis_panes;
}


Bool
OPObjectBrowser::SetVisibleValue (Real value)
{
	OP_METHOD (SetVisibleValue, 1);
	OP_ARG (1, Real, value);
	
	if (this->num_panes <= this->vis_panes) {
		this->visible_offset = 0.0;
		return TRUE;
	}
	
	Count pos = (Count)(value * (this->num_panes - this->vis_panes) + 0.5);
	
	this->SetZeroPane (pos, TRUE, FALSE);
	this->visible_offset = pos - this->ReturnVisibleValue ();
	this->QuickReport (notify_VisibleValueChanged);
	
	return TRUE;
}

Real
OPObjectBrowser::ReturnVisibleValue () const
{
	OP_METHOD (ReturnVisibleValue, 0);
	
	if (this->num_panes <= this->vis_panes) {
		return 0.0;
	}
	
	return (Real)(this->zero_pane) / (this->num_panes - this->vis_panes)
			+ this->visible_offset;
}

Real
OPObjectBrowser::ReturnVisibleRatio () const
{
	OP_METHOD (ReturnVisibleRatio, 0);
	
	if (this->num_panes <= this->vis_panes) {
		return 1.0;
	}
	
	return (Real)(this->vis_panes) / (this->num_panes);
}

Bool
OPObjectBrowser::GetIcon (Count nth, OPIconOut icon, OPCoordOut x) const
{
	OP_METHOD (GetIcon, 3);
	OP_ARG (1, Count, nth);
	OP_ARG (2, OPIconOut, icon);
	OP_ARG (3, OPCoordOut, x);
	
	if ( (nth >= this->zero_pane)
	  && (nth < this->zero_pane+this->vis_panes) ) {
		
		icon = this->icons[nth];
		x    = this->ReturnHorizontalPanePos (nth);
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Return the specified pane.
 */

Bool
OPObjectBrowser::GetPane (Count nth, OPObjectPaneOut pane) const
{
	OP_METHOD (GetPane, 2);
	OP_ARG (1, Count, nth);
	OP_ARG (2, OPObjectPaneOut, pane);
	
	if ( (nth >= this->zero_pane)
	  && (nth < this->zero_pane+this->vis_panes) ) {
		pane = this->panes[nth];
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Find the provider associated with the specified pane.
 */

Bool
OPObjectBrowser::GetProvider (Count nth, OPGenericProviderOut provider) const
{
	OP_METHOD (GetProvider, 2);
	OP_ARG (1, Count, nth);
	OP_ARG (2, OPGenericProviderOut, provider);
	
	if ( (nth >= this->zero_pane)
	  && (nth < this->zero_pane+this->vis_panes) ) {
		this->panes[nth]->GetProvider (provider);
		return TRUE;
	}
	
	return FALSE;
}



void
OPObjectBrowser::MakePaneVisible (Count index, Bool notify_icons, Bool notify_value)
{
	OP_METHOD (MakePaneVisible, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Bool, notify_icons);
	OP_ARG (3, Bool, notify_value);
	
	OP_ASSERT (index < this->max_panes);
	OP_ASSERT (index < this->num_panes);
	
	Count zero    = this->zero_pane;
	int   column  = index - this->zero_pane;
	int   visible = this->vis_panes;
	
	if (column >= visible) {
		zero = index - this->vis_panes + 1;
	} else if (column < 0) {
		zero = index;
	}
	
	if (this->num_panes-zero < this->vis_panes) {
		if (this->num_panes <= this->vis_panes) {
			zero = 0;
		} else {
			zero = this->num_panes - this->vis_panes;
		}
	}
	
	this->SetZeroPane (zero, notify_icons, notify_value);
}


void
OPObjectBrowser::Cleanup ()
{
	OP_METHOD (Cleanup, 0);
}

Bool
OPObjectBrowser::CellActivated (OPObjectIn observer, Card32 id)
{
	OP_METHOD (CellActivated, 2);
	OP_ARG (1, OPObjectIn, observer);
	OP_ARG (2, Card32, id);
	
	OPGenericProviderRef provider;
	OPGenericProviderRef next;
	OPSelGridRef grid;
	
	Count x, y;
	
	this->panes[id]->GetCellGrid (grid);
	this->panes[id]->GetProvider (provider);
	
	if (grid->FindActiveHilite (x, y)) {
		
		OPLookIcon icon;
		provider->GetLargeIcon (y, icon);
		
		this->icons[id]->Initialise (icon);
		this->icons[id]->SetSize (dx, 50);
		
		if (provider->EnterSubtree (y, next)) {
			this->CreateEmptyPane (id+1);
			this->panes[id+1]->SetProvider (next);
			this->panes[id+1]->Update ();
			this->SetLevel (id+1, FALSE);
		}
		
	} else {
		
		this->icons[id]->Initialise (OP_ICON_NONE);
		this->SetLevel (id, FALSE);
	}
	
	this->QuickReport (notify_IconsChanged);
	this->Display ();
	
	return TRUE;
}

Bool
OPObjectBrowser::CellDeactivated (OPObjectIn observer, Card32 id)
{
	OP_METHOD (CellDeactivated, 2);
	OP_ARG (1, OPObjectIn, observer);
	OP_ARG (2, Card32, id);
	
	this->icons[id]->Initialise (OP_ICON_NONE);
	this->SetLevel (id, FALSE);
	this->QuickReport (notify_IconsChanged);
	this->Display ();
	
	return TRUE;
}

Bool
OPObjectBrowser::GetSelectedPath (OPStringOut path) const
{
	OP_METHOD (GetSelectedPath, 1);
	OP_ARG (1, OPStringOut, path);
	
	Count index = 0;
	path = OPStringNull;
	
	while (index < this->num_panes) {
		
		OPGenericProviderRef provider;
		OPSelGridRef grid;
		OPStringRef  text;
		Count x, y;
		
		this->panes[index]->GetCellGrid (grid);
		this->panes[index]->GetProvider (provider);
		
		if (grid->FindActiveHilite (x, y)) {
			if (provider->GetText (y, text)) {
				if (path.IsValid ()) {
					path->AppendString (".");
				} else {
					path = "";
				}
				path->AppendString (text);
			}
		}
		
		index++;
	}
	
	return TRUE;
}

Bool
OPObjectBrowser::EnableIcons (Bool yes)
{
	OP_METHOD (EnableIcons, 1);
	OP_ARG (1, Bool, yes);
	
	this->icons_enabled = yes;
	
	//	NIY: should hide/show the associated icons, if needed.
	
	return TRUE;
}

