/*
 *	objectref.cxx
 *
 *	Implementation of the base reference class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/except.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;

/****************************************************************/

Address
OPObjectRef::ReturnRawAddress () const
{
	return (Address)(this->object);
}

OPObject*
OPObjectRef::ReturnRawObject () const
{
	return this->object;
}

Bool
OPObjectRef::IsValid () const
{
	OPObject* object = this->Resolve ();
	return object ? TRUE : FALSE;
}

Bool
OPObjectRef::IsInvalid () const
{
	OPObject* object = this->Resolve ();
	return object ? FALSE : TRUE;
}


/*
 *	Return TRUE if a reference is based on a phantom object.
 *	A phantom will return a pointer to another object to the
 *	caller; only ReturnRawAddress and ReturnRawObject will
 *	allow the caller to access the phantom object itself.
 */

Bool
OPObjectRef::IsPhantom () const
{
	OPObject* object = this->Resolve ();
	return (object != this->object);
}

Bool
OPObjectRef::operator == (OPObjectIn ref) const
{
	return (this->Resolve () == ref.Resolve ());
}

Bool
OPObjectRef::operator != (OPObjectIn ref) const
{
	return (this->Resolve () != ref.Resolve ());
}

/****************************************************************/

/*
 *	Assign an other object to this reference. We need not
 *	check type conformity, since the caller guarantees it.
 */

void
OPObjectRef::_Set (OPObjectIn obj, const OPaC_CallFrame* frame)
{
//	OP_MEMBER (_Set, 2);
//	OP_ARG (1, OPObjectIn, obj);
//	OP_ARG (2, const OPaC_CallFrame*, frame);
	
	if (obj.object == this->object) {
		return;
	}
	
	if (this->object) {
		
		//	Remove ourself from the object's reference list.
		
		(this->next ? this->next->prev : this->object->ref_tail) = this->prev;
		(this->prev ? this->prev->next : this->object->ref_head) = this->next;
		
		this->next   = 0;
		this->prev   = 0;
		this->c_name = 0;
		
		if (this->object->ref_count-- == 1) {
			
			OPObjectRef::RequireDeletion (this->object);
		}
	}
	
	this->object = obj.object;
	this->next   = 0;
	this->prev   = 0;
	this->c_name = 0;
#ifdef _DEBUG
	if (OPApp::IsOnStack ((Address)(this))) {
		if (frame == 0) frame = this->frame;
		if (frame) this->frame_copy = *frame;
	} else {
		frame = 0;
	}
#endif
	this->frame  = frame;
	
	if (this->object) {
		
		//	Insert ourself into the object's reference list.
		
		this->prev = this->object->ref_tail;
		
		(this->object->ref_tail ? this->object->ref_tail->next : this->object->ref_head) = this;
		(this->object->ref_tail) = this;
		
		this->object->ref_count++;
		
		const OPRuntimeRef& rt = this->object->ReturnRuntime ();
		this->c_name = rt.IsValid () ? rt->ReturnClassName () : "?";
	}
	
	if (this->embedded_in_object) {
		this->embedded_in_object->QuickReportEmbeddedChange (*this);
	}
}

void
OPObjectRef::_Set (OPValueIn val, const OPaC_CallFrame* frame)
{
	OPObject* obj = (OPObject*) val.ReturnRawAddress ();
	
	if (val.IsValid ()) {
		if (val->IsObject ()) {
			OPObjectRef::_Set (val->ReturnObject (), frame);
			return;
		}
	}
	
	OPObjectRef::_Set (obj, frame);
}

/*
 *	Assign an other object to this reference. Check to see
 *	if the types match. If not, raise an exception.
 *
 *	Throws :
 *
 *	  - OPExcObject (MISMATCH) if the type of the inserted
 *		object does not match the type of the reference.
 */

void
OPObjectRef::_Set (OPObjectIn obj, OPRuntimeIn runtime, const OPaC_CallFrame* frame)
{
//	OP_MEMBER (_Set, 3);
//	OP_ARG (1, OPObjectIn, obj);
//	OP_ARG (2, OPRuntimeIn, runtime);
//	OP_ARG (3, const OPaC_CallFrame*, frame);
	
	if (obj.object == this->object) {
		return;
	}
	
	if ( (obj.object != 0)
	  && (obj.object->ReturnRuntime () != runtime) ) {
		
		//	The object which is being inserted has not exactly the same type as
		//	the reference. We accept the object to be a child class of "runtime"
		//	type.
		
		if (obj->IsKindOf (runtime) == FALSE) {
			if (obj->IsKindOf (OP_RUNTIME (OPValue))) {
				OPValueRef val = obj;
				this->_Set (val->ReturnObject (), runtime, frame);
				return;
			}
			throw OPExcObject (OPExcObject::MISMATCH);
		}
	}
	
	if (this->object) {
		
		//	Remove ourself from the object's reference list.
		
		(this->next ? this->next->prev : this->object->ref_tail) = this->prev;
		(this->prev ? this->prev->next : this->object->ref_head) = this->next;
		
		this->next   = 0;
		this->prev   = 0;
		this->c_name = 0;
		
		if (this->object->ref_count-- == 1) {
			OPObjectRef::RequireDeletion (this->object);
		}
	}
	
	this->object = obj.object;
	this->next   = 0;
	this->prev   = 0;
	this->c_name = runtime->ReturnClassName ();
#ifdef _DEBUG
	if (OPApp::IsOnStack ((Address)(this))) {
		if (frame == 0) frame = this->frame;
		if (frame) this->frame_copy = *frame;
	} else {
		frame = 0;
	}
#endif
	this->frame  = frame;
	
	if (this->object) {
		
		//	Insert ourself into the object's reference list.
		
		this->prev = this->object->ref_tail;
		
		(this->object->ref_tail ? this->object->ref_tail->next : this->object->ref_head) = this;
		(this->object->ref_tail) = this;
		
		this->object->ref_count++;
	}
	
	if (this->embedded_in_object) {
		this->embedded_in_object->QuickReportEmbeddedChange (*this);
	}
}

void
OPObjectRef::_Set (OPValueIn val, OPRuntimeIn runtime, const OPaC_CallFrame* frame)
{
	OPObject* obj = (OPObject*) val.ReturnRawAddress ();
	
	if (val.IsValid ()) {
		if (val->IsObject ()) {
			OPObjectRef::_Set (val->ReturnObject (), runtime, frame);
			return;
		}
	}
	
	OPObjectRef::_Set (obj, runtime, frame);
}


/*
 *	Return the size of the object referred to.
 */

Card32
OPObjectRef::SizeOf () const
{
	return sizeof (OPObject);
}

/****************************************************************/

#ifdef _DEBUG
OPObjectRef*	OPObjectRef::g_head = 0;
OPObjectRef*	OPObjectRef::g_tail = 0;
#endif

/****************************************************************/

OPObjectRef::OPObjectRef (OPRuntimeIn rt, const OPaC_CallFrame* frame)
{
	this->object = 0;
	this->next   = 0;
	this->prev   = 0;
	this->c_name = 0;
	this->embedded_in_object = 0;
#ifdef _DEBUG
	if (OPApp::IsOnStack ((Address)(this))) {
		OP_ASSERT (frame);
		this->frame_copy = *frame;
	} else {
		frame = 0;
	}
#endif
	this->frame  = frame;
	
#ifdef _DEBUG
	this->g_next = OPObjectRef::g_head;
	this->g_prev = 0;
	(this->g_next ? this->g_next->g_prev : OPObjectRef::g_tail) = this;
	OPObjectRef::g_head = this;
#endif
	
	if (OPObjectRef::report_creation) {
    	OPObjectRef::ReportCreation (this, rt);
	}
}

OPObjectRef::OPObjectRef (OPObject* ptr)
{
	this->object = ptr;
	this->next   = 0;
	this->prev   = ptr ? ptr->ref_tail : 0;
	this->c_name = 0;
	this->embedded_in_object = 0;
	this->frame  = 0;
	
#ifdef _DEBUG
	this->g_next = OPObjectRef::g_head;
	this->g_prev = 0;
	(this->g_next ? this->g_next->g_prev : OPObjectRef::g_tail) = this;
	OPObjectRef::g_head = this;
#endif
	
	if (ptr) {
		ptr->ref_count++;
		(ptr->ref_tail ? ptr->ref_tail->next : ptr->ref_head) = this;
		ptr->ref_tail = this;
		const OPRuntimeRef& rt = this->object->ReturnRuntime ();
		this->c_name  = rt.IsValid () ? rt->ReturnClassName () : "?";
	}
}

OPObjectRef::OPObjectRef (OPObjectIn obj)
{
	OPObject* ptr = obj.object;
	
	this->object = ptr;
	this->next   = 0;
	this->prev   = ptr ? ptr->ref_tail : 0;
	this->c_name = 0;
	this->embedded_in_object = 0;
	this->frame  = 0;
	
#ifdef _DEBUG
	this->g_next = OPObjectRef::g_head;
	this->g_prev = 0;
	(this->g_next ? this->g_next->g_prev : OPObjectRef::g_tail) = this;
	OPObjectRef::g_head = this;
#endif
	
	if (ptr) {
		ptr->ref_count++;
		(ptr->ref_tail ? ptr->ref_tail->next : ptr->ref_head) = this;
		ptr->ref_tail = this;
		const OPRuntimeRef& rt = this->object->ReturnRuntime ();
		this->c_name  = rt.IsValid () ? rt->ReturnClassName () : "?";
	}
}

OPObjectRef::OPObjectRef ()
{
    this->object = 0;
    this->next   = 0;
    this->prev   = 0;
	this->c_name = 0;
	this->embedded_in_object = 0;
	this->frame  = 0;
	
#ifdef _DEBUG
	this->g_next = OPObjectRef::g_head;
	this->g_prev = 0;
	(this->g_next ? this->g_next->g_prev : OPObjectRef::g_tail) = this;
	OPObjectRef::g_head = this;
#endif
	
	if (OPObjectRef::report_creation) {
    	OPObjectRef::ReportCreation (this, OPObjectRef::ReturnRuntime ());
    }
}

/****************************************************************/

OPObjectRef::~OPObjectRef ()
{
#ifdef _DEBUG
	if ( (this->g_next)
	  || (this->g_prev)
	  || (OPObjectRef::g_tail == this)
	  || (OPObjectRef::g_head == this) ) {
		if (OPObjectRef::g_head) {
			(this->g_next ? this->g_next->g_prev : OPObjectRef::g_tail) = this->g_prev;
			(this->g_prev ? this->g_prev->g_next : OPObjectRef::g_head) = this->g_next;
		}
	}
#endif
	
	if (this->object) {
		
		(this->next ? this->next->prev : this->object->ref_tail) = this->prev;
		(this->prev ? this->prev->next : this->object->ref_head) = this->next;
		
		if (this->object->ref_count-- == 1) {
			OPObjectRef::RequireDeletion (this->object);
		}
	}
}

/****************************************************************/

OPObject*
OPObjectRef::ReturnEmbeddingObject () const
{
	OP_METHOD (ReturnEmbeddingObject, 0);
	return this->embedded_in_object;
}

const OPaC_CallFrame*
OPObjectRef::ReturnEmbeddingFrame () const
{
	OP_METHOD (ReturnEmbeddingFrame, 0);
	return this->frame;
}

const LatinChar*
OPObjectRef::ReturnClassName () const
{
	OP_METHOD (ReturnClassName, 0);
	return this->c_name;
}

/****************************************************************/

OPObjectRef&
OPObjectRef::operator = (OPObjectIn obj)
{
	OPObjectRef::_Set (obj);
	return *this;
}

OPObjectRef&
OPObjectRef::operator = (OPValueIn val)
{
	OPObjectRef::_Set (val);
	return *this;
}

/****************************************************************/

OPObject*
OPObjectRef::CheckAndResolve () const
{
	return (this->object) ? this->object->CheckAndResolve ()
						  : this->ReturnRuntime ()->ResolveNull ();
}

OPObject*
OPObjectRef::Resolve () const
{
	return (this->object) ? this->object->Resolve () : 0;
}

OPObject*
OPObjectRef::operator -> () const
{
	return this->CheckAndResolve ();
}

/****************************************************************/

#define	MAX_DELETE_INFO		100

struct OPaC_DeleteInfo
{
	static OPaC_DeleteInfo*	head;
	static OPaC_DeleteInfo*	tail;
	
	OPaC_DeleteInfo*		next;
	OPObject*				objects[MAX_DELETE_INFO];
	Count					num;
	
	static void AddDeleteInfo ();
	static void PurgeDeleteInfo ();
};

OPaC_DeleteInfo* OPaC_DeleteInfo::head = 0;
OPaC_DeleteInfo* OPaC_DeleteInfo::tail = 0;

void
OPaC_DeleteInfo::AddDeleteInfo ()
{
	OPaC_DeleteInfo* info = OP_ALLOC_FROM_ZONE (OPaC_sys_zone, OPaC_DeleteInfo);
	
	info->next = 0;
	info->num  = 0;
	
	(OPaC_DeleteInfo::tail ? OPaC_DeleteInfo::tail->next : OPaC_DeleteInfo::head) = info;
	OPaC_DeleteInfo::tail = info;
}

void
OPaC_DeleteInfo::PurgeDeleteInfo ()
{
	while (OPaC_DeleteInfo::head && OPaC_DeleteInfo::head->next) {
		OPaC_DeleteInfo* info = OPaC_DeleteInfo::head->next;
		OPaC_DeleteInfo::head = info->next;
		OP_FREE_FROM_ZONE (OPaC_sys_zone, info);
	}
	
	OPaC_DeleteInfo::tail = OPaC_DeleteInfo::head;
	
	if (OPaC_DeleteInfo::head) {
		OPaC_DeleteInfo::head->next = 0;
		OPaC_DeleteInfo::head->num  = 0;
	}
}

/****************************************************************/

/*
 *	When the reference counter reaches 0, the object has
 *	to be deleted. This is done in a delayed way. For now,
 *	just register the object as being deleted.
 */

void
OPObjectRef::RequireDeletion (OPObject*& obj)
{
	OP_MEMBER (RequireDeletion, 1);
	OP_ARG (1, OPObject*, obj);
	
	OPObject* pobj = obj;
	obj = 0;
	
	//	Temporarily set the reference count to some very huge value, so
	//	the object won't be re-referenced and then de-referenced, which
	//	would induce a new call to HandleDeletion.
	
	pobj->ref_count = 0x80000000;
	
	if ( (OPaC_DeleteInfo::tail == 0)
	  || (OPaC_DeleteInfo::tail->num == MAX_DELETE_INFO) ) {
		OPaC_DeleteInfo::AddDeleteInfo ();
	}
	
	OPaC_DeleteInfo::tail->objects[OPaC_DeleteInfo::tail->num++] = pobj;
}

/*
 *	Do the real deletion. This will be called by the main
 *	event loop, so that objects can be destroyed when they
 *	are in an inactive state (otherwise, one could destroy
 *	an object which is still being used -- e.g. when a self-
 *	destruction occurs).
 */

void
OPObjectRef::RealiseDeletions ()
{
	OP_MEMBER (RealiseDeletions, 0);
	
	Count index = 0;
	OPaC_DeleteInfo* info = OPaC_DeleteInfo::head;
	
	//	Walk through the list of marked objects. These will be destroyed and
	//	freed one after the other. While destroying objects, others can be
	//	marked as deleted and the current list might grow accordingly.
	
	while (info) {
		
		if (index >= info->num) {
			index = 0;
			info  = info->next;
			continue;
		}
		
		//	First, destroy the object :
		
		OPObject*  pobj = info->objects[index++];
		OPMemZone* zone = pobj->mem_zone;
		
		pobj->_Destroy ();
		
		//	Check that everything went the way it should (no reference change
		//	for the destroyed object).
		
		OP_ASSERT (pobj->ref_count == 0x80000000);
		
		pobj->ref_count = 0;
		
		if (OPApp::IsOnStack (pobj) == FALSE) {
			
			//	Finally, remove the memory associated to the instance. We assume that
			//	the memory zone is a legal one.
			
			OP_ASSERT (zone);
			zone->Free (pobj, __LINE__, __OPaC_src_file);
		}
	}
	
	//	Free unused memory. This is useful only if lots of deletions have
	//	been done and more than MAX_DELETE_INFO objects had to be stored in
	//	the list. We maintain a chunk of MAX_DELETE_INFO pointers ready for
	//	use.
	
	OPaC_DeleteInfo::PurgeDeleteInfo ();
}

/****************************************************************/

Bool
OPObjectRef::operator < (OPMessageIn msg) const
{
	return msg->Send (*this);
}

