/*
 *	phantom.cxx
 *
 *	A phantom object just overrides the "Resolve" and "CheckedResolve"
 *	methods in order to make the references think they refer to some
 *	other, real, object. The phantom will never be seen !
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/phantom.h>
#include <OPaC/values.h>

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPPhantom);

	
OPPhantom::OPPhantom ()
{
	this->is_cache_valid  = FALSE;
	this->is_const_object = FALSE;
}

OPPhantom::~OPPhantom ()
{
}

void
OPPhantom::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	static OPStorageSync info ("<OPPhantom root path is_const_object>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->root;
	storage << this->path;
	storage << this->is_const_object;
	storage << OPStorageSync::end;
}

void
OPPhantom::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->root;
	storage >> this->path;
	storage >> this->is_const_object;
	storage >> OPStorageSync::end;
}


void
OPPhantom::UpdateCache ()
{
	OP_METHOD (UpdateCache, 0);
	
	if ( (this->is_cache_valid == FALSE)
	  || (this->is_const_object == FALSE) ) {
		
		OPObjectRef root = this->root->Resolve ();
		
		if (OPObject::QueryData (root, this->path, this->cache) == FALSE) {
			OPValueRef value;
			OPObject::QueryData (root, this->path, value);
			this->cache = value;
		}
		
		this->is_cache_valid = this->cache.IsValid ();
	}
}

Bool
OPPhantom::Initialise (OPObjectIn root, OPStringIn path, Bool is_const_object)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPObjectIn, root);
	OP_ARG (2, OPStringIn, path);
	OP_ARG (3, Bool, is_const_object);
	
	this->is_cache_valid  = FALSE;
	this->is_const_object = is_const_object;
	this->root = root;
	this->path = path;
	
	return TRUE;
}


OPObject*
OPPhantom::Resolve ()
{
	this->UpdateCache ();
	
	if (this->is_cache_valid) {
		return this->cache.ReturnRawObject ();
	}
	
	return 0;
}

OPObject*
OPPhantom::CheckAndResolve ()
{
	this->UpdateCache ();
	
	if (this->is_cache_valid) {
		return this->cache.ReturnRawObject ();
	}
	
	return this->ReturnRuntime ()->ResolveNull ();
}

