/*
 *	respond.cxx
 *
 *	Base Responder class definition. The responder is the generic
 *	event receiver in an OPaC user interface.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/view.h>
#include <OPaC/new.h>
#include <OPaC/event.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPResponder);

/****************************************************************/


/*
 *	Create an empty responder.
 */

OPResponder::OPResponder ()
{
	OP_METHOD (OPResponder, 0);
	
	//	The next responder will be linked by some external process, eg. when a
	//	view gets put into a window, it will get its next responder...
	
	this->next_responder = 0;
}



/*
 *	Destroy the responder.
 */

OPResponder::~OPResponder ()
{
	OP_METHOD (~OPResponder, 0);
	
	this->next_responder = 0;
}



/*
 *	Nothing special to save ! The link with the next
 *	responder is being maintained by other means.
 */

void
OPResponder::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPResponder>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << OPStorageSync::end;
}



/*
 *	Nothing either to restore...
 */

void
OPResponder::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::end;
}



/*
 *	Connect to the next responder. This will get the
 *	events this responder isn't able to process.
 */

void
OPResponder::SetNextResponder (OPResponderIn responder)
{
	OP_METHOD (SetNextResponder, 1);
	OP_ARG (1, OPResponderIn, responder);
	
	this->next_responder = responder;
}



/*
 *	Return a reference to the next responder.
 */

Bool
OPResponder::GetNextResponder (OPResponderOut next) const
{
	OP_METHOD (GetNextResponder, 1);
	OP_ARG (1, OPResponderOut, next);
	
	next = this->next_responder;
	return next.IsValid ();
}



/*
 *	Return the reference to the associated view object.
 *	This will usually return "this" for the view classes
 *	and an invalid reference for the other cases.
 */

Bool
OPResponder::GetView (OPViewOut view)
{
	OP_METHOD (GetView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = 0;
	return FALSE;
}



/*
 *	Return TRUE if the responder would accept to become
 *	the Main Responder. Default is FALSE.
 */

Bool
OPResponder::WouldAcceptMain () const
{
	OP_METHOD (WouldAcceptMain, 0);
	return FALSE;
}



/*
 *	Tell the responder it has just become the Main
 *	Responder in the current window. This will only be
 *	called if the WouldAcceptBegin returned TRUE.
 */

void
OPResponder::BecomeMain (Bool clicked)
{
	OP_METHOD (BecomeMain, 1);
	OP_ARG (1, Bool, clicked);
}



/*
 *	The responder won't be the Main Responder any more.
 *	This gets called just before another one become the
 *	Main Responder.
 */

void
OPResponder::ResignMain ()
{
	OP_METHOD (ResignMain, 0);
}



/*
 *	Dispatch an event.
 */

void
OPResponder::DispatchEvent (OPEventIn event)
{
	OP_METHOD (DispatchEvent, 1);
	OP_ARG (1, OPEventIn, event);
	
	switch (event->ReturnType ()) {
		
		case OP_EVENT_KEY_DOWN:
		case OP_EVENT_KEY_REP:
			this->KeyDown (event);
			break;
		
		case OP_EVENT_KEY_UP:
			this->KeyUp (event);
			break;
		
		case OP_EVENT_KEY_CHANGE:
			this->KeyChange (event);
			break;
		
		case OP_EVENT_PRESS:
			this->ButtonPress (event);
			break;
		
		case OP_EVENT_RELEASE:
			this->ButtonRelease (event);
			break;
		
		case OP_EVENT_MOTION:
			event->ReturnButtons () ? this->DragNotify (event)
									: this->MotionNotify (event);
			break;
		
		default:
			this->Miscellanous (event);
			break;
	}
}



/*
 *	Handle OP_EVENT_KEY_DOWN and OP_EVENT_KEY_REP events.
 */

void
OPResponder::KeyDown (OPEventIn event)
{
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->KeyDown (event);
	}
}



/*
 *	Handle OP_EVENT_KEY_UP events.
 */

void
OPResponder::KeyUp (OPEventIn event)
{
	OP_METHOD (KeyUp, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->KeyUp (event);
	}
}



/*
 *	Handle OP_EVENT_KEY_CHANGE events.
 */

void
OPResponder::KeyChange (OPEventIn event)
{
	OP_METHOD (KeyChange, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->KeyChange (event);
	}
}



/*
 *	Handle OP_EVENT_PRESS events.
 */

void
OPResponder::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->ButtonPress (event);
	}
}



/*
 *	Handle OP_EVENT_RELEASE events.
 */

void
OPResponder::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->ButtonRelease (event);
	}
}



/*
 *	Handle OP_EVENT_MOTION events while no button is
 *	being pressed.
 */

void
OPResponder::MotionNotify (OPEventIn event)
{
	OP_METHOD (MotionNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->MotionNotify (event);
	}
}



/*
 *	Handle OP_EVENT_MOTION events while one (or more)
 *	button(s) is/are being pressed.
 */

void
OPResponder::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->DragNotify (event);
	}
}



/*
 *	Handle a tick event. This will be generated when a
 *	specified delay has elapsed.
 */

void
OPResponder::Tick ()
{
	OP_METHOD (Tick, 0);
}



/*
 *	Handle other non-recognized events.
 */

void
OPResponder::Miscellanous (OPEventIn event)
{
	OP_METHOD (Miscellanous, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->Miscellanous (event);
	}
}



/*
 *	The window or the application classes will tell the
 *	responder when the graphical pointer enters its box.
 */

void
OPResponder::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->EnterNotify (event);
	}
}



/*
 *	The window or the application classes will tell the
 *	responder when the graphical pointer leaves its box.
 */

void
OPResponder::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->next_responder.IsValid ()) {
		this->next_responder->LeaveNotify (event);
	}
}

