/*
 *	slider.cxx
 *
 *	Slider class.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/slider.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPSlider);

/****************************************************************/

OP_PUBLISH_SLOT (OPSlider, slider_value,  ReturnSliderValue,  SetSliderValue,  "Slider value", Real);
OP_PUBLISH_SLOT (OPSlider, slider_scale,  ReturnSliderScale,  SetSliderScale,  "Slider scale", Real);
OP_PUBLISH_SLOT (OPSlider, slider_offset, ReturnSliderOffset, SetSliderOffset, "Slider offset", Real);

OP_NOTIFY_CHANGE (OPSlider, slider_value, Scroll);

/****************************************************************/

/*
 *	The scroller automatically sets up two scroll buttons
 *	which will be used to scroll up/down automatically.
 */

OPSlider::OPSlider ()
{
	OP_METHOD (OPSlider, 0);
	
	this->slider_scale   = 1.0;
	this->slider_offset  = 0.0;
	this->init_arrow_pos = OP_SCROLL_ARROW_NONE;
	this->real_arrow_pos = OP_SCROLL_ARROW_NONE;
}



/*
 *	Clean up the scroller : delete both buttons. There is
 *	no need to disconnect the buttons; this will be done
 *	automatically.
 */

OPSlider::~OPSlider ()
{
	OP_METHOD (~OPSlider, 0);
}



/*
 *	Store the scroller...
 */

void
OPSlider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	static OPStorageSync info ("<OPSlider slider_scale slider_offset>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->slider_scale << this->slider_offset;
	storage << OPStorageSync::end;
}



/*
 *	Restore the scroller...
 */

void
OPSlider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->slider_scale >> this->slider_offset;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the scroller for destruction. Give up both
 *	arrow buttons.
 */

void
OPSlider::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->arrow_incr.IsValid ()) {
		this->arrow_incr->Delete ();
		this->arrow_incr = 0;
	}
	
	if (this->arrow_decr.IsValid ()) {
		this->arrow_decr->Delete ();
		this->arrow_decr = 0;
	}
	
	super::Delete ();
}


/*
 *	Update the current knob size (depending on the visibility
 *	ratio).
 */

void
OPSlider::UpdateKnobSize ()
{
	OP_METHOD (UpdateKnobSize, 0);
	
	this->knob_size = this->look_breadth - this->look_frame * 2;
}	

/*
 *	Compute the new button and knob positions/sizes after
 *	a geometry modification.
 */

void
OPSlider::CheckSpaceForParts ()
{
	OP_METHOD (CheckSpaceForParts, 0);
	
	super::CheckSpaceForParts ();
}



/*
 *	Draw the scroller's knob. I am still not sure if this
 *	solution is good or if the knob should be implemented
 *	as a separate object.
 */

void
OPSlider::DrawKnob (OPGrailPortIn port)
{
	OP_METHOD (DrawKnob, 1);
	OP_ARG (1, OPGrailPortIn, port);
	
	OP_ASSERT (this->is_knob_ok);
	
	//	The adorner will be used to do the drawing. This provides some easy way
	//	to change the overall look of OPaC applications.
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	//	Compute the bounding box where the knob may be located. This depends on
	//	the scroller's bounding box and on the buttons which have been used.
	
	OPCoord x1 = (this->is_horiz) ? (this->knob_min) : (this->box.X1 () + this->look_frame);
	OPCoord y1 = (this->is_horiz) ? (this->box.Y1 () + this->look_frame) : (this->knob_min);
	
	//	Now find out where the knob has to be drawn. The knob may be pressed
	//	and/or hilited...
	
	(this->is_horiz ? x1 : y1) += this->knob_pos;
	(this->is_horiz ? y1 : x1) -= this->look_frame;
	
	adorner->DrawIcon (x1, y1, this->is_horiz ? OP_ICON_V_SLIDER_SMALL : OP_ICON_H_SLIDER_SMALL);
}


/*
 *	Draw the scroller's body.
 */

void
OPSlider::DrawBody (OPGrailPortIn port)
{
	OP_METHOD (DrawBody, 1);
	OP_ARG (1, OPGrailPortIn, port);
	
	OP_ASSERT (this->is_knob_ok);
	
	//	The adorner will be used to do the drawing. This provides some easy way
	//	to change the overall look of OPaC applications.
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	//	Compute the bounding box where the knob may be located. This depends on
	//	the scroller's bounding box and on the buttons which have been used.
	
	OPCoord x1 = (this->is_horiz) ? (this->knob_min) : (this->box.X1 () + this->look_frame);
	OPCoord y1 = (this->is_horiz) ? (this->box.Y1 () + this->look_frame) : (this->knob_min);
	OPCoord x2 = (this->is_horiz) ? (this->knob_max) : (this->box.X2 () - this->look_frame);
	OPCoord y2 = (this->is_horiz) ? (this->box.Y2 () - this->look_frame) : (this->knob_max);
	
	OPRect inside (x1, y1, x2, y2);
	
	//	Draw the slider's slit.
	
	adorner->FillRect (inside, OP_COLOR_WIDGET_BACK, 6);
}



/*
 *	Initialise an arrow button.
 */

void
OPSlider::InitArrow (OPIconInOut arrow, Card32 id)
{
	OP_METHOD (InitArrow, 1);
	OP_ARG (1, OPIconInOut, arrow);
	
}

/*
 *	Initialise the icons after having allocated the arrows.
 */

void
OPSlider::InitIcons (Bool is_horiz)
{
	OP_METHOD (InitIcons, 1);
	OP_ARG (1, Bool, is_horiz);
}


/*
 *	Set the arrow button's position.
 */

void
OPSlider::SetArrowPos (OPScrollArrowPos pos)
{
	OP_METHOD (SetArrowPos, 1);
	OP_ARG (1, OPScrollArrowPos, pos);
	
	this->init_arrow_pos = OP_SCROLL_ARROW_NONE;
	this->CheckSpaceForParts ();
}


/*
 *	Scroll the knob, since the user pressed one of the
 *	arrow buttons.
 */

void
OPSlider::Scroll (Card32 id)
{
	OP_METHOD (Scroll, 1);
	OP_ARG (1, Card32, id);
	
	super::Scroll (id);
}



/*
 *	Resizing a scroller might change its shape, since the
 *	arrows may get suppressed.
 */

void
OPSlider::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
}


/*
 *	Set value which should be displayed by the slider, taking
 *	into account the scale and the offset. Define these before
 *	calling this method.
 */

Bool
OPSlider::SetSliderValue (Real value)
{
	OP_METHOD (SetSliderValue, 1);
	OP_ARG (1, Real, value);
	
	if (this->slider_scale) {
		return this->SetVisibleValue ((value - this->slider_offset) / this->slider_scale);
	} else {
		return this->SetVisibleValue (value);
	}
}

Bool
OPSlider::SetSliderScale (Real scale)
{
	OP_METHOD (SetSliderScale, 1);
	OP_ARG (1, Real, scale);
	
	Real value = this->ReturnSliderValue ();
	this->slider_scale = scale;
	return this->SetSliderValue (value);
}

Bool
OPSlider::SetSliderOffset (Real offset)
{
	OP_METHOD (SetSliderOffset, 1);
	OP_ARG (1, Real, offset);
	
	Real value = this->ReturnSliderValue ();
	this->slider_offset = offset;
	return this->SetSliderValue (value);
}


/*
 *	Return the current visible value. See SetVisibleValue.
 */

Real
OPSlider::ReturnSliderValue () const
{
	OP_METHOD (ReturnSliderValue, 0);
	
	return (this->slider_scale) ? this->value * this->slider_scale + this->slider_offset
								: this->value;
}

Real
OPSlider::ReturnSliderScale () const
{
	OP_METHOD (ReturnSliderScale, 0);
	return this->slider_scale;
}

Real
OPSlider::ReturnSliderOffset () const
{
	OP_METHOD (ReturnSliderOffset, 0);
	return this->slider_offset;
}
