/*
 *	springboard.cxx
 *
 *	A spring board is a collection of spring bands.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/array.h>
#include <OPaC/spring.h>
#include <OPaC/springboard.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPSpringBoard);

/****************************************************************/


/*
 *	Create an empty spring board.
 */

OPSpringBoard::OPSpringBoard ()
{
	OP_METHOD (OPSpringBoard, 0);
	this->band = OP_NEW (OPArray);
}



/*
 *	Destroy.
 */

OPSpringBoard::~OPSpringBoard ()
{
	OP_METHOD (~OPSpringBoard, 0);
}



/*
 *	Store every band into the archive.
 */

void
OPSpringBoard::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	super::StoreInstance (storage);

	static OPStorageSync info ("<OPSpringBoard band>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->band;
	storage << OPStorageSync::end;
}



/*
 *	Restore every band from the archive.
 */

void
OPSpringBoard::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->band;
	storage >> OPStorageSync::end;
}



/*
 *	Remove every band from the board.
 */

void
OPSpringBoard::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);

	this->band->SetSize (0);
}


/*
 *	Prepare the object for its destruction.
 */

void
OPSpringBoard::Delete ()
{
	OP_METHOD (Delete, 0);

	super::Delete ();
	this->RemoveAll ();
}



/*
 *	Tell every band that the parent window did resize. This
 *	will move the subviews and do lots of complicated work.
 */

void
OPSpringBoard::Resize (OPRectIn rect)
{
	OP_METHOD (Resize, 1);
	OP_ARG (1, OPRectIn, rect);

	Count index = 0;
	Count num   = this->band->TotalCount ();

	while (num--) {
		OPSpringBandRef spring_band = this->band[index++];
		spring_band->Resize (rect);
	}
}


/*
 *	Remove the specified view (if it existed in any of the
 *	board's bands, return TRUE).
 */

Bool
OPSpringBoard::RemoveView (OPViewIn view)
{
	OP_METHOD (RemoveView, 1);
	OP_ARG (1, OPViewIn, view);
	
	Count index = 0;
	Count num   = this->band->TotalCount ();
	Bool  found = FALSE;
	
	while (num--) {
		OPSpringBandRef spring_band = this->band[index++];
		found |= spring_band->RemoveView (view);
	}
	
	return found;
}


/*
 *	Return TRUE if the view is contained in one of the bands.
 *	The optional index specifies from which band to start
 *	searching and returns the first matching band.
 */

Bool
OPSpringBoard::FindView (OPViewIn view) const
{
	OP_METHOD (FindView, 1);
	OP_ARG (1, OPViewIn, view);
	
	Count index = 0;
	
	return this->FindView (view, index);
}


Bool
OPSpringBoard::FindView (OPViewIn view, CountInOut index) const
{
	OP_METHOD (FindView, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, CountInOut, index);
	
	Count num   = this->band->TotalCount ();
	Bool  found = FALSE;
	
	for ( /* start with given index */ ; index < num; index++) {
		if (this->band[index]->FindView (view)) return TRUE;
	}
	
	return FALSE;
}



/*
 *	Return the number of bands contained in this board.
 */

Count
OPSpringBoard::CountBands () const
{
	OP_METHOD (CountBands, 0);

	return this->band->TotalCount ();
}



/*
 *	Insert a band at a specified position in the band
 *	array. If out of array, add at the end of it.
 */

Bool
OPSpringBoard::InsertBand (Count index, OPSpringBandIn band)
{
	OP_METHOD (InsertBand, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringBandIn, band);
	
	if (index > this->band->TotalCount ()) {
		index = this->band->TotalCount ();
	}
	
	this->band->Insert (index);
	this->band[index] = band;
	
	return TRUE;
}


/*
 *	Remove a band from a specified position in the band
 *	array. Return FALSE if out of array.
 */

Bool
OPSpringBoard::RemoveBand (Count index)
{
	OP_METHOD (RemoveBand, 1);
	OP_ARG (1, Count, index);
	
	if (index > this->band->TotalCount ()) {
		return FALSE;
	}
	
	this->band->Remove (index);
	return TRUE;
}



/*
 *	Return the specified band. If out of array, return FALSE.
 */

Bool
OPSpringBoard::GetNthBand (Count index, OPSpringBandOut band) const
{
	OP_METHOD (GetNthBand, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringBandOut, band);

	if (index > this->band->TotalCount ()) {
		band = 0;
		return FALSE;
	}
	
	band = this->band[index];
	return TRUE;
}



/*
 *	Look for the band in the array. If it cannot be found,
 *	return FALSE.
 */

Bool
OPSpringBoard::FindBand (OPSpringBandIn band, CountOut index) const
{
	OP_METHOD (FindBand, 2);
	OP_ARG (1, OPSpringBandIn, band);
	OP_ARG (2, CountOut, index);

	index = 0;

	while (index < this->band->TotalCount ()) {
		if (this->band[index] == band) {
			return TRUE;
		}
		index++;
	}
	
	index = 0;
	return FALSE;
}

Bool
OPSpringBoard::ReplaceView (OPViewIn old_view, OPViewIn new_view)
{
	OP_METHOD (ReplaceView, 2);
	OP_ARG (1, OPViewIn, old_view);
	OP_ARG (2, OPViewIn, new_view);
	
	Count total = this->band->TotalCount ();
	Count index = 0;

	while (index < total) {
		if (this->band[index]->FindView (old_view)) {
			this->band[index]->ReplaceView (old_view, new_view);
		}
		index++;
	}
	
	return TRUE;
}

