/*
 *	storage.cxx
 *
 *	Generic storage class used to store and restore simple and complex data
 *	types. Look for the OPArchive and OPMessage classes to have a feel about
 *	some real implementations.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/colorpixel.h>
#include <OPaC/database.h>
#include <OPaC/dataslot.h>
#include <OPaC/ieeemath.h>
#include <OPaC/listree.h>
#include <OPaC/object.h>
#include <OPaC/phantom.h>
#include <OPaC/new.h>
#include <OPaC/values.h>

#include <math.h>
#include <string.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_ABSTRACT_CLASS (OPStorage);

/****************************************************************/

struct OPStorage::RefBase
{
	RefBase*		next;						//	next object description
	OPObjectRef		ref;						//	reference to the object
	Card32			id;							//	id of reference
	Bool			stored;						//	TRUE => object has been stored
};

struct OPStorage::NameBase
{
	NameBase*		next;						//	next name in data base
	Card32			check;						//	"sum" of characters (speed up)
	Card32			id;							//	id of that name (unique)
	LatinChar		name[1];					//	name
};


/*
 *	Some objects need not be saved, since they are known to be
 *	permanent. These can be found in the slot. To speed up the
 *	accesses, a temporary copy of the slot's contents is stored
 *	in the dictionnary.
 */

struct OPStorage::StorageInfos
{
	OPDataSlotRef		slot;					//	data slot associated to the storage
	OPDictRef			dict;					//	dictionnary to speed up data slot accesses
	OPLisTreeRef		more;					//	more info on unknown data
	Count				skip;					//	number of blocks to skip
};

/****************************************************************/

/*
 *	A discrete list of primitive data types exists. It is
 *	given below.
 */


#define	TYP_NULL		((Card8)(0x00))			//	 1: null reference, null object, etc.
#define	TYP_SYNC_BEGIN	((Card8)(0xFC))			//	 1: synchronisation for beginning block 
#define	TYP_SYNC_END	((Card8)(0xFD))			//	 1: synchronisation for ending block
#define	TYP_SYNC_MARK	((Card8)(0xFE))			//	 1: synchronisation mark
#define	TYP_EOF			((Card8)(0xFF))

#define	TYP_BOOL_FALSE	((Card8)(0x08))			//	 1: Boolean FALSE value
#define	TYP_BOOL_TRUE	((Card8)(0x09))			//	 1: Boolean TRUE value

#define	TYP_INT32		((Card8)(0x10))			//	 5: 32-bit (signed) integer
#define	TYP_CARD32		((Card8)(0x11))			//	 5: 32-bit (unsigned) integer
#define	TYP_ADDR32		((Card8)(0x12))			//	 5: 32-bit address -- used by obj. pointers

#define	TYP_INT32_8		(TYP_INT32+4)			//	 2: 32-bit (signed) integer
#define	TYP_INT32_16	(TYP_INT32+8)			//	 3: 32-bit (signed) integer
#define	TYP_INT32_ZERO	(TYP_INT32+12)			//	 1: zero

#define	TYP_CARD32_8	(TYP_CARD32+4)			//	 2: 32-bit (unsigned) integer
#define	TYP_CARD32_16	(TYP_CARD32+8)			//	 3: 32-bit (unsigned) integer
#define	TYP_CARD32_ZERO	(TYP_CARD32+12)			//	 1: zero

#define	TYP_INT64		((Card8)(0x20))			//	 9: 64-bit (signed) integer
#define	TYP_CARD64		((Card8)(0x21))			//	 9: 64-bit (unsigned) integer
#define	TYP_ADDR64		((Card8)(0x22))			//	 9: 64-bit address -- used by obj. pointers

#define	TYP_REAL		((Card8)(0x30))			//	13: real value (sign exp.mant.)
#define	TYP_REAL_INT	((Card8)(0x31))			//	1+: compact real value (as an int)
#define	TYP_REAL_ZERO	((Card8)(0x32))			//	 1: real value 0.0
#define	TYP_REAL_PONE	((Card8)(0x33))			//	 1: real value +1.0
#define	TYP_REAL_MONE	((Card8)(0x34))			//	 1: real value -1.0

#define	TYP_O_DECL_NAME	((Card8)(0x40))			//	nn: object declared by real name
#define	TYP_O_DECL_ID	((Card8)(0x41))			//	 5: object declared by id'd name
#define	TYP_O_INST_REF	((Card8)(0x42))			//	 5: object instanciated by reference
#define	TYP_O_DATA		((Card8)(0x43))			//	 5: object data by id
#define	TYP_O_SLOT_DATA	((Card8)(0x44))			//	1+: object in data slot, declared by name
#define	TYP_O_PHANTOM	((Card8)(0x45))			//	1+: object results from a complex query

#define	TYP_A_DECL_NAME	((Card8)(0x50))			//	nn: atom declared by name
#define	TYP_A_DECL_ID	((Card8)(0x51))			//	 5: atom declared by id'd name
#define	TYP_A_DECL_OBJ	((Card8)(0x52))			//	1+: atom declared by object id

#define	TYP_ARRAY		((Card8)(0x60))			//	 2: array of references
#define	TYP_STRING		((Card8)(0x61))			//	1+: string object

#define	TYP_BIN_8		((Card8)(0x70))			//	 2: binary data (8-bit size)
#define	TYP_BIN_16		((Card8)(0x71))			//	 3: binary data (16-bit size)
#define	TYP_BIN_32		((Card8)(0x72))			//	 5: binary data (32-bit size)

#define	TYP_TYPE_INFO	((Card8)(0x80))			//	1+: type information

// NIY: types REAL/ATOM/ARRAY have not been implemented properly.

/****************************************************************/

/*
 *	Initialise the abstract storage instance. Default will
 *	be a "binary" storage of the objects, which will have
 *	to be overridden by something more useful.
 */

OPStorage::OPStorage ()
{
	OP_METHOD (OPStorage, 0);
	this->storage_mode = STORE_BIN_ONLY;
	
	this->r_base = 0;
	this->n_base = 0;
	
	this->num_pending_store   = 0;
	this->num_recursion       = 0;
	this->storage_infos       = NEW_FROM_ZONE (this->ReturnMemZone ()) StorageInfos;
	this->storage_infos->slot = 0;
	this->storage_infos->dict = 0;
	this->storage_infos->more = 0;
	this->storage_infos->skip = 0;
	
	this->push_size = 0;
	this->id_base   = 0;
}



/*
 *	Destroy the abstract storage instance. There's nothing
 *	special to do... apart from destroying the object DB.
 */

OPStorage::~OPStorage ()
{
	OP_METHOD (~OPStorage, 0);
	
	//	Clean up the internal data bases.
	
	this->SetStorageMode (STORE_BIN_ONLY);

	if (this->storage_infos) {
		
		if (this->storage_infos->slot.IsValid ()) {
//			this->storage_infos->slot->Delete ();
			this->storage_infos->slot = 0;
		}
		if (this->storage_infos->dict.IsValid ()) {
			this->storage_infos->dict->Delete ();
			this->storage_infos->dict = 0;
		}
		
		DELETE_FROM_ZONE (this->ReturnMemZone (), this->storage_infos);
		
		this->storage_infos = 0;
	}
}	



/*
 *	Prepare the storage for deletion.
 */

void
OPStorage::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->storage_infos) {
		if (this->storage_infos->slot.IsValid ()) {
//			this->storage_infos->slot->Delete ();
			this->storage_infos->slot = 0;
		}
		if (this->storage_infos->dict.IsValid ()) {
			this->storage_infos->dict->Delete ();
			this->storage_infos->dict = 0;
		}
	}
	
	super::Delete ();
}



/*
 *	Search the data slot to see if we can find the specifed
 *	object. If so, we need not save it : its name will be
 *	enough to restore the link to it.
 */

Bool
OPStorage::FindNameFromSlot (OPObjectIn object, OPAtomOut name)
{
	OP_METHOD (FindNameFromSlot, 2);
	OP_ARG (1, OPObjectIn, object);
	OP_ARG (2, OPAtomOut, name);
	
	if ( (this->storage_infos->dict.IsValid ()) 
	  && (this->storage_infos->dict->GetFromValue (object, name)) ) {
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Find the specified object in the data slot and return
 *	its name. Return FALSE if the object is not in the data
 *	slot.
 */

Bool
OPStorage::FindObjectFromSlot (OPAtomIn name, OPObjectOut object)
{
	OP_METHOD (FindObjectFromSlot, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, object);
	
	if ( (this->storage_infos->slot.IsValid ()) 
	  && (this->storage_infos->slot->FindField (name, object)) ) {
		return TRUE;
	}
	
	object = 0;
	return FALSE;
}


/*
 *	Same as "RestoreData" provided by subclasses, but this
 *	provides the pushed data first...
 */

void
OPStorage::PopAndRestoreData (void* data, Card32 size)
{
	OP_METHOD (PopAndRestoreData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);
	
	Byte* ptr = (Byte*)(data);
	
	while (size && this->push_size) {
		*ptr++ = this->push_buffer[--this->push_size];
		size--;
	}
	
	if (size) {
		this->RestoreData (ptr, size);
	}
}


/*
 *	If too much data has been read by the caller, it can
 *	be pushed back (only for internal use).
 */

void
OPStorage::PushBackData (void* data, Card32 size)
{
	OP_METHOD (PushBackData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);
	
	Byte* ptr = (Byte*)(data);
	
	OP_ASSERT (this->push_size + size <= sizeof (this->push_buffer));
	
	while (size) {
		this->push_buffer[this->push_size++] = ptr[--size];
	}
}


/*
 *	Store an ID. This will be saved as an 8-bit ID if it fits into the
 *	active "ID page". Otherwise, we will emit a page change first.
 */

void
OPStorage::StoreID (Card32 id)
{
	OP_METHOD (StoreID, 1);
	OP_ARG (1, Card32, id);
	
	Card32 high = id & 0xFFFFFF00;
	Card32 low  = id & 0x000000FF;
	
	if ( (high == this->id_base)
	  && (low  != 0xFF) ) {
		Card8 data[1] = { (Card8)(id >> 0) };
		this->StoreData (data, 1);
		return;
	}
	
	Card8 data[5] = {
		(Card8)(0xFF),
		(Card8)(id >> 24),
		(Card8)(id >> 16),
		(Card8)(id >>  8),
		(Card8)(id >>  0)
	};
	
	this->StoreData (data, 5);
	this->id_base = high;
}

/*
 *	Restore an ID. This is the opposit of "StoreID" and will handle
 *	properly the 8-bit and prefixed 32-bit IDs.
 */

Card32
OPStorage::RestoreID ()
{
	OP_METHOD (RestoreID, 0);
	
	Card32 id;
	Card8  data[4];
	
	this->PopAndRestoreData (data, 1);
	
	if (data[0] != 0xFF) {
		return this->id_base | data[0];
	}
	
	this->PopAndRestoreData (data, 4);
	
	id = ((Card32)(data[0]) << 24)
	   + ((Card32)(data[1]) << 16)
	   + ((Card32)(data[2]) <<  8)
	   + ((Card32)(data[3]) <<  0);
	
	this->id_base = id & 0xFFFFFF00;
	
	return id;
}

/*
 *	Define the data slot which is associated with the storage
 *	class.
 */

void
OPStorage::SetDataSlot (OPDataSlotIn slot)
{
	OP_METHOD (SetDataSlot, 1);
	OP_ARG (1, OPDataSlotIn, slot);
	OP_ASSERT (this->storage_infos);

	this->storage_infos->slot = 0;
	this->storage_infos->dict = 0;
	
	if (slot.IsValid () == FALSE) {
		return;
	}

	this->storage_infos->slot = slot;
	this->storage_infos->dict = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPDict);
	
	Count       index  = 0;
	OPAtomRef   name   = "";
	OPObjectRef object = 0;
	
	const OPDataBaseRef& db = this->storage_infos->slot->ReturnFieldBase ();
	
	//	For each field stored in the data slot, store an entry into our
	//	own dictionary. The dictionary will be used more efficiently than
	//	the data slot.
	
	while ( (db->GetAttribute ("FieldName", index, "FieldName", name))
		 && (db->GetAttribute ("FieldName", index, "FieldData", object)) ) {
		this->storage_infos->dict->Insert (name, object);
		index++;
	}
}



/*
 *	Remove the data slot which is associated with the storage
 *	class.
 */

void
OPStorage::RemoveDataSlot (OPDataSlotIn slot)
{
	OP_METHOD (RemoveDataSlot, 1);
	OP_ARG (1, OPDataSlotIn, slot);
	OP_ASSERT (this->storage_infos);

	if (slot.IsValid () == FALSE) {
		return;
	}
	
	if (this->storage_infos->slot == slot) {
		this->storage_infos->slot = 0;
		this->storage_infos->dict = 0;
	}

	//	NIY
}



/*
 *	Just return the slot associated to the storage class.
 */

const OPDataSlotRef&
OPStorage::ReturnDataSlot (Count index) const
{
	OP_METHOD (ReturnDataSlot, 1);
	OP_ARG (1, Count, index);
	
	OP_ASSERT (this->storage_infos);
	OP_ASSERT (index == 0);

	return this->storage_infos->slot;
}



/*
 *	Look for the specified object in the object data base
 *	and update it if needed.
 */

Card32
OPStorage::LookForRef (OPObjectIn obj, BoolOut is_new)
{
	OP_METHOD (LookForRef, 2);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, BoolOut, is_new);
	
	OP_ASSERT (obj.IsValid ());
	
	is_new = FALSE;
	
	//	Travel through the list of stored references until some match occurs
	//	or until the end of the list has been reached.
	
	RefBase* iter = this->r_base;
	
	while ((iter != 0) && (iter->ref != obj)) {
		iter = iter->next;
	}
	
	//	Couldn't we find the right object ? If not, we have to allocate some
	//	room in the data base and record it...
	
	if (iter == 0) {
		
		is_new = TRUE;
		iter   = NEW_FROM_ZONE (this->ReturnMemZone ()) RefBase;
		
		iter->next   = this->r_base;
		iter->ref    = obj;
		iter->id     = (this->r_base) ? (this->r_base->id + 1) : (1);
		iter->stored = FALSE;
		
		this->r_base = iter;
		
		//	Update the number of pending stores.
		
		this->num_pending_store++;
	}
	
	return iter->id;
}


/*
 *	Compute some checksum for a string, which will enable
 *	us to speed up comparisons by rejecting string which
 *	do not match.
 */

static Card32
ComputeCheckSum (const LatinChar* name)
{
	Card32 check = 0;
	Card32 shift = 0;
	Card32 value = 0;
	
	while (*name) {
		value  = (Card8)(*name++);
		check += (value) << (shift++);
		shift &= 0x0F;
	}
	
	return check;
}



/*
 *	Look for the specified name in the name data base and
 *	update it if needed.
 */

Card32
OPStorage::LookForName (const LatinChar* name, BoolOut is_new)
{
	OP_METHOD (LookForName, 2);
	OP_ARG (1, const LatinChar*, name);
	OP_ARG (2, BoolOut, is_new);
	
	OP_ASSERT (name);
	
	is_new = FALSE;
	
	
	//	Travel through the list of stored names until some match occurs or
	//	until the end of the list has been reached.
	
	NameBase* iter  = this->n_base;
	Card32    check = ComputeCheckSum (name);
	
	while ( (iter != 0)
	     && ( (iter->check != check)
	       || (strcmp (iter->name, name) != 0) ) ) {
		iter = iter->next;
	}
	
	//	Couldn't we find the right name ? If not, we have to allocate some
	//	room in the data base and record it...
	
	if (iter == 0) {
		
		Card32 len  = strlen (name);
		Card32 size = sizeof (NameBase) + len;
		
		is_new = TRUE;
		iter   = (NameBase*)(OP_ALLOC_ARRAY_FROM_ZONE (this->ReturnMemZone (), char, size));
		
		iter->next  = this->n_base;
		iter->check = check;
		iter->id    = (this->n_base) ? (this->n_base->id + 1) : (1);
		
		this->n_base = iter;
		strcpy (iter->name, name);		//	strcpy is safe, since we allocated enough room...
	}
	
	return iter->id;
}



/*
 *	Resolve the reference given an id. This will always
 *	succeed, since it is only called when it is possible
 *	to know about the id.
 */

void
OPStorage::ResolveRef (Card32 id, OPObjectOut obj)
{
	OP_METHOD (ResolveRef, 2);
	OP_ARG (1, Card32, id);
	OP_ARG (2, OPObjectOut, obj);
	
	RefBase* iter = this->r_base;
	
	while (iter) {
		if (iter->id == id) {
			obj = iter->ref;
			return;
		}
		iter = iter->next;
	}
	
	OP_ABORT ("Object ID not found in storage data base.");
}



/*
 *	Resolve the name given an id. This will always suc-
 *	ceed, since it is only called when it is possible to
 *	know about the id.
 */

void
OPStorage::ResolveName (Card32 id, const LatinChar*& name)
{
	OP_METHOD (ResolveName, 2);
	OP_ARG (1, Card32, id);
	OP_ARG (2, const LatinChar*&, name);
	
	NameBase* iter = this->n_base;
	
	while (iter) {
		if (iter->id == id) {
			name = iter->name;
			return;
		}
		iter = iter->next;
	}
	
	OP_ABORT ("Name ID not found in storage data base.");
}



/*
 *	We cannot possibly store a storage object, can we ?
 */

void
OPStorage::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store a storage object !");
}



/*
 *	We cannot possibly restore a storage object, can we ?
 */

void
OPStorage::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore a storage object !");
}



/*
 *	Define the current storage mode used to handle the
 *	object references. Also reset internal data bases.
 */

void
OPStorage::SetStorageMode (StorageMode mode)
{
	OP_METHOD (SetStorageMode, 1);
	OP_ARG (1, StorageMode, mode);
	
	if ( (mode == STORE_BIN_ONLY)
	  || (mode == STORE_SHALLOW)
	  || (mode == STORE_DEEP) ) {
		
		//	Change the storage mode. We won't accept any non-defined value here,
		//	or this might break everything later on !
		
		this->storage_mode = mode;
		
	} else {
		
		OP_ABORT ("Unknown storage mode specified.");
	}
	
	//	Reset the storage class to a clean state. No longer remember about the
	//	previous opertations.

	RefBase*   r_next = 0;
	NameBase*  n_base = 0;
	OPMemZone* zone   = this->ReturnMemZone ();
	
	while (this->r_base) {
		r_next = this->r_base->next;
		DELETE_FROM_ZONE (zone, this->r_base);
		this->r_base = r_next;
	}
	
	while (this->n_base) {
		n_base = this->n_base->next;
		OP_FREE_FROM_ZONE (zone, this->n_base);
		this->n_base = n_base;
	}

	this->r_base = 0;
	this->n_base = 0;
	
	this->num_pending_store = 0;
	this->num_recursion     = 0;
}



/*
 *	Return the storage mode.
 */

OPStorage::StorageMode
OPStorage::ReturnStorageMode () const
{
	OP_METHOD (ReturnStorageMode, 0);
	
	return this->storage_mode;
}

#define	STORE_8(ref,data)								\
	do {												\
		(&(ref))[0] = (Card8)(data >>  0);				\
	} while (0)

#define	STORE_16(ref,data)								\
	do {												\
		(&(ref))[0] = (Card8)(data >>  8);				\
		(&(ref))[1] = (Card8)(data >>  0);				\
	} while (0)

#define	STORE_32(ref,data)								\
	do {												\
		(&(ref))[0] = (Card8)(data >> 24);				\
		(&(ref))[1] = (Card8)(data >> 16);				\
		(&(ref))[2] = (Card8)(data >>  8);				\
		(&(ref))[3] = (Card8)(data >>  0);				\
	} while (0)

#define	STORE_64(ref,data)								\
	do {												\
		(&(ref))[0] = (Card8)(data >> 56);				\
		(&(ref))[1] = (Card8)(data >> 48);				\
		(&(ref))[2] = (Card8)(data >> 40);				\
		(&(ref))[3] = (Card8)(data >> 32);				\
		(&(ref))[4] = (Card8)(data >> 24);				\
		(&(ref))[5] = (Card8)(data >> 16);				\
		(&(ref))[6] = (Card8)(data >>  8);				\
		(&(ref))[7] = (Card8)(data >>  0);				\
	} while (0)




/*
 *	Store object. This will be handled according to the
 *	storage mode.
 */

void
OPStorage::StoreObject (OPObjectIn _object)
{
	OP_METHOD (StoreObject, 1);
	OP_ARG (1, OPObjectIn, _object);
	
	static OPAtomRef atom_phantom_object = "OPaC.phantom.object";
	
	OPObject*   object = _object.ReturnRawObject ();
	OPAtomRef   slot_name;
	OPObjectRef temp;
	
	Card8   ref[128];
	Address address = (Address)(object);
	Card32  addr32  = (Card32)(address);
	Card32  id      = 0;
	
	Bool is_new     = FALSE;
	Bool is_phantom = _object.IsPhantom ();
	
	//	Hmmm. I should have implemented this with 64-bit pointers in mind, but I
	//	must admit that this is rather difficult, since not every C compiler has
	//	64-bit integer support.
	
	OP_ASSERT (sizeof (void*) == 4);
	
	//	We handle special "zero" objects, since they can be saved quite easily.
	//	And while storing data in binary mode only, we may not store references
	//	to objects, so just store some null reference.
	
	if ( (address == 0)
	  || (this->storage_mode == STORE_BIN_ONLY) ) {
		
		STORE_8 (ref[0], TYP_NULL);
		
		this->StoreData (ref, 1);
		return;
	}
	
	if (this->FindNameFromSlot (object, slot_name)) {
		
		STORE_8 (ref[0], TYP_O_SLOT_DATA);
		
		this->StoreData (ref, 1);
		this->StoreAtom (slot_name);
		return;
	}
	
	//	Objects tagged with the "automatic_object" attribute must be handled in a
	//	special way : their attribute must be saved in place of themselves.
	
	if ( (this->storage_mode == STORE_DEEP)
	  && (object->GetAttribute (atom_phantom_object, temp)) ) {
		
		is_phantom = temp.IsPhantom ();
		object     = temp.ReturnRawObject ();
	}
	
	//	Phantom objects are stored in a special way, so we really save their
	//	description rather than the object they represent.
	
	if ( (this->storage_mode == STORE_DEEP)
	  && (is_phantom) ) {
		
		STORE_8 (ref[0], TYP_O_PHANTOM);
		
		this->StoreData (ref, 1);
		object->StoreInstance (this);
		return;
	}
	
	//	We handle instances of classe OPValue in a special way, if they wrap
	//	simple types of data. This reduces the space required for the storage.
	//	It will allow the automatic conversion between simple types and values
	//	without the user being aware of it.
	
	if ( (this->storage_mode == STORE_DEEP)
	  && (object->IsKindOf (OP_RUNTIME (OPValue)))
	  && (object->IsObjectConnected () == FALSE) ) {
		
		OPValueRef value = (OPValue*)object;
		
		switch (value->ReturnType ()) {
			case OP_TYPE_BOOL:	this->StoreBool (value->ReturnBool ());   return;
			case OP_TYPE_INT:	this->StoreInt  (value->ReturnInt32 ());  return;
			case OP_TYPE_CARD:	this->StoreCard (value->ReturnCard32 ()); return;
			case OP_TYPE_REAL:	this->StoreReal (value->ReturnReal ());   return;
		}
	}
	
	//	Atoms and strings will be stored immediately in a special way. This
	//	avoids complicated handling for objects such as phantoms, which rely
	//	on their strings to provide their information.
	
	if (object->ReturnRuntime () == OP_RUNTIME (OPAtom)) {
		
//		STORE_8 (ref[0], TYP_A_DECL_OBJ);
//		this->StoreData (ref, 1);
		
		this->StoreAtom (_object);
		return;
	}

	if ( (object->ReturnRuntime () == OP_RUNTIME (OPString))
	  && (object->more == 0) ) {
		
		this->StoreString (_object);
		return;
	}
	
	switch (this->storage_mode) {
		
		case STORE_SHALLOW:
			
			//	We just save the reference's pointer, and nothing more. Pointers
			//	are currently considered to be 32-bit, but support for 64-bit
			//	pointers can be easily added.

			//	***BEWARE*** If some garbage collector has to be used, we need
			//	to track down objects which have been stored in a "shallow"
			//	way.
			
			OP_POSSIBLE_PROBLEM ("Garbage collection not safe !");
			
			if (address == addr32) {
				STORE_8  (ref[0], TYP_ADDR32);
				STORE_32 (ref[1], addr32);
				this->StoreData (ref, 5);
			} else {
				STORE_8  (ref[0], TYP_ADDR64);
				STORE_64 (ref[1], address);
				this->StoreData (ref, 9);
			}
			break;
		
		case STORE_DEEP:
			
			//	We have to save the reference and also the object which is being
			//	referenced, but it is not done immediately, in order to decrease
			//	the recursion depth.
			
			id = this->LookForRef (object, is_new);
			
			if (is_new) {
				
				const OPRuntimeRef& r_time = object->ReturnRuntime ();
				const LatinChar*    name   = r_time->ReturnClassName ();
				
				Card32 name_id = this->LookForName (name, is_new);
				
				if (is_new) {
					
					Card32 length = strlen (name);
					OP_ASSERT (length < 100);
					
					//	The name was not in the data base, so we will have to store
					//	the name's ID, the object's ID and the name's string !
					
					STORE_8  (ref[0], TYP_O_DECL_NAME);
					STORE_8  (ref[1], length);
					
					this->StoreData (ref, 2);
					this->StoreID (name_id);
					this->StoreID (id);
					this->StoreData (name, length);
					
				} else {
					
					//	The name was already in the data base, so we will just have
					//	to store the name's ID and the object's ID.
					
					STORE_8  (ref[0], TYP_O_DECL_ID);
					
					this->StoreData (ref, 1);
					this->StoreID (name_id);
					this->StoreID (id);
				}
			
			} else {
				
				//	The object was already in the data base, so we will just have to
				//	store the object's ID and that will be all.
				
				STORE_8  (ref[0], TYP_O_INST_REF);
				
				this->StoreData (ref, 1);
				this->StoreID (id);
			}
			break;
			
		default:
			OP_ABORT ("Unhandled storage mode");
			break;
	}
	
	//	The objects get not saved in a recursive way, since this could use up to
	//	several hundred KB of stack, as has been experienced with OPaC 0.8...
	
	while ( (this->num_pending_store > 0)
		 && (this->num_recursion == 0) ) {
		
		RefBase* iter = this->r_base;

		while ( (iter != 0)
		     && (iter->stored == TRUE) ) {
			iter = iter->next;
		}
		
		//	There has to be an object or otherwise, I would not know why the
		//	number of pending stores wouldn't be 0.
		
		OP_ASSERT (iter != 0);
		
		this->num_recursion++;
		this->num_pending_store--;
		iter->stored = TRUE;
		
		//	Store the id of the object which will be saved...
		
		OPStorageRef self = this;
		
		STORE_8  (ref[0], TYP_O_DATA);
		
		this->StoreData (ref, 1);
		this->StoreID (iter->id);
		
		//	It would be catastrophic if there were an exception, since the recursion
		//	count would not be updated correctly...
		
		OPLisTreeRef unknown;
		
		iter->ref->GetUnknownData (unknown);
		
		if (unknown.IsValid ()) {
			unknown->Clone (this->storage_infos->more);
			this->storage_infos->more->Rewind ();
		}
		
		iter->ref->StoreInstance (self);
		
		this->storage_infos->more = 0;
		
		OP_POSSIBLE_PROBLEM ("Exception will be fatal here !");
		
		this->num_recursion--;
	}
}



/*
 *	Store array. This will be handled according to the
 *	storage mode.
 */

void
OPStorage::StoreArray (OPArrayIn array)
{
	OP_METHOD (StoreArray, 1);
	OP_ARG (1, OPArrayIn, array);
	
	if (array.IsInvalid ()) {
		Card8 ref[1] = { TYP_NULL };
		this->StoreData (ref, 1);
		return;
	}
	
	Count  index     = 0;
	Card32 num_elem  = array->TotalCount ();
	Card8  header[2] = { TYP_ARRAY, 0 };
	
	this->StoreData (header, 2);
	this->StoreCard (num_elem);
	
	while (num_elem--) {
		OPObjectRef& obj = array[index++];
		this->StoreObject (obj);
	}
}



/*
 *	Restore object. This will be handled according to the
 *	storage mode.
 */

void
OPStorage::RestoreObject (OPObjectOut object, OPRuntimeIn runtime)
{
	OP_METHOD (RestoreObject, 2);
	OP_ARG (1, OPObjectOut, object);
	OP_ARG (2, OPRuntimeIn, runtime);
	
	static OPAtomRef slot_name;
	
	Card8   ref[128];
	Address address = 0;
	Card32  length  = 0;
	Card32  id      = 0;
	Card32  name_id = 0;
	Card32  new_id  = 0;
	
	Bool is_new = FALSE;
	
	const LatinChar* name  = 0;
	OPStorageRef     self  = this;
	OPValueRef       data  = OPValueNull;
	OPArrayRef       array = OPArrayNull;
	
	this->PopAndRestoreData (ref+0, 1);
		
	switch (ref[0]) {
		
		case TYP_SYNC_MARK:
		case TYP_SYNC_END:
			this->PushBackData (ref+0, 1);
			object = 0;
			break;
		
		case TYP_NULL:
			object = 0;
			break;
		
		//	Simple types map to an OPValue instance if an object was expected in
		//	place of it. This replacing is also done automatically when storing
		//	objects of type OPValue.
		
		case TYP_INT32:
		case TYP_INT32_8:
		case TYP_INT32_16:
		case TYP_INT32_ZERO:
			this->PushBackData (ref+0, 1);
			data   = (Int32) this->RestoreInt ();
			object = data;
			break;
		
		case TYP_CARD32:
		case TYP_CARD32_8:
		case TYP_CARD32_16:
		case TYP_CARD32_ZERO:
			this->PushBackData (ref+0, 1);
			data   = (Card32) this->RestoreCard ();
			object = data;
			break;
			
		case TYP_INT64:
			this->PushBackData (ref+0, 1);
			data   = this->RestoreInt ();
			object = data;
			break;
			
		case TYP_CARD64:
			this->PushBackData (ref+0, 1);
			data   = this->RestoreCard ();
			object = data;
			break;
			
		case TYP_STRING:
			object = OP_NEW (OPString);
			object->RestoreInstance (this);
			break;
		
		case TYP_ARRAY:
			this->PushBackData (ref+0, 1);
			this->RestoreArray (array, runtime);
			object = array;
			break;
		
		case TYP_REAL:
		case TYP_REAL_INT:
		case TYP_REAL_ZERO:
		case TYP_REAL_PONE:
		case TYP_REAL_MONE:
			this->PushBackData (ref+0, 1);
			data = this->RestoreReal ();
			object = data;
			break;
		
		case TYP_BOOL_FALSE:
			data = FALSE;
			object = data;
			break;
		
		case TYP_BOOL_TRUE:
			data = TRUE;
			object = data;
			break;
		
		//	Complex data types will be restored here :
		
		case TYP_ADDR32:
//		case TYP_ADDR32_8:
//		case TYP_ADDR32_16:
			this->PopAndRestoreData (ref+1, 4);
			
			address = ((Address)(ref[1]) << 24)
			        + ((Address)(ref[2]) << 16)
			        + ((Address)(ref[3]) <<  8)
			        + ((Address)(ref[4]) <<  0);
				
			object = (OPObject*)(address);
			break;
		
		case TYP_ADDR64:
			this->PopAndRestoreData (ref+1, 8);
			
			address = ((Address)(ref[1]) << 56)
			        + ((Address)(ref[2]) << 48)
			        + ((Address)(ref[3]) << 40)
			        + ((Address)(ref[4]) << 32)
			        + ((Address)(ref[5]) << 24)
			        + ((Address)(ref[6]) << 16)
			        + ((Address)(ref[7]) <<  8)
			        + ((Address)(ref[8]) <<  0);
			
			object = (OPObject*)(address);
			break;
		
		case TYP_O_DECL_NAME:
			
			//	The object is defined by its ID and the name of the runtime class
			//	it belongs to. The name is given by a string.
			
			this->PopAndRestoreData (ref+1, 1);
			
			length  =  (Card32)(ref[1]);
			name_id = this->RestoreID ();
			id      = this->RestoreID ();
			
			this->PopAndRestoreData (ref, length);
			ref[length] = 0;
			name = (LatinChar*)(ref);
			
			//	Check that the name we got is really unique and that its ID is
			//	correct with respect to the data base. By the way, store the name
			//	for further references to it.
			
			new_id = this->LookForName (name, is_new);
			
			OP_ASSERT (new_id == name_id);
			OP_ASSERT (is_new);
			goto new_instance;
		
		case TYP_O_DECL_ID:
			
			//	The object is defined by its ID and by the ID of the name of the
			//	runtime class it belongs to.
			
			name_id = this->RestoreID ();
			id      = this->RestoreID ();
			
			this->ResolveName (name_id, name);
			
			//	Instanciate the object before it is being added to the data base
			//	of known objects.
			
new_instance:
			object = OPClassMan::New (name, this->ReturnMemZone (), __LINE__, __OPaC_src_file);
			
			//	Store the object into the object data base and verify that it is
			//	unique and that the ID is correct with respect to the data base.
			
			new_id = this->LookForRef (object, is_new);
			
			OP_ASSERT (new_id == id);
			OP_ASSERT (is_new);
			break;
		
		case TYP_O_INST_REF:
			
			//	The object is the same as a previously saved one. Just find out
			//	about it.
			
			id = this->RestoreID ();
			this->ResolveRef (id, object);
			break;			
		
		case TYP_O_DATA:
			
			//	The object is the same as a previously saved one. Just find out
			//	about it; we then restore its data from the stream.
			
			id = this->RestoreID ();
			this->ResolveRef (id, object);
			object->RestoreInstance (self);
			break;
		
		case TYP_O_SLOT_DATA:
			
			//	The object is stored in the data slot under a known name. Just
			//	find out which name it is and what object is referenced by that
			//	one.
			
			this->RestoreAtom (slot_name);
			this->FindObjectFromSlot (slot_name, object);
			break;
		
		case TYP_O_PHANTOM:
			object = OP_NEW (OPPhantom);
			object.ReturnRawObject ()->RestoreInstance (this);
			break;
		
		case TYP_A_DECL_NAME:
		case TYP_A_DECL_ID:
			
			//	Restore an atom and return it as an object. There is nothing
			//	wrong with it.
			
			this->PushBackData (ref+0, 1);
			this->RestoreAtom (slot_name);
			object = slot_name;
			break;
		
		case TYP_A_DECL_OBJ:
			
			//	The object is an atom... this should happen only if an atom has
			//	been cast to an object before saving.
			
			this->RestoreAtom (slot_name);
			object = slot_name;
			break;
		
		default:
			OP_ABORT ("Wrong data type, cannot restore object");
			break;
	}
	
	//	The objects get not restored in a recursive way, since this could use up
	//	to several hundred KB of stack, as has been experienced with OPaC 0.8...
	
	while ( (this->num_pending_store > 0)
	     && (this->num_recursion == 0) ) {
		
		RefBase* iter = this->r_base;
		this->storage_infos->more = OP_NEW (OPLisTree);
		
		while ( (iter != 0)
		     && (iter->stored == TRUE) ) {
			iter = iter->next;
		}
		
		//	There has to be an object or otherwise, I would not know why the
		//	number of pending (re)stores wouldn't be 0.
		
		OP_ASSERT (iter != 0);
		
		this->num_recursion++;
		this->num_pending_store--;
		iter->stored = TRUE;
		
		//	Restore the id of the object which had been saved...
		
		this->PopAndRestoreData (ref, 1);
		
		id = this->RestoreID ();
		
		OP_ASSERT (ref[0] == TYP_O_DATA);
		OP_ASSERT (id == iter->id);
		
		OPStorageRef self = this;
		
		//	It would be catastrophic if there is an exception, since the recursion
		//	count would not be updated correctly...
		
		iter->ref->RestoreInstance (self);
		iter->ref->SetUnknownData (this->storage_infos->more);
		
		OP_POSSIBLE_PROBLEM ("Exception will be fatal here !");
		
		this->num_recursion--;
	}
	
	//	Check if the object we have just restored matches the proper runtime type
	//	specified by the caller.
	
	Bool is_text = runtime.IsValid () && runtime->IsKindOf (OP_RUNTIME (OPText));
	Bool is_val  = runtime.IsValid () && runtime->IsKindOf (OP_RUNTIME (OPValue));
	
	while ( (object.IsValid ())
		 && (runtime.IsValid ()) ) {
		
		if (object.IsPhantom ()) {
			return;
		}
		
		if (object->IsKindOf (runtime)) {
			return;
		}
		
		if ( (is_text)
		  && (object->IsKindOf (OP_RUNTIME (OPText))) ) {
			if (runtime->IsKindOf (OP_RUNTIME (OPAtom))) {
				OPTextRef o_text = object;
				OPAtomRef n_text = o_text->ReturnString ();
				object = n_text;
			} else {
				OPTextRef o_text = object;
				OPTextRef n_text = OP_DYNAMIC_NEW (runtime, OPText);
				n_text->SetString (o_text->ReturnString ());
				object = n_text;
			}
			return;
		}
		
		if ( (is_val)
		  && (object->IsKindOf (OP_RUNTIME (OPText))) ) {
			OPValueRef value = object;
			object = value;
			return;
		}
		
		if (object->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef value = object;
			object = value->ReturnObject ();
		} else if (runtime->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef value = OP_NEW (OPValue);
			value->SetObject (object);
			object = value;
			return;
		} else {
			OP_ABORT ("Restoring object of bad type");
		}
	}
}



/*
 *	Restore array. This will be handled according to the
 *	storage mode.
 */

void
OPStorage::RestoreArray (OPArrayOut array, OPRuntimeIn runtime)
{
	OP_METHOD (RestoreArray, 2);
	OP_ARG (1, OPArrayOut, array);
	OP_ARG (2, OPRuntimeIn, runtime);
	
	//	The "runtime" information is provided in order to be able to insure that
	//	the objects put into the array are compatible. Is is used by the object
	//	restoration method.
	
	Count  index     = 0;
	Card32 num_elem  = 0;
	Card8  header[2] = { 0, 0 };
	
	this->PopAndRestoreData (header, 2);
	
	if (header[0] == TYP_NULL) {
		this->PushBackData (header+1, 1);
		array = OPArrayNull;
		return;
	}
	
	if ( (header[0] == TYP_SYNC_MARK)
	  || (header[0] == TYP_SYNC_END) ) {
		this->PushBackData (header, 2);
		array = OP_NEW (OPArray);
		array->SetSize (0);
		return;
	}
	
	OP_ASSERT (header[0] == TYP_ARRAY);
	
	num_elem = (Card32) this->RestoreCard ();
	
	array = OP_NEW (OPArray);
	array->SetSize (num_elem);
	
	while (num_elem--) {
		OPObjectRef& obj = array[index++];
		this->RestoreObject (obj, OP_RUNTIME (OPObject));
	}
}



/*
 *	Store a Boolean.
 */

void
OPStorage::StoreBool (Bool x)
{
	OP_METHOD (StoreBool, 1);
	OP_ARG (1, Bool, x);
	
	Card8 val[1];
	
	val[0] = (x) ? TYP_BOOL_TRUE : TYP_BOOL_FALSE;
	
	this->StoreData (val, 1);
}



/*
 *	Store a 32-bit integer. The integer is always being
 *	recorded as a Big Endian number, whatever the host's
 *	numbering scheme is.
 */

void
OPStorage::StoreInt (Int64 x64)
{
	OP_METHOD (StoreInt, 1);
	OP_ARG (1, Int64, x64);
	
	Int32 x = (Int32)(x64);
	
	Card8  val[5];
	Card8* ptr  = val;
	Count  size = 5;
	
	if (x == 0) {
		val[0] = TYP_INT32_ZERO;
		this->StoreData (ptr, 1);
		return;
	}
	
	val[0] = TYP_INT32;
	val[1] = (Card8)(x >> 24);
	val[2] = (Card8)(x >> 16);
	val[3] = (Card8)(x >>  8);
	val[4] = (Card8)(x >>  0);
	
	//	Optimize the case when the number is small and the most significant
	//	bits are all set to zero : save the number either as an 8 or 16-bit
	//	quantities. Handle also the case of small negative numbers where the
	//	bits are all set to one.
	
	if ( ((x & 0xFFFFFF00) == 0x00000000)
	  || ((x & 0xFFFFFF00) == 0xFFFFFF00) ) {
		ptr    = val+3;
		ptr[0] = TYP_INT32_8;
		size   = 2;
	} else if ( ((x & 0xFFFF0000) == 0x00000000)
			 || ((x & 0xFFFF0000) == 0xFFFF0000) ) {
		ptr    = val+2;
		ptr[0] = TYP_INT32_16;
		size   = 3;
	}
	
	this->StoreData (ptr, size);
}



/*
 *	Store a 32-bit cardinal (unsigned int, if you like).
 */

void
OPStorage::StoreCard (Card64 x64)
{
	OP_METHOD (StoreCard, 1);
	OP_ARG (1, Card64, x64);
	
	Card32 x = (Card32)(x64);
	
	//	NIY : storing 64-bit values
	
	Card8  val[5];
	Card8* ptr  = val;
	Count  size = 5;
	
	if (x == 0) {
		val[0] = TYP_CARD32_ZERO;
		this->StoreData (ptr, 1);
		return;
	}
	
	val[0] = TYP_CARD32;
	val[1] = (Card8)(x >> 24);
	val[2] = (Card8)(x >> 16);
	val[3] = (Card8)(x >>  8);
	val[4] = (Card8)(x >>  0);
	
	//	Optimize the case when the number is small and the most significant
	//	bits are all set to zero : save the number either as an 8 or 16-bit
	//	quantities.
	
	if ((x & 0xFFFFFF00) == 0x00000000) {
		ptr    = val+3;
		ptr[0] = TYP_CARD32_8;
		size   = 2;
	} else if ((x & 0xFFFF0000) == 0x00000000) {
		ptr    = val+2;
		ptr[0] = TYP_CARD32_16;
		size   = 3;
	}
	
	this->StoreData (ptr, size);
}



/*
 *	Store a real value (double precision floating point,
 *	if you like).
 */

void
OPStorage::StoreReal (Real x)
{
	OP_METHOD (StoreReal, 1);
	OP_ARG (1, Real, x);
	
	if (x == 0.0) {
		Card8 header[1] = { TYP_REAL_ZERO };
		this->StoreData (header, 1);
		return;
	}
	
	if (x == 1.0) {
		Card8 header[1] = { TYP_REAL_PONE };
		this->StoreData (header, 1);
		return;
	}
	
	if (x == -1.0) {
		Card8 header[1] = { TYP_REAL_MONE };
		this->StoreData (header, 1);
		return;
	}
	
	Real rest = fmod (x, 1.0);
	
	if ( (rest == 0.0)
	  && (fabs (x) < (Int32)(0x7FFFFFFF)) ) {
		
		//	The real number can be represented by a (small) integer value, since
		//	it has no fractional part.
		
		Int32 i = (Int32)(x);
		Card8 header[1] = { TYP_REAL_INT };
		this->StoreData (header, 1);
		this->StoreInt (i);
		return;
	}
	
	Card8  header[2]   = { TYP_REAL, 0 };
	Card32 exponent    = 0;
	Card32 mantissa[2] = { 0, 0 };
	
	OPaC_ExportDouble (x, exponent, mantissa);
	
	this->StoreData (header, 2);
	this->StoreCard (exponent);
	this->StoreCard (mantissa[0]);
	this->StoreCard (mantissa[1]);
}



/*
 *	Store an atom. This won't be handled in the same way
 *	as an other object would, since atoms are special
 *	entities.
 */

void
OPStorage::StoreAtom (OPAtomIn atom)
{
	OP_METHOD (StoreAtom, 1);
	OP_ARG (1, OPAtomIn, atom);
	
	//	This is a special case where the super-class won't be called ! The atom
	//	is not considered to be a real object.
	
	if (atom.IsValid () == FALSE) {
		Card8 header[2] = { TYP_A_DECL_NAME, 0 };
		this->StoreData (header, 2);
		return;
	}

	LatinChar* text = atom->text;
	
	Bool   is_new  = FALSE;
	Card32 name_id = this->LookForName (text, is_new);

	if (is_new == FALSE) {
		
		//	The name is already known in the "name database", which is shared with
		//	the class names. Just store the ID of the atom. This will do...
		
		Card8 ref[1] = { TYP_A_DECL_ID };
		
		this->StoreData (ref, 1);
		this->StoreID (name_id);
		return;
	}
	
	//	The name was not know in the "name database" (now, it is). Save the name
	//	of the atom.
	
	while (*text++) {
		
		//	Just skip to end of text... so we know how many characters there are
		//	in the atom's name.
	}
	
	Card32 length    = (Card32)(text - atom->text);
	Card8  header[4] = { TYP_A_DECL_NAME, (Card8)(length) };
	Count  size = 2;
	
	if (length >= 240) {
		header[1] = 0xff;
		header[2] = (Card8)(length >> 8);
		header[3] = (Card8)(length);
		size = 4;
	}
	
	this->StoreData (header, size);
	this->StoreData (atom->text, length);
}



/*
 *	Store an atom array. This won't be handled in the same
 *	way as an other array would, since atoms are special
 *	entities.
 */

void
OPStorage::StoreAtomArray (OPAtomArrayIn array)
{
	OP_METHOD (StoreAtomArray, 1);
	OP_ARG (1, OPAtomArrayIn, array);
	
	Count  index     = 0;
	Card32 num_elem  = array->TotalCount ();
	Card8  header[2] = { TYP_ARRAY, 0 };
	
	this->StoreData (header, 2);
	this->StoreCard (num_elem);
	
	while (num_elem--) {
		OPAtomRef& atom = array[index++];
		this->StoreAtom (atom);
	}
}


void
OPStorage::StoreString (OPStringIn text)
{
	OP_METHOD (StoreString, 1);
	OP_ARG (1, OPStringIn, text);
	
	OPObjectRef object = text;
	
	Card8 header[1] = { TYP_STRING };
	
	this->StoreData (header, 1);
	object->StoreInstance (this);
}

void
OPStorage::StorePhantom (OPPhantomIn phantom)
{
	OP_METHOD (StorePhantom, 1);
	OP_ARG (1, OPPhantomIn, phantom);
}


/*
 *	Store a chunk of bytes as binary data. This will be
 *	prefixed by the size of the data, so no overrun will
 *	occurr when restoring the data.
 */

void
OPStorage::StoreBinData (const void* data, Card32 size)
{
	OP_METHOD (StoreBinData, 2);
	OP_ARG (1, const void*, data);
	OP_ARG (2, Card32, size);
	
	Card8 header[5] = { 0, 0, 0, 0, 0 };
	
	if (size < 0x00000100) {
		header[0] = TYP_BIN_8;
		header[1] = (Card8)(size);
		this->StoreData (header, 2);
		this->StoreData (data, size);
	} else if (size < 0x00010000) {
		header[0] = TYP_BIN_16;
		header[1] = (Card8)(size >> 8);
		header[2] = (Card8)(size >> 0);
		this->StoreData (header, 3);
		this->StoreData (data, size);
	} else {
		header[0] = TYP_BIN_32;
		header[1] = (Card8)(size >> 24);
		header[2] = (Card8)(size >> 16);
		header[3] = (Card8)(size >>  8);
		header[4] = (Card8)(size >>  0);
		this->StoreData (header, 5);
		this->StoreData (data, size);
	}
}


/*
 *	Restore a boolean.
 */

Bool
OPStorage::RestoreBool ()
{
	OP_METHOD (RestoreBool, 0);
	
	Card8  val[1];
	
	this->PopAndRestoreData (val, 1);
	
	if ( (val[0] == TYP_SYNC_MARK)
	  || (val[0] == TYP_SYNC_END) ) {
		this->PushBackData (val, 1);
		return FALSE;
	}
	
	if ( (val[0] == TYP_BOOL_TRUE)
	  || (val[0] == TYP_BOOL_FALSE) ) {
		return val[0] == TYP_BOOL_TRUE;
	}
	
	//	Seems we have got a value...
	
	OPValueRef  value;
	OPObjectRef object;
	
	this->PushBackData (val, 1);
	this->RestoreObject (object, OP_RUNTIME (OPValue));
	
	OP_ASSERT (object.IsValid ());
	
	value = object;
	
	return value->ReturnBool ();
}



/*
 *	Restore a 32-bit integer.
 */

Int64
OPStorage::RestoreInt ()
{
	OP_METHOD (RestoreInt, 0);
	
	Card8  type;
	Card8  val[8] = { 0 };
	
	//	Integer values are saved as "Iabcd" where abcd are the high to low-order
	//	bytes of the 32-bit integer. The "I" stands for 32-bit "Int".
	
	this->PopAndRestoreData (&type, 1);
	
	if ( (type == TYP_SYNC_MARK)
	  || (type == TYP_SYNC_END) ) {
		this->PushBackData (&type, 1);
		return 0;
	}
	
	if (type == TYP_INT64) {
		this->PopAndRestoreData (val+0, 8);
	} else if (type == TYP_INT32) {
		this->PopAndRestoreData (val+4, 4);
	} else if (type == TYP_INT32_16) {
		this->PopAndRestoreData (val+6, 2);
	} else if (type == TYP_INT32_8) {
		this->PopAndRestoreData (val+7, 1);
	} else if (type == TYP_INT32_ZERO) {
		return 0;
	} else if (type != TYP_INT32) {
		
		//	Seems we have got a value...
		
		OPValueRef  value;
		OPObjectRef object;
		
		this->PushBackData (&type, 1);
		this->RestoreObject (object, OP_RUNTIME (OPValue));
		
		OP_ASSERT (object.IsValid ());
		
		value = object;
		
		return value->ReturnInt64 ();
	}
	
	return (Int64)( ((Card64)(val[0]) << 56) +
					((Card64)(val[1]) << 48) +
					((Card64)(val[2]) << 40) +
					((Card64)(val[3]) << 32) +
					((Card64)(val[4]) << 24) +
					((Card64)(val[5]) << 16) +
					((Card64)(val[6]) <<  8) +
					((Card64)(val[7]) <<  0) );
}



/*
 *	Restore a 32-bit cardinal (unsigned int, if you like).
 */

Card64
OPStorage::RestoreCard ()
{
	OP_METHOD (RestoreCard, 0);
	
	Card8  type;
	Card8  val[8] = { 0 };
	
	//	Cardinal values are saved as "Cabcd" where abcd are the high to low-order
	//	bytes of the 32-bit integer. The "C" stands for 32-bit "Card".
	
	this->PopAndRestoreData (&type, 1);
	
	if ( (type == TYP_SYNC_MARK)
	  || (type == TYP_SYNC_END) ) {
		this->PushBackData (&type, 1);
		return 0;
	}
	
	if (type == TYP_CARD64) {
		this->PopAndRestoreData (val+0, 8);
	} else if (type == TYP_CARD32) {
		this->PopAndRestoreData (val+4, 4);
	} else if (type == TYP_CARD32_16) {
		this->PopAndRestoreData (val+6, 2);
	} else if (type == TYP_CARD32_8) {
		this->PopAndRestoreData (val+7, 1);
	} else if (type == TYP_CARD32_ZERO) {
		return 0;
	} else if (type != TYP_CARD32) {
		
		//	Seems we have got a value...
		
		OPValueRef  value;
		OPObjectRef object;
		
		this->PushBackData (&type, 1);
		this->RestoreObject (object, OP_RUNTIME (OPValue));
		
		OP_ASSERT (object.IsValid ());
		
		value = object;
		
		return value->ReturnCard64 ();
	}
	
	return ( ((Card64)(val[0]) << 56) +
			 ((Card64)(val[1]) << 48) +
			 ((Card64)(val[2]) << 40) +
			 ((Card64)(val[3]) << 32) +
			 ((Card64)(val[4]) << 24) +
			 ((Card64)(val[5]) << 16) +
			 ((Card64)(val[6]) <<  8) +
			 ((Card64)(val[7]) <<  0) );
}



/*
 *	Restore a real value (double precision floating point,
 *	if you like).
 */

Real
OPStorage::RestoreReal ()
{
	OP_METHOD (RestoreReal, 0);
	
	Card8  header[2]   = { 0, 0 };
	Real   val         = 0.0;
	Card32 exponent    = 0;
	Card32 mantissa[2] = { 0, 0 };
	
	this->PopAndRestoreData (header, 2);
	
	if ( (header[0] == TYP_SYNC_MARK)
	  || (header[0] == TYP_SYNC_END) ) {
		this->PushBackData (header, 2);
		return 0.0;
	}
	
	if (header[0] != TYP_REAL) {
		
		switch (header[0]) {
			case TYP_REAL_INT:
				this->PushBackData (header+1, 1);
				return (Real)this->RestoreInt ();
			case TYP_REAL_ZERO:
				this->PushBackData (header+1, 1);
				return 0.0;
			case TYP_REAL_PONE:
				this->PushBackData (header+1, 1);
				return 1.0;
			case TYP_REAL_MONE:
				this->PushBackData (header+1, 1);
				return -1.0;
		}
		
		//	Seems we have got a value...
		
		OPValueRef  value;
		OPObjectRef object;
		
		this->PushBackData (header, 2);
		this->RestoreObject (object, OP_RUNTIME (OPValue));
		
		OP_ASSERT (object.IsValid ());
		
		value = object;
		
		return value->ReturnReal ();
	}
	
	exponent    = (Card32) this->RestoreCard ();
	mantissa[0] = (Card32) this->RestoreCard ();
	mantissa[1] = (Card32) this->RestoreCard ();
	
	OPaC_ImportDouble (val, exponent, mantissa);
	
	return val;
}



/*
 *	Restore an atom. This won't be handled in the same
 *	way an other object would, since atoms are special
 *	entities.
 */

void
OPStorage::RestoreAtom (OPAtomOut atom)
{
	OP_METHOD (RestoreAtom, 1);
	OP_ARG (1, OPAtomOut, atom);
	
	Card8      header[4];
	LatinChar  buffer[256];
	LatinChar* ptr = buffer;
	
	this->PopAndRestoreData (header, 1);
	
	if ( (header[0] == TYP_SYNC_MARK)
	  || (header[0] == TYP_SYNC_END) ) {
		this->PushBackData (header, 1);
		atom = "";
		return;
	}
	
	if (header[0] == TYP_A_DECL_ID) {
		
		//	The atom had previously been restored. Just look into the name
		//	database.
		
		const LatinChar* name = 0;
		Card32 name_id = this->RestoreID ();
		
		this->ResolveName (name_id, name);
		
		atom = name;
		return;
	}
	
	if (header[0] != TYP_A_DECL_NAME) {
		
		//	Seems we haven't got an atomic atom at this place. Try to restore
		//	the atom as an object...
		
		this->PushBackData (header, 1);
		this->RestoreObject (atom, OP_RUNTIME (OPAtom));
		
		return;
	}
	
	this->PopAndRestoreData (header+1, 1);
	
	if (header[1] == 0) {
		atom = OPAtomNull;
		return;
	}
	
	Card32 size = (Card32)(header[1]);
	
	//	Handle the special case where the atom name is longer than 239 characters.
	//	This will happen with the type descriptions, which are stored like atoms.
	
	if (size == 0xff) {
		this->PopAndRestoreData (header+2, 2);
		size = (header[2] << 8) + header[3];
		ptr  = (LatinChar*) alloca (size);
	}
	
	//	Restore the text which corresponds to the atom. This will be used to
	//	create/find the adequate atom :
	
	this->PopAndRestoreData (ptr, size);
	atom = ptr;
		
	//	Insert the name into the name database.
	
	Bool   is_new  = FALSE;
	Card32 new_id  = 0;
	
	new_id = this->LookForName (ptr, is_new);
	
	OP_ASSERT (is_new);
}



/*
 *	Restore atom array. This is a special case for the
 *	atom array.
 */

void
OPStorage::RestoreAtomArray (OPAtomArrayOut array)
{
	OP_METHOD (RestoreAtomArray, 1);
	OP_ARG (1, OPAtomArrayOut, array);
	
	Count  index     = 0;
	Card32 num_elem  = 0;
	Card8  header[2] = { 0, 0 };
	
	this->PopAndRestoreData (header, 2);
	
	if ( (header[0] == TYP_SYNC_MARK)
	  || (header[0] == TYP_SYNC_END) ) {
		this->PushBackData (header, 2);
		array = OP_NEW (OPArray);
		array->SetSize (0);
		return;
	}
	
	OP_ASSERT (header[0] == TYP_ARRAY);
	
	num_elem = (Card32) this->RestoreCard ();
	
	array = OP_NEW (OPArray);
	array->SetSize (num_elem);
	
	while (num_elem--) {
		OPAtomRef& atom = array[index++];
		this->RestoreAtom (atom);
	}
}


void
OPStorage::RestoreString (OPStringOut text)
{
	OP_METHOD (RestoreString, 1);
	OP_ARG (1, OPStringOut, text);
}

void
OPStorage::RestorePhantom (OPPhantomOut phantom)
{
	OP_METHOD (RestorePhantom, 1);
	OP_ARG (1, OPPhantomOut, phantom);
}


/*
 *	Restore a chunk of bytes as binary data. The size which
 *	is requested must match the one specified when storing.
 */

void
OPStorage::RestoreBinData (void* data, Card32 size)
{
	OP_METHOD (RestoreBinData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);
	
again:
	Card8  header[5] = { 0, 0, 0, 0, 0 };
	Card32 length    = 0; 
	
	this->PopAndRestoreData (header, 1);
	
	switch (header[0]) {
		case TYP_SYNC_MARK:
		case TYP_SYNC_END:
			this->PushBackData (header, 1);
			return;
		case TYP_BIN_8:
			this->PopAndRestoreData (header+4, 1);
			break;
		case TYP_BIN_16:
			this->PopAndRestoreData (header+3, 2);
			break;
		case TYP_BIN_32:
			this->PopAndRestoreData (header+1, 4);
			break;
		
		default:
			//	There is no data chunk available. Perhaps we didn't really expect
			//	data, so this could just be ignored.
			
			if (size == 0) {
				this->PushBackData (header, 1);
				return;
			}
			
			OP_ABORT ("Bad data type");
	}
	
	length = ( (Card32)(header[1]) << 24)
		   + ( (Card32)(header[2]) << 16)
		   + ( (Card32)(header[3]) <<  8)
		   + ( (Card32)(header[4]) <<  0);
	
	if ( (data == 0)
	  && (size == 0) ) {
		
		//	Discard the binary data if the caller did not give us a
		//	pointer to a valid buffer (RestoreToSyncMark uses this
		//	facility to skip a chunk of binary data).
		
		if (this->push_size) {
			if (length > this->push_size) {
				length -= this->push_size;
				this->push_size = 0;
			} else {
				this->push_size -= length;
				length = 0;
			}
		}
		while (length) {
			Card8 buffer[1000];
			size = (length > 1000) ? 1000 : length;
			this->RestoreData (buffer, size);
			length -= size;
		}
		
		return;
	}
	
	if (length > size) {
		
		//	There is more data available than the user is willing to
		//	restore. Restore what is requested and pretend there is
		//	a second chunk of data following immediately.
		
		this->PopAndRestoreData (data, size);
		
		length -= size;
		
		header[0] = TYP_BIN_32;
		header[1] = (Card8)(length >> 24);
		header[2] = (Card8)(length >> 16);
		header[3] = (Card8)(length >>  8);
		header[4] = (Card8)(length >>  0);
		
		this->PushBackData (header, 5);
		return;
	}
	
	if (length < size) {
		
		//	The chunk does not contain enough data for this restore
		//	operation. Restore what is available, then try again with
		//	the following chunk.
		
		this->PopAndRestoreData (data, length);
		
		size -= length;
		data  = (void*)((Address)(data) + length);
		
		goto again;
	}
	
	this->PopAndRestoreData (data, length);
}


/*
 *	Awake the miscellanous objects which have been read
 *	from the storage. They will get their awakening method
 *	called.
 */

void
OPStorage::FinallyAwake ()
{
	OP_METHOD (FinallyAwake, 0);
	
	RefBase* iter = this->r_base;
	
	while (iter) {
		iter->ref->Awake ();
		iter = iter->next;
	}
}

void
OPStorage::FinallyAwakeFromApp (OPAppIn app)
{
	OP_METHOD (FinallyAwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);
	
	RefBase* iter = this->r_base;
	
	while (iter) {
		iter->ref->AwakeFromApp (app);
		iter = iter->next;
	}
}

void
OPStorage::FinallyAwakeFromObject (OPObjectIn object)
{
	OP_METHOD (FinallyAwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	RefBase* iter = this->r_base;
	
	while (iter) {
		iter->ref->AwakeFromObject (object);
		iter = iter->next;
	}
}

/****************************************************************/

OPStorageSync::OPStorageSync (const LatinChar* name) : internal (name), is_name_ok (FALSE)
{
}

OPStorageSync::~OPStorageSync ()
{
}

void
OPStorageSync::UpdateInternalName () const
{
	OPAtomRef   name = this->internal;
	this->name       = name;
	this->is_name_ok = TRUE;
}

const OPStorageSync OPStorageSync::mark ("mark");
const OPStorageSync OPStorageSync::begin ("begin");
const OPStorageSync OPStorageSync::repeat ("repeat");
const OPStorageSync OPStorageSync::end ("end");
const OPStorageSync OPStorageSync::end_repeat ("end_repeat");
const OPStorageSync OPStorageSync::type ("type");

OPStorageIn
OPStorageRef::operator << (OPStorageSyncIn x) const
{
	if (&x == &OPStorageSync::mark) {
		(*this)->StoreSyncMark ();
		return (*this);
	}
	if ( (&x == &OPStorageSync::begin)
	  || (&x == &OPStorageSync::repeat) ) {
		(*this)->StoreBlockBegin ();
		return (*this);
	}
	if ( (&x == &OPStorageSync::end)
	  || (&x == &OPStorageSync::end_repeat) ) {
		(*this)->StoreBlockEnd ();
		return (*this);
	}
	
	Card8 header[1] = { TYP_TYPE_INFO };
	
	if (! x.is_name_ok) {
		x.UpdateInternalName ();
	}
	
	(*this)->StoreData (header, 1);
	(*this)->StoreAtom (x.name);
	
	return (*this);
}

OPStorageIn
OPStorageRef::operator >> (OPStorageSyncIn x) const
{
	if (&x == &OPStorageSync::mark) {
		(*this)->RestoreToSyncMark ();
		return (*this);
	}
	if ( (&x == &OPStorageSync::begin)
	  || (&x == &OPStorageSync::repeat) ) {
		(*this)->RestoreBlockBegin ();
		return (*this);
	}
	if ( (&x == &OPStorageSync::end)
	  || (&x == &OPStorageSync::end_repeat) ) {
		(*this)->RestoreToBlockEnd ();
		return (*this);
	}
	
	if (&x == &OPStorageSync::type) {
		OPAtomRef name;
		Card8 val[1];
		(*this)->PopAndRestoreData (val, 1);
		(*this)->RestoreAtom (name);
		return (*this);
	}
	
	return (*this);
}

/****************************************************************/

OPStorageIn
OPStorageRef::operator << (OPColorPixelIn x) const
{
	x.StoreInstance (*this);
	return (*this);
}

OPStorageIn
OPStorageRef::operator >> (OPColorPixelOut x) const
{
	x.RestoreInstance (*this);
	return (*this);
}

/****************************************************************/

/*
 *	Store a synchronisation mark into the output stream. This
 *	is used to mark the end of the fields stored by an object.
 */

void
OPStorage::StoreSyncMark ()
{
	OP_METHOD (StoreSyncMark, 0);
	
	if ( (this->storage_infos->more.IsValid ())
	  && (this->storage_infos->skip == 0) ) {
		
		Card8 val[1] = { 0 };
		Count skip_end  = 0;
		Bool  more_data = TRUE;
		
		do {
			OPObjectRef object;
			OPLisTree::Type type = this->storage_infos->more->PopHead (object);
			
			switch (type) {
				case OPLisTree::OP_LIST_EMPTY:
					more_data = FALSE;
					break;
				case OPLisTree::OP_LIST_MARK:
					if (skip_end) {
						val[0] = TYP_SYNC_MARK;
						this->StoreData (val, 1);
					} else {
						more_data = FALSE;
					}
					break;
				case OPLisTree::OP_LIST_DATA:
					this->StoreObject (object);
					break;
				case OPLisTree::OP_LIST_NEW_LEVEL:
					val[0] = TYP_SYNC_BEGIN;
					this->StoreData (val, 1);
					skip_end++;
					break;
				case OPLisTree::OP_LIST_END_LEVEL:
					val[0] = TYP_SYNC_END;
					this->StoreData (val, 1);
					skip_end--;
					break;
				default:
					OP_ABORT ("Invalid type");
					break;
			}
		} while (more_data);
	}
	
	Card8 val[1] = { TYP_SYNC_MARK };
	this->StoreData (val, 1);
}

void
OPStorage::StoreBlockBegin ()
{
	OP_METHOD (StoreBlockBegin, 0);
	
	if (this->storage_infos->skip) {
		this->storage_infos->skip++;
	}
	
	if ( (this->storage_infos->more.IsValid ())
	  && (this->storage_infos->skip == 0) ) {
		
		Card8 val[1] = { 0 };
		Count skip_end  = 0;
		Bool  more_data = TRUE;
		
		OPLisTree::Type type = this->storage_infos->more->PeekHeadType ();
		
		if (type == OPLisTree::OP_LIST_END_LEVEL) {
			this->storage_infos->skip++;
			goto store;
		}
		
		do {
			OPObjectRef object;
			OPLisTree::Type type = this->storage_infos->more->PopHead (object);
			
			switch (type) {
				case OPLisTree::OP_LIST_EMPTY:
					more_data = FALSE;
					break;
				case OPLisTree::OP_LIST_MARK:
					val[0] = TYP_SYNC_MARK;
					this->StoreData (val, 1);
					break;
				case OPLisTree::OP_LIST_DATA:
					this->StoreObject (object);
					break;
				case OPLisTree::OP_LIST_NEW_LEVEL:
					more_data = FALSE;
					break;
				case OPLisTree::OP_LIST_END_LEVEL:
					OP_ABORT ("Unexpected end of block");
					break;
				default:
					OP_ABORT ("Invalid type");
					break;
			}
		} while (more_data);
	}
	
store:
	Card8 val[1] = { TYP_SYNC_BEGIN };
	this->StoreData (val, 1);
}

void
OPStorage::StoreBlockEnd ()
{
	OP_METHOD (StoreBlockEnd, 0);
	
	if ( (this->storage_infos->more.IsValid ())
	  && (this->storage_infos->skip == 0) ) {
		
		Card8 val[1] = { 0 };
		Count skip_end  = 0;
		Bool  more_data = TRUE;
		
		do {
			OPObjectRef object;
			OPLisTree::Type type = this->storage_infos->more->PopHead (object);
			
			switch (type) {
				case OPLisTree::OP_LIST_EMPTY:
					more_data = FALSE;
					break;
				case OPLisTree::OP_LIST_MARK:
					val[0] = TYP_SYNC_MARK;
					this->StoreData (val, 1);
					break;
				case OPLisTree::OP_LIST_DATA:
					this->StoreObject (object);
					break;
				case OPLisTree::OP_LIST_NEW_LEVEL:
					val[0] = TYP_SYNC_BEGIN;
					this->StoreData (val, 1);
					skip_end++;
					break;
				case OPLisTree::OP_LIST_END_LEVEL:
					if (skip_end) {
						val[0] = TYP_SYNC_END;
						this->StoreData (val, 1);
						skip_end--;
					} else {
						more_data = FALSE;
					}
					break;
				default:
					OP_ABORT ("Invalid type");
					break;
			}
		} while (more_data);
	}
	
	Card8 val[1] = { TYP_SYNC_END };
	this->StoreData (val, 1);
	
	if (this->storage_infos->skip) {
		this->storage_infos->skip--;
	}
}


/*
 *	Restore (and discard) all data found in the stream, until
 *	a matching synchronisation mark is reached.
 *
 *	If normal data is restored and a synchronisation mark is
 *	reached in the stream, null/empty objects will be returned
 *	instead, until a call to RestoreToSyncMark is done. This
 *	allows for missing data in a stream.
 */

void
OPStorage::RestoreToSyncMark ()
{
	OP_METHOD (RestoreToSyncMark, 0);
	
	//	Are we already locked on a sync mark, but in a non-existing block ?
	//	If so, remember about the mark but don't advance in the stream until
	//	we have left the block...
	
	if (this->storage_infos->skip) {
		if (this->storage_infos->more.IsValid ()) {
			this->storage_infos->more->InsertMark ();
		}
		return;
	}
	
	Card8 val[1];
	this->PopAndRestoreData (val, 1);
	
	//	Don't cross block boundaries : if the mark happens to miss inside a
	//	block, don't leave the block !
	
	if (val[0] == TYP_SYNC_END) {
		this->PushBackData (val, 1);
		if (this->storage_infos->more.IsValid ()) {
			this->storage_infos->more->InsertMark ();
		}
		return;
	}
	
	if (val[0] != TYP_SYNC_MARK) {
		this->PushBackData (val, 1);
		this->RestoreToMark (TYP_SYNC_MARK);
	}
	
	if (this->storage_infos->more.IsValid ()) {
		this->storage_infos->more->InsertMark ();
	}
}


void
OPStorage::RestoreBlockBegin ()
{
	OP_METHOD (RestoreBlockBegin, 0);
	
	Card8 val[1];
	
	this->PopAndRestoreData (val, 1);
	
	//	Have we reached a mark ? If so, we need to remember that the
	//	entire block does not exist and that when restoring marks inside
	//	of the block, we just have to discard them until the block has
	//	been left.
	
	if (val[0] == TYP_SYNC_MARK) {
		this->PushBackData (val, 1);
		if (this->storage_infos->more.IsValid ()) {
			this->storage_infos->more->InsertNewLevel ();
		}
		this->storage_infos->skip++;
		return;
	}
	
	OP_ASSERT (val[0] == TYP_SYNC_BEGIN);
	
	if (this->storage_infos->more.IsValid ()) {
		this->storage_infos->more->InsertNewLevel ();
	}
}

void
OPStorage::RestoreToBlockEnd ()
{
	OP_METHOD (RestoreToBlockEnd, 0);
	
	//	Are we already locked on a sync mark in a non-existing block ? If so,
	//	we have just left such a block...
	
	if (this->storage_infos->skip) {
		if (this->storage_infos->more.IsValid ()) {
			this->storage_infos->more->InsertEndLevel ();
		}
		this->storage_infos->skip--;
		return;
	}
	
	this->RestoreToMark (TYP_SYNC_END);
	
	if (this->storage_infos->more.IsValid ()) {
		this->storage_infos->more->InsertEndLevel ();
	}
}


/*
 *	Restore (and eventually discard) data until the proper mark
 *	is reached. The discarded data is stored into a list tree,
 *	which will be attached to the restored object.
 */

void
OPStorage::RestoreToMark (Card8 mark)
{
	OP_METHOD (RestoreToMark, 1);
	OP_ARG (1, Card8, mark);
	
	Card8 ref[8];
	
	OPObjectRef  object  = OPObjectNull;
	OPArrayRef   array   = OPArrayNull;
	OPAtomRef    atom    = OPAtomNull;
	OPRuntimeRef runtime = OP_RUNTIME (OPObject);
	
	for (;;) {
		
		this->PopAndRestoreData (ref, 1);
		
		if (ref[0] == mark) {
			return;
		}
		
		switch (ref[0]) {
			
			case TYP_SYNC_MARK:
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->InsertMark ();
				}
				break;
			
			case TYP_SYNC_BEGIN:
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->InsertNewLevel ();
				}
				this->RestoreToBlockEnd ();
				break;
			case TYP_SYNC_END:
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->InsertEndLevel ();
				}
				break;
			
			case TYP_INT32:
			case TYP_CARD32:
			case TYP_ADDR32:
			case TYP_INT32_8:
			case TYP_CARD32_8:
			case TYP_INT32_16:
			case TYP_CARD32_16:
			case TYP_INT64:
			case TYP_CARD64:
			case TYP_ADDR64:
			case TYP_NULL:
			case TYP_BOOL_FALSE:
			case TYP_BOOL_TRUE:
			case TYP_INT32_ZERO:
			case TYP_CARD32_ZERO:
			case TYP_REAL:
			case TYP_REAL_INT:
			case TYP_REAL_ZERO:
			case TYP_REAL_PONE:
			case TYP_REAL_MONE:
				this->PushBackData (ref, 1);
				this->RestoreObject (object, runtime);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (object);
				}
				break;
			case TYP_O_DECL_NAME:
			case TYP_O_DECL_ID:
			case TYP_O_INST_REF:
			case TYP_O_DATA:
			case TYP_O_SLOT_DATA:
			case TYP_A_DECL_OBJ:
				this->PushBackData (ref, 1);
				this->RestoreObject (object, runtime);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (object);
				}
				break;
			case TYP_ARRAY:
				this->PushBackData (ref, 1);
				this->RestoreArray (array, runtime);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (array);
				}
				break;
			case TYP_STRING:
				object = OP_NEW (OPString);
				object->RestoreInstance (this);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (object);
				}
				break;
			case TYP_TYPE_INFO:
				this->RestoreAtom (atom);
				break;
			case TYP_O_PHANTOM:
				object = OP_NEW (OPPhantom);
				object.ReturnRawObject ()->RestoreInstance (this);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (object);
				}
				break;
			case TYP_A_DECL_NAME:
			case TYP_A_DECL_ID:
				this->PushBackData (ref, 1);
				this->RestoreAtom (atom);
				if (this->storage_infos->more.IsValid ()) {
					this->storage_infos->more->Insert (atom);
				}
				break;
			case TYP_BIN_8:
			case TYP_BIN_16:
			case TYP_BIN_32:
				this->PushBackData (ref, 1);
				this->RestoreBinData (0, 0);
				break;
			default:
				OP_ABORT ("RestoreToMark did not find mark.");
		}
	}
}

/****************************************************************/

Bool
OPStorage::CheckForMark ()
{
	OP_METHOD (CheckForMark, 0);
	Card8 ref[1];
	this->PopAndRestoreData (ref, 1);
	this->PushBackData (ref, 1);
	return ref[0] == TYP_SYNC_MARK;
}

Bool
OPStorage::CheckForBegin ()
{
	OP_METHOD (CheckForBegin, 0);
	Card8 ref[1];
	this->PopAndRestoreData (ref, 1);
	this->PushBackData (ref, 1);
	return ref[0] == TYP_SYNC_BEGIN;
}

Bool
OPStorage::CheckForEnd ()
{
	OP_METHOD (CheckForEnd, 0);
	Card8 ref[1];
	this->PopAndRestoreData (ref, 1);
	this->PushBackData (ref, 1);
	return ref[0] == TYP_SYNC_END;
}

/****************************************************************/
