/*
 *	turbosrc.cxx
 *
 *	Turbo Search implementation for OPaC. A very fast way to find
 *	patterns in long texts.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/turbosrc.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPTurboSearch);

/****************************************************************/

OPTurboSearch::OPTurboSearch ()
{
	this->len     = 0;
	this->last    = 0;
	this->pattern = 0;
}

OPTurboSearch::~OPTurboSearch ()
{
	if (this->pattern) {
		OP_FREE (this->pattern);
		this->pattern = 0;
	}
}

/****************************************************************/

/*
 *	Initialise the search string. It must be at least two character
 *	long, or otherwise, it would be useless to use this algorithm.
 */

void
OPTurboSearch::Setup (const LatinChar* pattern, Count len)
{
	if (this->pattern) {
		OP_FREE (this->pattern);
		this->pattern = 0;
	}
	
	this->pattern = OP_ALLOC_ARRAY (Card8, len);
	this->len     = len;
	this->last    = len-1;
	
	Count i;
	
	for (i = 0; i < len; i++) {
		this->pattern[i] = (Card8)pattern[i];
	}
	
	for (i = 0; i < 256; i++) {
		this->shift[i] = len + 1;
	}
	
	for (i = 0; i < len; i++) {
		this->shift[this->pattern[i]] = this->last;
	}
	
	this->look_at = 0;
	
	while (this->look_at < this->last) {
		if (this->pattern[this->last] == this->pattern[this->len-this->look_at-2]) {
			break;
		}
		this->look_at++;
	}

	this->look_after = this->len + this->look_at;
}

/*
 *	T-search main function. The caller must insure that the pattern
 *	to search for exists at the end of the text. If it were not the
 *	case, the algorithm would never stop.
 */

Count
OPTurboSearch::Search (const LatinChar* text)
{
	Count i;
	const Card8* t = (const Card8*)text;
	
	for (;;) {
		
		i = this->last;
		
		//	Skip very quickly through the text as long as the pattern
		//	does not even match with its last character.
		
		while (t[i] != this->pattern[i]) {
			t += this->shift[t[this->len]];
		}
		
		i--;
		
		//	Compare the rest of the pattern...
		
		while (t[i] == pattern[i]) {
			if (i-- <= 0) break;
		}
		
		//	Skip as far as possible if there is a mismatch.
		
		t += this->look_at + this->shift[t[this->look_after]];
	}
	
	return (Count)((Address)(t)-(Address)(text));
}

