/*
 *	view.cxx
 *
 *	Base View class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/container.h>
#include <OPaC/except.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/springboard.h>
#include <OPaC/targetmacro.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPView);

/****************************************************************/

#define	DEF_ENUM_BEGIN(T)	OP_IMPLEMENT_TYPEENUM(T) {
#define	DEF_ENUM(T,n)		OP_IMPLEMENT_TYPEENUM_DEF(T)
#define	DEF_ENUM_END(T)		OP_IMPLEMENT_TYPEENUM_END() };

#include <OPaC/enum/dragdrop.h>

/****************************************************************/

OP_TARGET_ACTOR (OPView, Display);
OP_TARGET_ACTOR (OPView, MakeMainResponder);
OP_TARGET_ACTOR (OPView, ReorderFirst);
OP_TARGET_ACTOR (OPView, ReorderLast);

OP_NOTIFY_EVENT (OPView, SizeChanged);
OP_NOTIFY_EVENT (OPView, OriginChanged);

/****************************************************************/

static const OPaC_BitsetDataDef local_state_info[] = {
	{ 0xFFFFFFFF,			"View state",			OP_BITSET_DATA_MASK },
	{ OP_VIEW_IS_ENABLED,	"enabled",				OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_ALIVE,		"may become active",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_FROZEN,	"frozen (NIY)",			OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_HOT,		"hot (NIY)",			OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_ACTIVE,	"active (has focus)",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_DRAGGING,	"is drag source",		OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_DROPPING,	"is drop source",		OP_BITSET_DATA_BIT },
	{ OP_VIEW_IS_DISABLED,	"globally disabled",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_DRAG,		"accepts dragging",		OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_DROP,		"accepts dropping",		OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_DETECT,	"accepts detection",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_MAIN,		"accepts main resp.",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_UI_EDIT,	"accepts GUI edition",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_UI_MOVE,	"accepts GUI moving",	OP_BITSET_DATA_BIT },
	{ OP_VIEW_OK_DRAW,      "uses standard drawing",OP_BITSET_DATA_BIT }
};

/****************************************************************/

OP_PUBLISH_SLOT (OPView, state,    ReturnState,   SetState, "View state", Card32);
OP_PUBLISH_SLOT (OPView, user_tag, ReturnUserTag, SetUserTag, "User tag", Card32);
OP_PUBLISH_SLOT (OPView, user_name, ReturnUserName, SetUserName, "User name", OPText);
OP_PUBLISH_SLOT (OPView, active_rank, ReturnActiveRank, SetActiveRank, "Active rank", Int32);
OP_PUBLISH_SLOT (OPView, box, ReturnBox, SetBox, "Bounding box", OPRect);

OP_PUBLISH_BITSET_INFO (OPView, state, local_state_info);

/****************************************************************/

OPAtomRef OPView::atom_look_man_changed;

/****************************************************************/

/*
 *	Initialise a view : no super-view, no children...
 */

OPView::OPView ()
{
	OP_METHOD (OPView, 0);
	
	this->window = 0;
	this->parent = 0;
	this->spring = 0;
	
	this->sibling.next = 0;
	this->sibling.prev = 0;
	
	this->child.head = 0;
	this->child.tail = 0;
	
	this->num_overlays = 0;
	this->no_draw_back = FALSE;
	
	this->box.MakeEmpty ();
	
	this->state = OP_VIEW_IS_ENABLED			//	may get events
				| OP_VIEW_IS_ALIVE				//	may become active
				| OP_VIEW_OK_DETECT				//	may be clicked
				| OP_VIEW_OK_DROP				//	may accept a dropping
				| OP_VIEW_OK_MAIN				//	may become main responder
				| OP_VIEW_OK_UI_EDIT			//	may be edited by UI edition tools
				| OP_VIEW_OK_UI_MOVE			//	may be moved by UI edition tools
				| OP_VIEW_OK_DRAW;				//	uses standard drawing
	
	this->tag = 0;
	
	this->look_cursor  = OP_CURSOR_DEFAULT;
	this->named_cursor = OPAtomNull;
	
	if (this->atom_look_man_changed.IsInvalid ()) {
		this->atom_look_man_changed = "LookManChanged";
	}
}



/*
 *	Destroy a view. Nothing special needs to be done here,
 *	since this can only be called if there are no more
 *	references to this view.
 */

OPView::~OPView ()
{
	OP_METHOD (~OPView, 0);
	
	OP_ASSERT (this->child.head.IsInvalid ());
	OP_ASSERT (this->child.tail.IsInvalid ());
	OP_ASSERT (this->sibling.next.IsInvalid ());
	OP_ASSERT (this->sibling.prev.IsInvalid ());
	OP_ASSERT (this->active_list.IsInvalid ());
}



/*
 *	Prepare a view for its destruction. This will remove
 *	the links to the other views and to the window...
 */

void
OPView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPViewRef self = this;
	OPViewRef iter;
	OPViewRef next;
	
	//	Disconnect from the active list.
	
	this->RemoveActive ();
	
	//	Disconnect the children and delete them explicitely.
	
	iter = this->child.head;
	
	while (iter.IsValid ()) {
		
		iter->GetNextSibling (next);
		iter->SetParent (OPViewNull);
		iter->Delete ();
		
		iter = next;
	}
	
	//	If there is still some parent around, remove the pending connection to
	//	it. This will remove connection with the siblings...
	
	if (this->parent.IsValid ()) {
		this->SetParent (OPViewNull);
	}
	
	//	The views which have been inserted in a window will notify the window
	//	before being destroyed. A window hasn't to notify itself about its own
	//	destruction !
	
	if ( (this->window.IsValid ())
	  && (self != this->window) ) {
		this->window->NotifyRemoveView (self);
	}

	this->spring = OPSpringBoardNull;
	
	//	Remove the active view list.
	
	if (this->active_list.IsValid ()) {
		this->active_list->Delete ();
		this->active_list = OPListNull;
	}
	
	super::Delete ();
}



/*
 *	Store the view and its sub-views. This will also save
 *	the siblings and the parent...
 *
 *	This might store much more than you would expect at
 *	first sight (usually, storing a visible view will
 *	store the full window).
 *
 *	If you just want to store the view and its sub-views,
 *	disconnect the view from its parent/window...
 */

void
OPView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	//	Save individually each child, thus removing the necessity to maintain
	//	the sibling list.
	
	static OPStorageSync info ("<OPView child_head child_tail { child }"
							   " box state tag spring num_overlays active_list"
							   " look_cursor named_cursor name>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->child.head;
	storage << this->child.tail;
	
	storage << OPStorageSync::repeat;
	
	if (this->child.head.IsValid ()) {
		
		OPViewRef child = this->child.head;
		
		do {
			storage << child;
		} while (child->GetNextSibling (child));
	}
	
	storage << OPStorageSync::end_repeat;
	
	storage << this->box;
	storage << this->state;
	storage << this->tag;
	storage << this->spring;
	storage << this->num_overlays;
	storage << this->active_list;
	
	storage << this->look_cursor;
	storage << this->named_cursor;
	storage << this->name;
	
	storage << OPStorageSync::end;
}



/*
 *	Restore the view, its sub-views, its siblings and its
 *	parent.
 */

void
OPView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	//	Restore individually each child, thus removing the necessity to maintain
	//	the sibling list. By the way, also restore the active list...
	
	OPViewRef child_head;
	OPViewRef child_tail;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> child_head;
	storage >> child_tail;
	
	storage >> OPStorageSync::repeat;
	
	if (child_tail.IsValid ()) {
		
		OPViewRef child;
		OPViewRef self = this;
		
		//	Restore each subview till the last subview in the list has been
		//	read. This will rebuild the children list, thanks to SetParent.
		
		do {
			storage >> child;
			child->SetParent (self);
		} while (child != child_tail);
	}
	
	storage >> OPStorageSync::end_repeat;
	
	storage >> this->box;
	storage >> this->state;
	storage >> this->tag;
	storage >> this->spring;
	storage >> this->num_overlays;
	storage >> this->active_list;	
	
	storage >> this->look_cursor;
	storage >> this->named_cursor;
	storage >> this->name;
	
	storage >> OPStorageSync::end;
	
	if (this->window.IsValid ()) {
		if (this->state & OP_VIEW_IS_ACTIVE) {
			this->window->SetMainResponder (this, FALSE);
		}
	}
}



/*
 *	This will be called when the view's geometry changes
 *	with (Set/Change)(Origin/Size).
 *
 *	This will move every child view if needed and will
 *	also notify them of the change of size, if any.
 */

void
OPView::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	OPViewRef  self = this;
	OPRelCoord dx   = box.X1 () - this->box.X1 ();
	OPRelCoord dy   = box.Y1 () - this->box.Y1 ();
	
	if ( (this->box.DX () != box.DX ())
	  || (this->box.DY () != box.DY ()) ) {
		
		//	First of all, tell the spring board that it has to resize and move
		//	its views.
		
		if (this->spring.IsValid ()) {
			
			OPRect the_box = box;
			
			if (self == this->window) {
				the_box.SetOrigin (0, 0);
			}
			
			this->spring->Resize (the_box);
		}
			
		//	If the geometry change induced some resizing, then we will have to
		//	notify every child view. This notification is done before the real
		//	resizing.
		
		OPCoord dx = box.DX ();
		OPCoord dy = box.DY ();
		
		OPViewRef iter = this->child.head;
		
		while (iter.IsValid ()) {
			iter->ParentResizing (dx, dy);
			iter->GetNextSibling (iter);
		}
		
		//	We can now tell our parent about the change of size. Perhaps
		//	the parent will once again change our size... Hope we don't
		//	cycle forever.
		
		this->box = box;
		
		if (this->parent.IsValid ()) {
			this->parent->ChildChangedSize (this);
		}
		
		this->Report (notify_SizeChanged);
		
	} else {
		
		this->box = box;
	}
	
	if ( (dx || dy)
	  && (self != this->window) )  {
		
		//	If the geometry change induced some translation, then we have to
		//	move every child view...
		
		OPViewRef iter = this->child.head;
		
		while (iter.IsValid ()) {
			iter->ChangeOrigin (dx, dy);
			iter->GetNextSibling (iter);
		}
		
		this->Report (notify_OriginChanged);
	}
}



/*
 *	Enable the view. This will allow it to become the main
 *	responder, to get events, etc...
 *
 *	This method will change the children's state when the
 *	broadcast parameter is being set to TRUE.
 *
 *	No redrawing is performed. The caller should Display
 *	the view in order to get things up to date.
 *
 *	NB: Enabling views this way sets OP_VIEW_IS_ENABLED
 *		but it does not affect OP_VIEW_IS_DISABLED, which
 *		is more global and overrides the enabled state.
 */

void
OPView::SetEnableState (Bool ok, Bool broadcast)
{
	OP_METHOD (SetEnableState, 2);
	OP_ARG (1, Bool, ok);
	OP_ARG (2, Bool, broadcast);
	
	this->state &= ~ (OP_VIEW_IS_ENABLED);
	this->state |= ok ? OP_VIEW_IS_ENABLED : 0;
	
	if (broadcast == FALSE) {
		return;
	}
	
	OPViewRef iter = this->child.head;
	
	while (iter.IsValid ()) {
		iter->SetEnableState (ok, broadcast);
		iter->GetNextSibling (iter);
	}
}



/*
 *	Remove the view from the active list.
 */

Bool
OPView::RemoveActive ()
{
	OP_METHOD (RemoveActive, 0);
	
	if ( (this->parent.IsInvalid ())
	  || (this->window.IsInvalid ()) ) {
		return FALSE;
	}
	
	//	Tell the window that this view won't be in the active list any more.
	//	This gives the window the opportunity to update the default/current
	//	active view.
	
	this->window->NoLongerActive (this);
	this->parent->RemoveActive (this);
	
	return TRUE;
}

Bool
OPView::RemoveActive (OPViewIn child)
{
	OP_METHOD (RemoveActive, 1);
	OP_ARG (1, OPViewIn, child);
	
	if (this->active_list.IsValid ()) {
		this->active_list->Remove (child);
		this->NotifyActiveListChange ();
		return TRUE;
	}
	
	return FALSE;
}

const OPListRef&
OPView::ReturnAndCreateActiveList ()
{
	OP_METHOD (ReturnAndCreateActiveList, 0);
	
	if (this->active_list.IsInvalid ()) {
		this->active_list = OP_NEW (OPList);
		this->active_list->SetInsertMode (FALSE);
	}
	
	return this->active_list;
}

const OPListRef&
OPView::ReturnActiveList () const
{
	OP_METHOD (ReturnActiveList, 0);
	
	return this->active_list;
}

void
OPView::NotifyActiveListChange ()
{
	OP_METHOD (NotifyActiveListChange, 0);
	
	if ( (this->active_list.IsValid ())
	  && (this->active_list->TotalCount () == 0) ) {
		this->active_list->Delete ();
		this->active_list = OPListNull;
	}
}


/*
 *	Return a reference to the view, that's "this" !
 */

Bool
OPView::GetView (OPViewOut view)
{
	OP_METHOD (GetView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this;
	return TRUE;
}



/*
 *	Return TRUE if the view accepts to become the main
 *	responder.
 */

Bool
OPView::WouldAcceptMain () const
{
	OP_METHOD (WouldAcceptMain, 0);
	
	return ( (this->state & OP_VIEW_IS_ENABLED)
		  && ((this->state & OP_VIEW_IS_DISABLED) == 0)
		  && (this->ReturnVisible ())
		  && (this->state & OP_VIEW_OK_MAIN) ) ? TRUE : FALSE;
}


/*
 *	Make the view the main responder.
 */

void
OPView::MakeMainResponder ()
{
	if (this->window.IsValid ()) {
		this->window->SetMainResponder (this, FALSE);
	}
}

/*
 *	Notify the view that it has become the main responder
 *	and redraw it.
 */

void
OPView::BecomeMain (Bool clicked)
{
	OP_METHOD (BecomeMain, 1);
	OP_ARG (1, Bool, clicked);
	
	super::BecomeMain (clicked);
	
	this->state |= OP_VIEW_IS_ACTIVE;
	this->Display ();
}



/*
 *	Notify the view that it no longer is the main responder
 *	and redraw it.
 */

void
OPView::ResignMain ()
{
	OP_METHOD (ResignMain, 0);
	
	super::ResignMain ();
	
	this->state &= ~ (OP_VIEW_IS_ACTIVE);
	this->Display ();
}



/*
 *	Define which window this view belongs to. This gets
 *	called automatically when the parent view gets defined.
 *	This will also set the window for every child view.
 */

Bool
OPView::SetWindow (OPWindowIn window)
{
	OP_METHOD (SetWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	//	If the window which has been specified is equal to the current window,
	//	don't do anything...
	
	if (this->window == window) {
		return TRUE;
	}
	
	this->RemoveActive ();
	
	//	Tell the window about the view being added or removed.
	
	if (this->window.IsValid ()) {
		this->window->NotifyRemoveView (this);
	}
	
	if (window.IsValid ()) {
		window->NotifyInsertView (this);
	}
	
	this->window = window;
	
	//	Tell the sub-views about the window. This may get recursive, if the sub-
	//	views have also sub-views...
	
	OPViewRef iter = this->child.head;
	
	while (iter.IsValid ()) {
		
		OPView* i0c = (OPView*)iter.CheckAndResolve ();
		OPView* i0n = (OPView*)i0c->sibling.next.Resolve ();
		OPView* i0p = (OPView*)i0c->sibling.prev.Resolve ();
		
		iter->SetWindow (window);
		
		OPView* i1c = (OPView*)iter.CheckAndResolve ();
		OPView* i1n = (OPView*)i1c->sibling.next.Resolve ();
		OPView* i1p = (OPView*)i1c->sibling.prev.Resolve ();
		
		OP_ASSERT (i0c == i1c);
		OP_ASSERT (i0n == i1n);
		OP_ASSERT (i0p == i1p);
		
		iter->GetNextSibling (iter);
	}
	
	return TRUE;
}



/*
 *	Define the parent and link with the brothers and the
 *	window. If the reference was invalid, unlink with the
 *	brothers and remove any reference to the window.
 */

Bool
OPView::SetParent (OPViewIn parent)
{
	OP_METHOD (SetParent, 1);
	OP_ARG (1, OPViewIn, parent);
	
	if (this->parent == parent) {
		return TRUE;
	}
	
	if (this->IsKindOf (OP_RUNTIME (OPWindow))) {
		return TRUE;
	}
	
	this->RemoveActive ();
	
	if (parent.IsValid ()) {
		
		//	Ask the parent if it would accept the new child. If not, we may not
		//	insert into the parent's subviews.

		if (parent->WouldAcceptChild (this) == FALSE) {
			return FALSE;
		}

		//	If the parent changes from a valid one to another valid one, we first
		//	have to remove the previous parent -- do this with a small recursion !
		
		if (this->parent.IsValid ()) {
			this->SetParent (OPViewNull);
		}
		
		this->parent = parent;
		this->SetWindow (parent->window);
		
		//	Insert "self" into parent's children list...
		
		this->sibling.next = OPViewNull;
		this->sibling.prev = parent->child.tail;
		
		if (this->sibling.prev.IsValid ()) {
			this->sibling.prev->sibling.next = this;
			parent->child.tail               = this;
		} else {
			parent->child.head = this;
			parent->child.tail = this;
		}

		parent->NotifyInsertChild (this);
		
	} else {
		
		this->parent->NotifyRemoveChild (this);
		
		//	First of all, remove from the active list, before the parent vanishes
		//	or we would get in trouble.
		
		this->RemoveActive ();
		
		//	Remove from parent's children list...
		
		if (this->sibling.prev.IsValid ()) {
			this->sibling.prev->sibling.next = this->sibling.next;
		} else {
			this->parent->child.head = this->sibling.next;
		}
		
		if (this->sibling.next.IsValid ()) {
			this->sibling.next->sibling.prev = this->sibling.prev;
		} else {
			this->parent->child.tail = this->sibling.prev;
		}
		
		this->sibling.prev = OPViewNull;
		this->sibling.next = OPViewNull;
		this->parent = OPViewNull;
		
		this->SetWindow (OPWindowNull);
	}
	
	if (this->window.IsValid ()) {
		this->window->NotifyReparentView (this);
	}
	
	this->SetNextResponder (parent);

	return TRUE;
}



/*
 *	Define the spring curve.
 */

Bool
OPView::SetSpringBoard (OPSpringBoardIn spring)
{
	OP_METHOD (SetSpringBoard, 1);
	OP_ARG (1, OPSpringBoardIn, spring);
	
	this->spring = spring;
	
	return TRUE;
}



/*
 *	Ask the view if it would accept the child. If not, it
 *	will not insert itself as a subview...
 */

Bool
OPView::WouldAcceptChild (OPViewIn child) const
{
	OP_METHOD (WouldAcceptChild, 1);
	OP_ARG (1, OPViewIn, child);
	return TRUE;
}



/*
 *	Once the child has been inserted into the subview list
 *	we get a notification, which allows us to modify the
 *	child's attributes, if needed.
 */

void
OPView::NotifyInsertChild (OPViewIn child)
{
	OP_METHOD (NotifyInsertChild, 1);
	OP_ARG (1, OPViewIn, child);
}



/*
 *	We get a notification just before the child view gets
 *	removed from the subview list.
 */

void
OPView::NotifyRemoveChild (OPViewIn child)
{
	OP_METHOD (NotifyRemoveChild, 1);
	OP_ARG (1, OPViewIn, child);
	
	if (this->spring.IsValid ()) {
		this->spring->RemoveView (child);
	}
}



/*
 *	Reorder the views so that this view will be the first
 *	one with respect to the siblings.
 */

void
OPView::ReorderFirst ()
{
	OP_METHOD (ReorderFirst, 0);

	if ( (this->sibling.prev.IsValid ())
	  && (this->parent.IsValid ()) ) {
		
		OPViewRef self = this;

		//	Remove self from the list. The first test is not really needed,
		//	but it has been left there since we rely on smart compilers and
		//	dumb programmers.
		
		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = this->sibling.next;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = this->sibling.prev;
		
		//	Now insert self at the head of the list.
		
		this->sibling.prev = 0;
		this->sibling.next = this->parent->child.head;

		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = self;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = self;
	}
}



/*
 *	Reorder the views so that this view will be the last
 *	one with respect to the siblings.
 */

void
OPView::ReorderLast ()
{
	OP_METHOD (ReorderLast, 0);

	if ( (this->sibling.next.IsValid ())
	  && (this->parent.IsValid ()) ) {
		
		OPViewRef self = this;

		//	Remove self from the list. The first test is not really needed,
		//	but it has been left there since we rely on smart compilers and
		//	dumb programmers.
		
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = this->sibling.prev;
		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = this->sibling.next;
		
		//	Now insert self at the end of the list.
		
		this->sibling.next = 0;
		this->sibling.prev = this->parent->child.tail;

		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = self;
		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = self;
	}
}



/*
 *	Reorder the views so that this view comes just before
 *	the specified view in the viewing list.
 *
 *	If no view is specified, make this view the last one
 *	(it is before no view).
 */

void
OPView::ReorderBefore (OPViewIn view)
{
	OP_METHOD (ReorderBefore, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsInvalid ()) {
		this->ReorderLast ();
		return;
	}
	
	if (view->sibling.prev.IsInvalid ()) {
		this->ReorderFirst ();
		return;
	}
	
	if ( (this->sibling.next != view)
	  && (this->parent.IsValid ())
	  && (this->parent == view->parent) ) {
		
		OPViewRef self = this;

		//	Remove self from the list.
		
		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = this->sibling.next;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = this->sibling.prev;
		
		//	Now insert self just before the specified view.
		
		this->sibling.prev = view->sibling.prev;
		this->sibling.next = view;

		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = self;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = self;
	}
}



/*
 *	Reorder the views so that this view comes just after
 *	the specified view in the viewing list.
 *
 *	If no view is specified, make this view the first one
 *	(it is after no view).
 */

void
OPView::ReorderAfter (OPViewIn view)
{
	OP_METHOD (ReorderAfter, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsInvalid ()) {
		this->ReorderFirst ();
		return;
	}

	if (view->sibling.next.IsInvalid ()) {
		this->ReorderLast ();
		return;
	}

	if ( (this->sibling.prev != view)
	  && (this->parent.IsValid ())
	  && (this->parent == view->parent) ) {
		
		OPViewRef self = this;

		//	Remove self from the list.
		
		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = this->sibling.next;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = this->sibling.prev;
		
		//	Now insert self just before the specified view.
		
		this->sibling.prev = view;
		this->sibling.next = view->sibling.next;

		(this->sibling.prev.IsValid () ? this->sibling.prev->sibling.next
									   : this->parent->child.head) = self;
		(this->sibling.next.IsValid () ? this->sibling.next->sibling.prev
									   : this->parent->child.tail)  = self;
	}
}



/*
 *	Change the view's bounding box. This will automatically
 *	call "ChangeGeometry".
 */

Bool
OPView::SetBox (OPRectIn box)
{
	OP_METHOD (SetBox, 1);
	OP_ARG (1, OPRectIn, box);
	
	if (box != this->box) {
		this->ChangeGeometry (box);
	}

	return TRUE;
}



/*
 *	Set the view's origin. This will automatically call
 *	"ChangeGeometry".
 */

void
OPView::SetOrigin (OPCoord x, OPCoord y)
{	
	OP_METHOD (SetOrigin, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	if ( (this->box.X1 () == x)
	  && (this->box.Y1 () == y) ) {
		return;
	}
	
	OPRect box;
	
	box.SetOrigin (x, y);
	box.SetSize (this->box.DX (), this->box.DY ());
	
	this->ChangeGeometry (box);
}



/*
 *	Set the view's size. This will automatically call
 *	"ChangeGeometry".
 */

void
OPView::SetSize (OPCoord dx, OPCoord dy)
{
	OP_METHOD (SetSize, 2);
	OP_ARG (1, OPCoord, dx);
	OP_ARG (2, OPCoord, dy);
	
	if ( (this->box.DX () == dx)
	  && (this->box.DY () == dy) ) {
		return;
	}
	
	OPRect box;
	
	box.SetOrigin (this->box.X1 (), this->box.Y1 ());
	box.SetSize (dx, dy);
	
	this->ChangeGeometry (box);
}




/*
 *	Move the view's origin. This will automatically call
 *	"ChangeGeometry". The position is relative to the
 *	last one.
 */

void
OPView::ChangeOrigin (OPRelCoord dx, OPRelCoord dy)
{
	OP_METHOD (ChangeOrigin, 2);
	OP_ARG (1, OPRelCoord, dx);
	OP_ARG (2, OPRelCoord, dy);
	
	if ( (dx == 0)
	  && (dy == 0) ) {
		return;
	}
	
	OPRect box;
	
	box.SetOrigin ( (OPCoord)(this->box.X1 () + dx),
					(OPCoord)(this->box.Y1 () + dy) );
	box.SetSize (this->box.DX (), this->box.DY ());
	
	this->ChangeGeometry (box);
}



/*
 *	Change the view's size. This will automatically call
 *	"ChangeGeometry". The size will change according to
 *	the specified increments.
 */

void
OPView::ChangeSize (OPRelCoord dx, OPRelCoord dy)
{
	OP_METHOD (ChangeSize, 2);
	OP_ARG (1, OPRelCoord, dx);
	OP_ARG (2, OPRelCoord, dy);
	
	if ( (dx == 0)
	  && (dy == 0) ) {
		return;
	}
	
	OPRect box;
	
	box.SetOrigin (this->box.X1 (), this->box.Y1 ());
	box.SetSize ( (OPCoord)(this->box.DX () + dx),
				  (OPCoord)(this->box.DY () + dy) );
	
	this->ChangeGeometry (box);
}



/*
 *	Set the view to be the first active in the list.
 */

Bool
OPView::SetFirstActive ()
{
	OP_METHOD (SetFirstActive, 0);
	
	if (this->parent.IsValid ()) {
		
		const OPListRef& list = this->parent->ReturnAndCreateActiveList ();
		
		list->Remove (this);
		list->PushHead (this);
		
		this->parent->NotifyActiveListChange ();
		
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Set the view to be the last active in the list.
 */

Bool
OPView::SetLastActive ()
{
	OP_METHOD (SetLastActive, 0);
	
	if (this->parent.IsValid ()) {
		
		const OPListRef& list = this->parent->ReturnAndCreateActiveList ();
		
		list->Remove (this);
		list->PushHead (this);
		
		this->parent->NotifyActiveListChange ();
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	This view will become active before after the specified view.
 *	If the view is the first of the list, just make self the first
 *	active view. If no view is specified, make self the last
 *	active view (before no one).
 */

Bool
OPView::SetActiveBefore (OPViewIn view)
{
	OP_METHOD (SetActiveBefore, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsInvalid ()) {
		return this->SetLastActive ();
	}
	
	if (this->parent.IsValid ()) {
		
		Count pos = 0;
		const OPListRef& list = this->parent->ReturnAndCreateActiveList ();
		
		list->Remove (this);
		list->PushHead (this);
		
		if (list->Find (view, pos)) {
			list->MoveAtPosition (this, pos);
		}
		
		this->parent->NotifyActiveListChange ();
		
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	This view will become active after the specified view.
 *	If the view is the last of the list, just make self the last
 *	active view. If no view is specified, make self the first
 *	active view (after no one).
 */

Bool
OPView::SetActiveAfter (OPViewIn view)
{
	OP_METHOD (SetActiveAfter, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsInvalid ()) {
		return this->SetFirstActive ();
	}
	
	if (this->parent.IsValid ()) {
		
		Count pos = 0;
		const OPListRef& list = this->parent->ReturnAndCreateActiveList ();
		
		list->Remove (this);
		list->PushTail (this);
		
		if (list->Find (view, pos)) {
			list->MoveAtPosition (this, pos+1);
		}
		
		this->parent->NotifyActiveListChange ();
		
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPView::SetActiveRank (Int32 pos)
{
	OP_METHOD (SetActiveRank, 1);
	OP_ARG (1, Int32, pos);
	
	if (pos < 0) {
		return this->RemoveActive ();
	}
	
	if (this->parent.IsValid ()) {
		
		const OPListRef& list = this->parent->ReturnAndCreateActiveList ();
		
		list->Remove (this);
		list->PushTail (this);
		list->MoveAtPosition (this, pos);
		
		this->parent->NotifyActiveListChange ();
		
		return TRUE;
	}
	
	return FALSE;
}

Int32
OPView::ReturnActiveRank () const
{
	OP_METHOD (ReturnActiveRank, 0);
	
	if (this->parent.IsValid ()) {
		
		const OPListRef& list = this->parent->ReturnActiveList ();
		Count pos = 0;
		
		if ( (list.IsValid ())
		  && (list->Find (const_cast<OPView*>(this), pos)) ) {
			return (Int32)(pos);
		}
	}
	
	return -1;
}


/*
 *	Define the user tag associated to the view. This can
 *	be used freely by the user to distinguish between his
 *	views and widgets...
 */

Bool
OPView::SetUserTag (Int32 tag)
{
	OP_METHOD (SetUserTag, 1);
	OP_ARG (1, Int32, tag);
	
	this->tag = tag;
	
	return TRUE;
}

Bool
OPView::SetUserName (OPTextIn name)
{
	OP_METHOD (SetUserName, 1);
	OP_ARG (1, OPTextIn, name);
	
	if (name.IsValid ()) {
		
		if (name->IsKindOf (OP_RUNTIME (OPAtom))) {
			this->name = name;
		} else if (name->IsKindOf (OP_RUNTIME (OPString))) {
			OPStringRef text = name;
			this->name = text;
		} else {
			this->name = name->ReturnString ();
		}
	} else {
		this->name = OPAtomNull;
	}
	
	return TRUE;
}


/*
 *	Return view's containing window.
 */

Bool
OPView::GetWindow (OPWindowOut window) const
{
	OP_METHOD (GetWindow, 1);
	OP_ARG (1, OPWindowOut, window);
	
	window = this->window;
	return window.IsValid ();
}



/*
 *	Return the view's spring curve.
 */

Bool
OPView::GetSpringBoard (OPSpringBoardOut spring) const
{
	OP_METHOD (GetSpringBoard, 1);
	OP_ARG (1, OPSpringBoardOut, spring);
	
	spring = this->spring;
	
	return TRUE;
}


/*
 *	Return the view's parent view. If there is no parent
 *	(eg. because it is a window), a null pointer will be
 *	returned.
 */

Bool
OPView::GetParent (OPViewOut view) const
{
	OP_METHOD (GetParent, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->parent;
	return view.IsValid ();
}




const OPWindowRef&
OPView::ReturnWindow () const
{
	OP_METHOD (ReturnWindow, 0);
	return this->window;
}

const OPViewRef&
OPView::ReturnParent () const
{
	OP_METHOD (ReturnParent, 0);
	return this->parent;
}

const OPSpringBoardRef&
OPView::ReturnSpringBoard () const
{
	OP_METHOD (ReturnSpringBoard, 0);
	return this->spring;
}




Bool
OPView::GetNextSibling (OPViewOut view) const
{
	OP_METHOD (GetNextSibling, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->sibling.next;
	return view.IsValid ();
}

Bool
OPView::GetPrevSibling (OPViewOut view) const
{
	OP_METHOD (GetPrevSibling, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->sibling.prev;
	return view.IsValid ();
}

Bool
OPView::GetFirstChild (OPViewOut view) const
{
	OP_METHOD (GetFirstChild, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->child.head;
	return view.IsValid ();
}

Bool
OPView::GetLastChild (OPViewOut view) const
{
	OP_METHOD (GetLastChild, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->child.tail;
	return view.IsValid ();
}


/*
 *	Return upper left corner coordinates.
 */

void
OPView::GetOrigin (OPCoordOut x, OPCoordOut y) const
{
	OP_METHOD (GetOrigin, 2);
	OP_ARG (1, OPCoordOut, x);
	OP_ARG (2, OPCoordOut, y);
	
	x = this->box.X1 ();
	y = this->box.Y1 ();
}



/*
 *	Return the view's size.
 */

void
OPView::GetSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = this->box.DX ();
	dy = this->box.DY ();
}



/*
 *	Return the view's minimal size. At least 2 pixels
 *	wide or else it would only be a point, not a view !
 */

void
OPView::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = 2;
	dy = 2;
}



/*
 *	Return the view's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPView::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = 10000;
	dy = 10000;
}



/*
 *	Return the view's containing box. This is a faster
 *	way to get the geometry than using GetOrigin/Size.
 */

const OPRect&
OPView::ReturnBox () const
{
	OP_METHOD (ReturnBox, 0);
	
	return this->box;
}

Bool
OPView::GetBox (OPRectOut box) const
{
	OP_METHOD (GetBox, 1);
	OP_ARG (1, OPRectOut, box);
	
	box = this->box;
	
	return TRUE;
}

/*
 *	Return first active view.
 */

OPViewRef
OPView::ReturnFirstActive () const
{
	OP_METHOD (ReturnFirstActive, 0);
	
	OPViewRef view;
	
	if (this->active_list.IsValid ()) {
		this->active_list->Get (0, view);
	}

	while ( (view.IsValid ())
		 && (view->WouldAcceptMain () == FALSE) ) {
		view = view->ReturnNextActive ();
	}
	
	if (view.IsValid ()) {
		OPListRef list = view->ReturnActiveList ();
		
		if ( (list.IsValid ())
		  && (list->TotalCount ()) ) {
			return view->ReturnFirstActive ();
		}
	}
	
	return view;
}



/*
 *	Return last active view.
 */

OPViewRef
OPView::ReturnLastActive () const
{
	OP_METHOD (ReturnLastActive, 0);
	
	OPViewRef view;
	
	if (this->active_list.IsValid ()) {
		Count num = this->active_list->TotalCount ();
		this->active_list->Get (num-1, view);
	}
	
	while ( (view.IsValid ())
		 && (view->WouldAcceptMain () == FALSE) ) {
		view = view->ReturnPrevActive ();
	}
	
	if (view.IsValid ()) {
		OPListRef list = view->ReturnActiveList ();
		
		if ( (list.IsValid ())
		  && (list->TotalCount ()) ) {
			return view->ReturnLastActive ();
		}
	}
	
	return view;
}



/*
 *	Return the view to activate after this one.
 */

OPViewRef
OPView::ReturnNextActive () const
{
	OP_METHOD (ReturnNextActive, 0);
	
	OPViewRef view;
	OPListRef list;
	
	if (this->parent.IsValid ()) {
		list = this->parent->ReturnActiveList ();
	}
	
	if (list.IsValid ()) {
		Count num = 0;
		list->Find (const_cast<OPView*>(this), num);
		list->Get (num+1, view);
	}
	
	return view;
}



/*
 *	Return the view to activate before this one.
 */

OPViewRef
OPView::ReturnPrevActive () const
{
	OP_METHOD (ReturnPrevActive, 0);
	
	OPViewRef view;
	OPListRef list;
	
	if (this->parent.IsValid ()) {
		list = this->parent->ReturnActiveList ();
	}
	
	if (list.IsValid ()) {
		Count num = 0;
		list->Find (const_cast<OPView*>(this), num);
		if (num) list->Get (num-1, view);
	}
	
	return view;
}



/*
 *	Return the view to activate after this one, taking
 *	into account the end of the list.
 */

Bool
OPView::FindNextActive (OPViewOut view) const
{
	OP_METHOD (FindNextActive, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->ReturnNextActive ();
	
	while ( (view.IsValid ())
		 && (view->WouldAcceptMain () == FALSE) ) {
		view = view->ReturnNextActive ();
	}
	
	if (view.IsInvalid ()) {
		
		if (this->parent.IsInvalid ()) {
			view = OPViewNull;
			return FALSE;
		}
		
		if (this->parent == this->window) {
			view = this->window->ReturnFirstActive ();
			return view.IsValid ();
		}
		
		return this->parent->FindNextActive (view);
	}
	
	OPListRef list = view->ReturnActiveList ();
	
	if ( (list.IsValid ())
	  && (list->TotalCount ()) ) {
		view = view->ReturnFirstActive ();
		return view.IsValid ();
	}
	
	return TRUE;
}



/*
 *	Return the view to activate before this one, taking
 *	into account the beginning of the list.
 */

Bool
OPView::FindPrevActive (OPViewOut view) const
{
	OP_METHOD (FindPrevActive, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->ReturnPrevActive ();
	
	while ( (view.IsValid ())
		 && (view->WouldAcceptMain () == FALSE) ) {
		view = view->ReturnPrevActive ();
	}
	
	if (view.IsInvalid ()) {
		
		if (this->parent.IsInvalid ()) {
			view = OPViewNull;
			return FALSE;
		}
		
		if (this->parent == this->window) {
			view = this->window->ReturnLastActive ();
			return view.IsValid ();
		}
		
		return this->parent->FindPrevActive (view);
	}
	
	OPListRef list = view->ReturnActiveList ();
	
	if ( (list.IsValid ())
	  && (list->TotalCount ()) ) {
		view = view->ReturnLastActive ();
		return view.IsValid ();
	}
	
	return TRUE;
}



/*
 *	Return the view's tag. See SetUserTag for details.
 */

Int32
OPView::ReturnUserTag () const
{
	OP_METHOD (ReturnUserTag, 0);
	return this->tag;
}

OPAtomRet
OPView::ReturnUserName () const
{
	OP_METHOD (ReturnUserName, 0);
	return this->name;
}



/*
 *	Set/clear OP_VIEW_OK_DETECT.
 */

void
OPView::SetFlagOkDetect (Bool ok)
{
	OP_METHOD (SetFlagOkDetect, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_DETECT, ok);
}



/*
 *	Set/clear OP_VIEW_OK_UI_EDIT.
 */

void
OPView::SetFlagOkUIEdit (Bool ok)
{
	OP_METHOD (SetFlagOkUIEdit, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_UI_EDIT, ok);
}



/*
 *	Set/clear OP_VIEW_OK_UI_MOVE.
 */

void
OPView::SetFlagOkUIMove (Bool ok)
{
	OP_METHOD (SetFlagOkUIMove, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_UI_MOVE, ok);
}



/*
 *	Set/clear OP_VIEW_OK_MAIN.
 */

void
OPView::SetFlagOkMain (Bool ok)
{
	OP_METHOD (SetFlagOkMain, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_MAIN, ok);
}


/*
 *	Set/clear OP_VIEW_OK_DRAG.
 */

void
OPView::SetFlagOkDrag (Bool ok)
{
	OP_METHOD (SetFlagOkDrag, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_DRAG, ok);
}


/*
 *	Set/clear OP_VIEW_OK_DROP.
 */

void
OPView::SetFlagOkDrop (Bool ok)
{
	OP_METHOD (SetFlagOkDrag, 1);
	OP_ARG (1, Bool, ok);
	this->ChangeState (OP_VIEW_OK_DROP, ok);
}




/*
 *	Set the local enable mode. This will automatically
 *	affect every child view. Affects OP_VIEW_IS_ENABLED.
 *	This will not redraw the view.
 */

void
OPView::SetEnable (Bool ok)
{
	OP_METHOD (SetEnable, 1);
	OP_ARG (1, Bool, ok);
	
	this->SetEnableState (ok, TRUE);
}


/*
 *	Define the visibility of a view. An invisible view
 *	will not respond to detection and drawing, nor will
 *	it get the keyboard focus.
 */

void
OPView::SetVisible (Bool ok)
{
	OP_METHOD (SetVisible, 1);
	OP_ARG (1, Bool, ok);
	
	this->state &= ~ (OP_VIEW_IS_HIDDEN);
	this->state |= ok ? 0 : OP_VIEW_IS_HIDDEN;
}



/*
 *	Return TRUE if the view is enabled; this will be the
 *	case if the global OP_VIEW_IS_DISABLED is cleared and
 *	if the local OP_VIEW_IS_ENABLED is set !
 */

Bool
OPView::ReturnEnable () const
{
	OP_METHOD (ReturnEnable, 0);
	return ( (this->state & OP_VIEW_IS_ENABLED)
	      && ((this->state & OP_VIEW_IS_DISABLED) == 0) ) ? TRUE : FALSE;
}


/*
 *	Return TRUE if the view is visible.
 */

Bool
OPView::ReturnVisible () const
{
	OP_METHOD (ReturnVisible, 0);
	return (this->state & OP_VIEW_IS_HIDDEN) ? FALSE : TRUE;
}



/*
 *	Set the global enable mode. This will not affect any
 *	child view. Affects OP_VIEW_IS_DISABLED. This will
 *	not redraw the view.
 */

void
OPView::SetGlobalEnable (Bool mode)
{
	OP_METHOD (SetGlobalEnable, 1);
	OP_ARG (1, Bool, mode);
	
	this->state &= ~ (OP_VIEW_IS_DISABLED);
	this->state |= (mode == FALSE) ? OP_VIEW_IS_DISABLED : 0;
}



/*
 *	Return TRUE if the view is globally enabled; this is
 *	the case if OP_VIEW_IS_DISABLED is cleared.
 */

Bool
OPView::ReturnGlobalEnable () const
{
	OP_METHOD (ReturnGlobalEnable, 0);
	return (this->state & OP_VIEW_IS_DISABLED) ? FALSE : TRUE;
}



/*
 *	Change the view's state (set or clear, according to
 *	the boolean parameter). This will not affect any child
 *	view nor redraw the view.
 */

void
OPView::ChangeState (OPViewState state, Bool set)
{
	OP_METHOD (ChangeState, 2);
	OP_ARG (1, OPViewState, state);
	OP_ARG (2, Bool, set);
	
	if (set) {
		this->state |= state;
	} else {
		this->state &= ~state;
	}
}


/*
 *	Define the view's state. Use with care...
 */

Bool
OPView::SetState (OPViewState state)
{
	OP_METHOD (SetState, 1);
	OP_ARG (1, OPViewState, state);
	
	this->state = state;
	return TRUE;
}

/*
 *	Return the view's state.
 */

OPViewState
OPView::ReturnState () const
{
	OP_METHOD (ReturnState, 0);
	return this->state;
}



/*
 *	Look for the specified user tag. Scan the view and
 *	every child... Return TRUE if found.
 */

Bool
OPView::FindUserTag (Int32 tag, OPViewOut view)
{
	OP_METHOD (FindUserTag, 2);
	OP_ARG (1, Int32, tag);
	OP_ARG (2, OPViewOut, view);
	
	if (this->tag == tag) {
		view = this;
		return TRUE;
	}
	
	//	Scan each child view...
	
	OPViewRef iter = this->child.head;
	
	while (iter.IsValid ()) {
		if (iter->FindUserTag (tag, view)) return TRUE;
		iter->GetNextSibling (iter);
	}
	
	return FALSE;
}

Bool
OPView::FindUserName (OPAtomIn name, OPViewOut view)
{
	OP_METHOD (FindUserName, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPViewOut, view);
	
	if (this->name == name) {
		view = this;
		return TRUE;
	}
	
	//	Scan each child view...
	
	OPViewRef iter = this->child.head;
	
	while (iter.IsValid ()) {
		if (iter->FindUserName (name, view)) return TRUE;
		iter->GetNextSibling (iter);
	}
	
	return FALSE;
}


/*
 *	Return the (sub)view containing the specified point.
 *	The view assumes that the subview all reside in the
 *	same box as itself. Return FALSE if not in a view.
 */

Bool
OPView::Detect (OPCoord x, OPCoord y, OPViewOut view)
{
	OP_METHOD (Detect, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPViewOut, view);
	
	if ( ( (this->state & OP_VIEW_OK_DETECT) == 0)
	  || (this->ReturnEnable () == FALSE)
	  || (this->ReturnVisible () == FALSE) ) {
		return FALSE;
	}
	
	OPViewRef self = this;
	
	if ( (self == this->window)
	  || (this->box.IsInside (x, y)) ) {
		
		//	Could be this view or one of the child views. Ask any
		//	child if the specified point matches...
		
		OPViewRef iter = this->child.tail;		//	start from end (last drawn first)
		
		while (iter.IsValid ()) {
			if (iter->Detect (x, y, view)) return TRUE;
			iter->GetPrevSibling (iter);
		}
		
		//	No matching child found, then the point is really in
		//	our view.
		
		view = this;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the rectangle (even partly) covers the
 *	view.
 */

Bool
OPView::Detect (OPRectIn rect)
{
	OP_METHOD (Detect, 1);
	OP_ARG (1, OPRectIn, rect);
	
	if ( ( (this->state & OP_VIEW_OK_DETECT) == 0)
	  || (this->ReturnEnable () == FALSE)
	  || (this->ReturnVisible () == FALSE) ) {
		return FALSE;
	}
	
	return this->box.DoOverlap (rect);
}



/*
 *	Return TRUE if the view is the current main responder.
 */

Bool
OPView::IsMainResponder () const
{
	OP_METHOD (IsMainResponder, 0);
	return (this->state & OP_VIEW_IS_ACTIVE) ? TRUE : FALSE;
}



/*
 *	The parent is signalling that it is undergoing a
 *	resizing operation. The new size will be "dx", "dy";
 *	the previous size is still available by performing
 *	a GetSize query to the parent.
 */

void
OPView::ParentResizing (OPCoord dx, OPCoord dy)
{
	OP_METHOD (ParentResizing, 2);
	OP_ARG (1, OPCoord, dx);
	OP_ARG (2, OPCoord, dy);
	
	//	Default : don't do anything...
}


/*
 *	A child is signalling that it has changed its size.
 *	The parent might decide to resize in turn.
 */

void
OPView::ChildChangedSize (OPViewIn child)
{
	OP_METHOD (ChildChangedSize, 1);
	OP_ARG (1, OPViewIn, child);
	
	//	Default : don't do anything
}


/*
 *	Broadcast a message to every subview. This might be used
 *	to signal a recolor or another phenomenon.
 */

void
OPView::Broadcast (OPAtomIn message, OPObjectIn value)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, message);
	OP_ARG (2, OPObjectIn, value);
	
	OPViewRef iter = this->child.head;
	
	while (iter.IsValid ()) {
		iter->Broadcast (message, value);
		iter->GetNextSibling (iter);
	}
}



/*
 *	Draw the full view, using the specified GRAIL port.
 *	This will paint into a buffer (OPaC is using double
 *	buffering).
 */

void
OPView::Draw (OPGrailPortIn port, OPDrawOp what)
{
	OP_METHOD (Draw, 2);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPDrawOp, what);
	
	if ( (this->no_draw_back)				// don't draw the view
	  || (this->window.IsInvalid ())
	  || (this->ReturnVisible () == FALSE)
	  || (port.IsInvalid ()) ) {
		return;
	}
	
	OPViewRef self = this;
	
	if (self == this->window) {
		
		//	Draw the window : use the window's bounding box for the drawing
		//	operation. The origin should be 0, but we make sure...
		
		OPRect rect = this->box;
		rect.SetOrigin (0, 0);
		this->Draw (port, rect, what);
		
	} else {
		
		//	Tell the window to draw the specified rectangle (that is, its own
		//	bounding box).
		
		this->window->Draw (port, this->box, what);
	}
}



/*
 *	Draw part of the view, using the specified GRAIL
 *	port. This will paint into a buffer (OPaC is using
 *	double buffering).
 *
 *	The current clipping rectangle can be used for any
 *	drawing operation.
 */

void
OPView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	OPViewRef self = this;
	
	if ( (self != this->window)
	  && (rect.DoOverlap (this->box) == FALSE) ) {
		return;
	}
	
	if ( (port.IsInvalid ())
	  || (this->ReturnVisible () == FALSE) ) {
		return;
	}

	
	OPViewRef iter = this->child.head;
	OPRect    old_rect;
	OPRect    new_rect;
	
	if (iter.IsValid ()) {
		
		port->GetClipRect (old_rect);
		
		while (iter.IsValid ()) {
			
			OPView* i0c = (OPView*)iter.CheckAndResolve ();
			OPView* i0n = (OPView*)i0c->sibling.next.Resolve ();
			OPView* i0p = (OPView*)i0c->sibling.prev.Resolve ();
			
			if ( (iter->state & OP_VIEW_IS_FROZEN) == 0) {
				
				new_rect  = rect;
				new_rect *= iter->ReturnBox ();
				
				if (new_rect.IsEmpty () == FALSE) {
					
					//	Select clipping rectangle and draw...
					
					port->SetClipRect (new_rect);
					iter->Draw (port, new_rect, what);
				}
			}
			
			OPView* i1c = (OPView*)iter.CheckAndResolve ();
			OPView* i1n = (OPView*)i1c->sibling.next.Resolve ();
			OPView* i1p = (OPView*)i1c->sibling.prev.Resolve ();
			
			OP_ASSERT (i0c == i1c);
			OP_ASSERT (i0n == i1n);
			OP_ASSERT (i0p == i1p);
			
			iter->GetNextSibling (iter);
		}
		
		//	Finally, restore original clipping rectangle. This might even
		//	be suppressed !
		
		port->SetClipRect (old_rect);
	}
}


/*
 *	Draw the overlay (if any) after a window has been painted.
 */

void
OPView::DrawOverlay (OPGrailPortIn port, OPRectIn rect)
{
	OP_METHOD (DrawOverlay, 2);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	
	OPViewRef self = this;
	
	if (self == this->window) {
		return;
	} else {
		if (this->state & OP_VIEW_IS_CONNECT) {
			OPAdornerRef adorner = port->ReturnAdorner ();
			adorner->DrawRect (this->box, OP_COLOR_VIEW_CONNECT, -1);
			adorner->DrawRect (this->box, OP_COLOR_VIEW_CONNECT, -2);
		}
		if (this->state & OP_VIEW_IS_DROPPING) {
			OPAdornerRef adorner = port->ReturnAdorner ();
			adorner->FillRect (this->box, OP_COLOR_VIEW_DROPPING);
		}
	}
}
	

/*
 *	Make a view display itself. This will invoke "Draw".
 */

void
OPView::Display ()
{
	OP_METHOD (Display, 0);
	
	const OPGrailPortRef& port = this->ReturnGrailPort ();
	const OPWindowRef&    wind = this->window;
	
	if ( (port.IsInvalid ())
	  || (wind.IsInvalid ())
	  || (this->ReturnVisible () == FALSE) ) {
		return;
	}
	
	OPViewRef self = this;
	OPRect    rect = this->box;
	OPViewRef iter = this;
	
	//	Work all the way up the display hierarchy in order to insure that
	//	the clipping rectangle is correct and that even the superview will
	//	redraw what it has to draw.
	
	if (iter != wind) {
		while ( (iter->GetParent (iter))
		     && (iter != wind)
		     && (rect.IsEmpty () == FALSE) ) {
			rect *= iter->ReturnBox ();
		}
	}
	
	if (self == wind) {
		rect.SetOrigin (0, 0);
	}
	
	if ( (rect.IsEmpty () == FALSE)
	  && (iter.IsValid ()) ) {
		port->BeginDraw (rect);
		iter->Draw (port, rect);
		port->EndDraw ();
	}
}



/*
 *	Make a view display part of itself. This will redraw
 *	the window in the specified rectangle. It will repaint
 *	each view contained in the specified rectangle.
 */

void
OPView::Display (OPRectIn rect)
{
	OP_METHOD (Display, 1);
	OP_ARG (1, OPRectIn, rect);
	
	if ( (this->window.IsValid ())
	  && (this->ReturnVisible () == FALSE) ) {
		this->window->Redraw (rect);
	}
}



/*
 *	Return a reference to the GRAIL port to use for any
 *	drawing operation. It will be the window's.
 */

const OPGrailPortRef&
OPView::ReturnGrailPort () const
{
	OP_METHOD (ReturnGrailPort, 0);
	
	return this->window.IsValid () ? this->window->ReturnGrailPort ()
								   : OPGrailPortNull;
}



/*
 *	Return the associated look manager. This simplifies
 *	the view's access to the LookMan instance.
 */

const OPLookManRef&
OPView::ReturnLookMan () const
{
	OP_METHOD (ReturnLookMan, 0);
	return OPLookMan::Current ();
}



/*
 *	Return the associated adorner. This simplifies the
 *	view's access to the Adorner instance.
 */

const OPAdornerRef&
OPView::ReturnAdorner () const
{
	OP_METHOD (ReturnAdorner, 0);
	
	static OPAdornerRef none = 0;
	
	const OPGrailPortRef& port = this->ReturnGrailPort ();
	return port.IsValid () ? port->ReturnAdorner ()
						   : none;
}


/*
 *	Handle the drag & drop protocol. This is used for real drag
 *	and drop, inter-object connection, etc.
 */

Bool
OPView::DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp)
{
	OP_METHOD (DragAndDrop, 3);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPObjectIn, initiator);
	OP_ARG (3, OPDragDropProtocol, ddp);
	
	if (this->ReturnVisible () == FALSE) {
		return FALSE;
	}
	
	OPObjectRef object;
	OPViewRef   parent = this->parent;
	OPViewRef   self   = this;
	
	if (self == this->window) {
		parent = OPViewNull;
	}
	
	//	Handle the supported operations
	
	switch (ddp) {
		
		case OP_DDP_DRAGnDROP_CHECK:
			return ((this->state & OP_VIEW_OK_DROP) ? TRUE : FALSE)
				&& (parent.IsInvalid () || parent->DragAndDrop (box, initiator, ddp));
		
		case OP_DDP_DRAGnDROP_ENTER:
			
			this->state |= OP_VIEW_IS_DROPPING;
			
			if (this->num_overlays++) {
				this->Display ();
			} else {
				this->window->NotifyInsertOverlay (this);
				this->window->Display ();
			}
			return TRUE;
		
		case OP_DDP_DRAGnDROP_LEAVE:
		case OP_DDP_DRAGnDROP_CANCEL:
		case OP_DDP_DRAGnDROP_COMMIT:
			
			this->state &= ~OP_VIEW_IS_DROPPING;
			
			if (--this->num_overlays) {
				this->Display ();
			} else {
				this->window->NotifyRemoveOverlay (this);
				this->window->Display ();
			}
			break;
		
		case OP_DDP_DRAGnDROP_MOTION:
			return TRUE;
		
		case OP_DDP_CONNECT_CHECK:
			return (parent.IsInvalid () || parent->DragAndDrop (box, initiator, ddp));
		
		case OP_DDP_CONNECT_ENTER:
			
			this->state |= OP_VIEW_IS_CONNECT;
			
			if (this->num_overlays++) {
				this->Display ();
			} else {
				this->window->NotifyInsertOverlay (this);
				this->window->Display ();
			}
			return TRUE;
		
		case OP_DDP_CONNECT_LEAVE:
		case OP_DDP_CONNECT_CANCEL:
		case OP_DDP_CONNECT_COMMIT:
			
			this->state &= ~OP_VIEW_IS_CONNECT;
			
			if (--this->num_overlays) {
				this->Display ();
			} else {
				this->window->NotifyRemoveOverlay (this);
				this->window->Display ();
			}
			break;
		
		case OP_DDP_CONNECT_MOTION:
			return TRUE;
		
		case OP_DDP_UI_EDIT_CHECK:
			return (this->state & OP_VIEW_OK_UI_EDIT)
				&& (this->state & OP_VIEW_OK_UI_MOVE)
				&& (this->window.IsValid ())
				&& (this->window->ReturnApp ()->IsBuildEnabled ())
				&& (parent.IsInvalid () || parent->DragAndDrop (box, initiator, ddp));
		
		case OP_DDP_UI_EDIT_COMMIT:
			break;
		
		//	The substitution process returns TRUE if no substitution has taken
		//	place, FALSE otherwise. Substitution should always start first with
		//	the topmost parent.
		
		case OP_DDP_DRAGnDROP_SUBST:
		case OP_DDP_CONNECT_SUBST:
		case OP_DDP_UI_EDIT_SUBST:
			return (parent.IsInvalid () || parent->DragAndDrop (box, initiator, ddp));
		
		default:
			return FALSE;
	}
	
	//	We might want to do a second pass in order to handle the cancel or
	//	commit operations.
	
	OPMessageRef doCommit;
	OPAtomRef    att_name;
	OPObjectRef  data = this;
	
	switch (ddp) {
		
		case OP_DDP_DRAGnDROP_COMMIT:
			doCommit = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			att_name = "embedded_object_for_drop";
			doCommit->SetMethod ("CommitDrop");
			this->GetAttribute (att_name, data);
			initiator <- doCommit << data;
			break;
			
		case OP_DDP_CONNECT_COMMIT:
			doCommit = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			att_name = "embedded_object_for_connect";
			doCommit->SetMethod ("CommitConnect");
			this->GetAttribute (att_name, data);
			initiator <- doCommit << data;
			break;
		
		case OP_DDP_UI_EDIT_COMMIT:
			doCommit = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			att_name = "embedded_object_for_ui_edit";
			doCommit->SetMethod ("CommitUIEdit");
			this->GetAttribute (att_name, data);
			initiator <- doCommit << data;
			break;
		
		default:
			break;
	}
	
	return TRUE;
}


void
OPView::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPViewRef view = this;
	
	while (view.IsValid ()) {
		
		if (view->look_cursor == OP_CURSOR_USER) {
			
			//	NIY
			
			break;
			
		} else if (this->look_cursor == OP_CURSOR_DEFAULT) {
			
			view = view->ReturnParent ();
			
		} else {
			
			OPLookMan::Current ()->SelectCursor (this->look_cursor);
			break;
		}
	}
	
	if (view.IsInvalid ()) {
		OPLookMan::Current ()->SelectCursor (OP_CURSOR_NORMAL);
	}
	
	super::EnterNotify (event);
}

void
OPView::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::LeaveNotify (event);
}
