/*
 *	gfxfontcache.cxx
 *
 *	The Font Cache provides the necessary support to use the FreeType
 *	engine to draw glyphs on a GfxSurface.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

#include "gfxtypes.h"
#include "gfxfontcache.h"
#include "gfxsurface.h"

#include "../freetype/lib/freetype.h"

#include <string.h>

/****************************************************************/

#define	MAX_UNICODE	1024

static const int	BAND_INCREMENT	= 10;
static const int	FONT_DPI		= 72;

/****************************************************************/

struct GfxFontCache::FreeType
{
	TT_Engine			engine;
	TT_Face				face;
	TT_Instance			instance;
	TT_Glyph			glyph;
	TT_CharMap			char_map;
	TT_Glyph_Metrics	metrics;
	TT_Outline			outline;
	TT_Face_Properties	properties;
	TT_Instance_Metrics imetrics;
	TT_Raster_Map		bit;
	int					glyph_code[MAX_UNICODE];
};

/****************************************************************/

/*
 *	The GlyphInfo must be initialised properly. The best way still is to use
 *	a constructor (even if it is a simple structure, not a class).
 */

GfxFontCache::GlyphInfo::GlyphInfo ()
{
	this->next = 0;
	
	this->x = this->y = this->dx = this->dy = 0;
	this->geom.x_min = this->geom.x_max = 0;
	this->geom.y_min = this->geom.y_max = 0;
	
	this->unicode = 0xFFFF;
	this->band_id = 0;
	this->kern_id = 0;
	this->liga_id = 0;
	this->geom.x_bearing = 0;
	this->geom.y_bearing = 0;
	this->geom.x_advance = 0;
	this->geom.y_advance = 0;
}

/*
 *	Deleting a glyph information will also delete every "child" with the
 *	same sub-page code (low byte of Unicode code).
 */

GfxFontCache::GlyphInfo::~GlyphInfo ()
{
	if (this->next) {
		DELETE (this->next);
		this->next = 0;
	}
}



GfxFontCache::GfxFontCache ()
{
	this->bands     = 0;
	this->num_bands = 0;
	this->band_id = 0;
	this->band_ox = 0;
	this->band_dx = 0;
	this->band_dy = 0;
	
	this->font = 0;
	this->size = 0;
}

GfxFontCache::~GfxFontCache ()
{
	if (this->font) this->CloseFace ();
	
	for (Card32 i = 0; i < this->num_bands; i++) {
		if (this->bands[i]) GfxSurface::Delete (this->bands[i]);
	}
	
	if (this->bands) DELETE (this->bands);
}


/*
 *	Create the specified glyph info record (empty, but for the unicode
 *	field).
 */

GfxFontCache::GlyphInfo*
GfxFontCache::NewGlyphInfo (Card16 unicode)
{
	if (unicode >= MAX_UNICODE) return 0;
	
	GlyphInfo* info = & this->glyphs[unicode & 0x00FF];
	
	//	The Unicode glyphs which are not in the first page will never be stored in
	//	the first entry of the glyphs table.
	
	if (unicode & 0xFF00) {
		
		while (info->next) {
			info = info->next;
		}
		
		//	Allocate a new information structure.
		
		info->next = NEW GlyphInfo;
		info = info->next;
	}
	
	info->unicode = unicode;
	return info;
}


/*
 *	Find the specified glyph info. If it does not exist, return a null
 *	pointer.
 */

GfxFontCache::GlyphInfo*
GfxFontCache::FindGlyphInfo (Card16 unicode)
{
	GlyphInfo* info = & this->glyphs[unicode & 0x00FF];
	
	while (info && (info->unicode != unicode)) {
		info = info->next;
	}
	
	return info;
}


/*
 *	Create a new band for caching.
 */

void
GfxFontCache::NewBand ()
{
	//	Allocate the graphic surface (8-bit) used as the cache for the
	//	glaphs.
	
	GfxSurface* surface = GfxSurface::New (8);
	surface->Initialise (this->band_dx, this->band_dy);
	
	Card32 count = 1;
	
	if (this->bands == 0) this->GrowBands ();
	
	//	Find the first free entry in the table where the surface will be
	//	registered. Skip the entry zero...
	
	while (this->bands[count]) {
		
		count++;
		
		if (count == this->num_bands) {
			this->GrowBands ();
		}
	}
	
	this->bands[count] = surface;
	this->band_id = (Card16) count;
	this->band_ox = 0;
}


/*
 *	Add room for additional bands.
 */

void
GfxFontCache::GrowBands ()
{
	Card32   new_count = this->num_bands+BAND_INCREMENT;
	GfxSurface** bands = NEW GfxSurface* [new_count];
	
	for (Card32 i = 0; i < this->num_bands; i++) {
		bands[i] = this->bands[i];
	}
	for (; i < new_count; i++) {
		bands[i] = 0;
	}
	
	if (this->bands) DELETE (this->bands);
	
	this->bands     = bands;
	this->num_bands = new_count;
}


/*
 *	Find where the glyph data will be stored (offset and band ID). A new
 *	band will be allocated when needed.
 */

void
GfxFontCache::NewGlyphInBand (Card16 dx, Card16& x, Card16& band_id)
{
	if (this->band_id == 0) {
		this->NewBand ();
	}
	if (this->band_ox+dx > this->band_dx) {
		this->NewBand ();
	}
	
	band_id = this->band_id;
	x       = this->band_ox;
	
	this->band_ox += dx;
}


/*
 *
 */

void
GfxFontCache::FillGlyphInfo (GlyphInfo* info)
{
	if (this->font == 0) return;
	
	int index = this->font->glyph_code[info->unicode];
	
	TT_Load_Glyph (this->font->instance, this->font->glyph, index,
				   TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH);
	
	TT_Get_Glyph_Outline (this->font->glyph, &this->font->outline);
	TT_Get_Glyph_Metrics (this->font->glyph, &this->font->metrics);
	
	TT_F26Dot6  xmin, ymin, xmax, ymax;
	
	//	We begin by grid-fitting the bounding box
	
	xmin = (this->font->metrics.bbox.xMin+0)  & -64;
	ymin = (this->font->metrics.bbox.yMin+0)  & -64;
	xmax = (this->font->metrics.bbox.xMax+63) & -64;
	ymax = (this->font->metrics.bbox.yMax+63) & -64;
	
	Card16 dx = (Card16)((xmax - xmin) / 64);
	Card16 dy = (Card16)((ymax - ymin) / 64);
	
	this->font->bit.width  = dx;
	this->font->bit.cols   = (dx+31) & -32;
	this->font->bit.rows   = dy;
	this->font->bit.flow   = TT_Flow_Down;
	this->font->bit.size   = this->font->bit.cols * this->font->bit.rows;
	this->font->bit.bitmap = NEW char[this->font->bit.size];
	
	memset (this->font->bit.bitmap, 0, this->font->bit.size);
	
	info->geom.x_min = (Int16)(xmin / 64);
	info->geom.x_max = (Int16)(xmax / 64);
	info->geom.y_min = (Int16)(ymin / 64);
	info->geom.y_max = (Int16)(ymax / 64);
	info->dx = dx;
	info->dy = dy;
	info->y  = 0;
	info->geom.x_bearing = (Int16)((this->font->metrics.bearingX+63) / 64);
	info->geom.y_bearing = (Int16)((this->font->metrics.bearingY+63) / 64);
	info->geom.x_advance = (Int16)((this->font->metrics.advance+63) / 64);
	info->geom.y_advance = 0;
	
	this->NewGlyphInBand (dx, info->x, info->band_id);
	
	TT_Get_Glyph_Pixmap (this->font->glyph, &this->font->bit, -xmin, -ymin);
	
	for (int i = 0; i < this->font->bit.cols * dy; i++) {
		((Card8*)(this->font->bit.bitmap))[i] ^= 0xFF;
	}
	
	GfxSurface surface (8);
	surface.Initialise (dx, dy, this->font->bit.bitmap, this->font->bit.cols);
	
	this->bands[info->band_id]->RectCopy (info->x, info->y, info->dx, info->dy,
										  0, 0, &surface);
	
	
	DELETE (this->font->bit.bitmap);
	this->font->bit.bitmap = 0;
}

/****************************************************************/

void
GfxFontCache::CharToUnicode ()
{
	int    i, n;
	Card16 platform, encoding;
	
	//	First, look for a Unicode charmap.
	
	n = TT_Get_CharMap_Count (this->font->face);
	
	for (i = 0; i < n; i++) {
		TT_Get_CharMap_ID (this->font->face, i, &platform, &encoding);
		
		if ( (platform == 3 && encoding == 1)
		  || (platform == 0 && encoding == 0) ) {
			TT_Get_CharMap (this->font->face, i, &this->font->char_map );
			i = n + 1;
		}
	}
	
	//	Check if a Unicode charmap has been found.
	
	if (i == n) return;
	
	for (i = 0; i < sizeof (this->font->glyph_code) / sizeof (int); i++ ) {
		this->font->glyph_code[i] = TT_Char_Index (this->font->char_map, i);
	}
}


Bool
GfxFontCache::OpenFace (const char* ttf_name, int size)
{
	static unsigned char virtual_palette[5] = { 0x00, 0x40, 0x80, 0xC0, 0xFF };
	
	int error = 0;
	
	if (this->font) {
		this->CloseFace ();
	}
	
	//	Allocate structure for the free type font description.
	
	this->font = NEW FreeType;
	this->size = size;
	
	memset (this->font, 0, sizeof (FreeType));
	
	for (;;) {
		
		error = TT_Init_FreeType (&this->font->engine);
		if (error) break;
		
		error = TT_Open_Face (this->font->engine, ttf_name, &this->font->face);
		if (error) break;
		
		TT_Get_Face_Properties (this->font->face, &this->font->properties);
		
		error = TT_New_Glyph (this->font->face, &this->font->glyph);
		if (error) break;
		
		error = TT_New_Instance (this->font->face, &this->font->instance);
		if (error) break;
		
		error = TT_Set_Instance_Resolutions (this->font->instance, FONT_DPI, FONT_DPI);
		if (error) break;
		
		TT_Set_Raster_Gray_Palette (this->font->engine, (Card8*) virtual_palette);
		
		TT_Set_Instance_CharSize (this->font->instance, size*64);
		TT_Get_Instance_Metrics (this->font->instance, &this->font->imetrics);
		
		this->band_dx = 64; //100 * this->font->imetrics.x_ppem;
		this->band_dy = this->font->imetrics.y_ppem;
		
		int typo_max = this->font->properties.horizontal->xMax_Extent * 72;
		
		int typo_a = this->font->properties.os2->sTypoAscender * FONT_DPI;
		int typo_d = - (Int16)(this->font->properties.os2->sTypoDescender) * FONT_DPI;
		int typo_l = this->font->properties.os2->sTypoLineGap * FONT_DPI;
		
		this->typo_ascender  = (typo_a * this->size + typo_max - 1) / typo_max + 1;
		this->typo_descender = (typo_d * this->size + typo_max - 1) / typo_max;
		this->typo_linegap   = (typo_l * this->size + typo_max - 1) / typo_max;
		
		this->CharToUnicode ();
		this->font->bit.bitmap = 0;

		return TRUE;
	}
	
	this->CloseFace ();
	
	return FALSE;
}

void
GfxFontCache::CloseFace ()
{
	if (this->font) {
		
		TT_Close_Face (this->font->face);
		TT_Done_FreeType (this->font->engine);
		
		DELETE (this->font);
		this->font = 0;
	}
}


void
GfxFontCache::AccessGlyph (Card16 unicode, GfxSurface*& surface,
						   Card16& x, Card16& y, GlyphGeom& geom)
{
	if (this->font == 0) {
		static const GlyphGeom zero_geom = { 0 };
		x    = 0;
		y    = 0;
		geom = zero_geom;
		return;
	}
	
	GlyphInfo* info = this->FindGlyphInfo (unicode);
	
	if (info == 0) {
		info = this->NewGlyphInfo (unicode);
		this->FillGlyphInfo (info);
	}
	
	surface = this->bands[info->band_id];
	x       = info->x;
	y       = info->y;
	geom    = info->geom;
}

void
GfxFontCache::DrawGlyph (Card16 unicode, Card32 x, Card32 y, Card32 pixel, GfxSurface* surface)
{
	if (this->font == 0) {
		return;
	}
	
	GlyphInfo* info = this->FindGlyphInfo (unicode);
	
	if (info == 0) {
		info = this->NewGlyphInfo (unicode);
		this->FillGlyphInfo (info);
	}
	
	surface->RectFillBlend (x+info->geom.x_min, y-info->geom.y_max, info->dx, info->dy,
							pixel, info->x, info->y, this->bands[info->band_id]);
}


void
GfxFontCache::GetGlyphGeometry (Card16 unicode, GlyphGeom& geom)
{
	if (this->font == 0) {
		static const GlyphGeom zero_geom = { 0 };
		geom = zero_geom;
		return;
	}
	
	GlyphInfo* info = this->FindGlyphInfo (unicode);
	
	if (info == 0) {
		info = this->NewGlyphInfo (unicode);
		this->FillGlyphInfo (info);
	}
	
	geom = info->geom;
}

