/*
 *	springview.cxx
 *
 *	The spring view draws the overlays for a specified spring
 *	band.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "springview.h"

#include <OPaC/app.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/spring.h>
#include <OPaC/springboard.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
APP_IMPLEMENT_CLASS (OPSpringView);

OP_PUBLISH_SLOT (OPSpringView, band, ReturnSpringBand, SetSpringBand, "Spring band", OPSpringBand);

/****************************************************************/

OPSpringView::OPSpringView ()
{
	OP_METHOD (OPSpringView, 0);
	this->SetAttribute ("OPaC.volatile.observer");
}

OPSpringView::~OPSpringView ()
{
	OP_METHOD (~OPSpringView, 0);
}

void
OPSpringView::Disconnect ()
{
	OP_METHOD (Disconnect, 0);
	
	if (this->window.IsValid ()) {
		this->window->NotifyRemoveOverlay (this);
		this->window = OPWindowNull;
	}
	
	this->band = OPSpringBandNull;
}


void
OPSpringView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->Disconnect ();
	
	super::Delete ();
}


Bool
OPSpringView::SetHiliteView (OPViewIn view)
{
	OP_METHOD (SetHiliteView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->hilite = view;
	
	return TRUE;
}

Bool
OPSpringView::SetHiliteAll ()
{
	OP_METHOD (SetHiliteAll, 0);
	
	this->hilite = OPViewNull;
	return TRUE;
}


Bool
OPSpringView::SetSpringBand (OPSpringBandIn band)
{
	OP_METHOD (SetSpringBand, 1);
	OP_ARG (1, OPSpringBandIn, band);
	
	this->Disconnect ();
	
	if ( (band.IsValid ())
	  && (band->GetAssociatedWindow (this->window)) ) {
		this->band = band;
		this->window->NotifyInsertOverlay (this);
		return TRUE;
	}
	
	return FALSE;
}

const OPSpringBandRef&
OPSpringView::ReturnSpringBand () const
{
	OP_METHOD (ReturnSpringBand, 0);
	return this->band;
}

void
OPSpringView::DrawOverlay (OPGrailPortIn port, OPRectIn rect)
{
	OP_METHOD (DrawOverlay, 2);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	
	if ( (this->band.IsValid ())
	  && (this->window.IsValid ()) ) {
		
		Bool hilite_all = this->hilite.IsInvalid ();
		
		Count num   = this->band->CountViews ();
		Count index = 0;
		Count total = 0;
		
		OPViewRef view;
		
		OPCoord mean_x = 0;
		OPCoord mean_y = 0;
		OPCoord wdx = 0;
		OPCoord wdy = 0;
		
		this->window->GetSize (wdx, wdy);
		
		OPColorPixel pix_black0 (0.0, 0.0, 0.0, 0.0);
		OPColorPixel pix_black1 (0.0, 0.0, 0.0, 0.4);
		OPColorPixel pix_black2 (0.0, 0.0, 0.0, 0.8);
		OPColorPixel pix_white0 (1.0, 1.0, 1.0, 0.8);
		
		for (index = 0; index < num; index++) {
			if (this->band->GetNthView (index, view)) {
				
				OPRect box = view->ReturnBox ();
				
				mean_x += box.X1 () + box.X2 ();
				mean_y += box.Y1 () + box.Y2 ();
				total++;
				
				if (box.DoOverlap (rect)) {
					port->SetForeColor (pix_black0);
					port->DrawRect (box, 0);
					port->SetForeColor (pix_black1);
					port->DrawRect (box, -1);
					port->SetForeColor (pix_black2);
					port->DrawRect (box, -2);
				}
				
				if ( (hilite_all)
				  || (this->hilite == view) ) {
					port->SetForeColor (pix_white0);
					port->FillRect (box, 1);
				}
			}
		}
		
		if (total == 0) return;
		
		mean_x /= 2 * total;
		mean_y /= 2 * total;
		
		OPCoord pos = 0;
		
		OPColorPixel col0 = OPColorPixel (0.0, 0.0, 0.4, 0.0);
		OPColorPixel col1 = OPColorPixel (0.0, 0.0, 0.4, 0.4);
		OPColorPixel col2 = OPColorPixel (0.0, 0.0, 0.4, 0.8);
		OPColorPixel grid = OPColorPixel (0.0, 0.0, 1.0, 0.8);
		
		for (index = 0; index <= num; index++) {
			if (this->band->GetNthView (index, view)) {
				
				OPRect box = view->ReturnBox ();
				
				if (this->band->IsVertical ()) {
					
//					OPCoord middle = (box.X1 () + box.X2 ()) / 2;
					
					port->SetForeColor (col2);
					port->DrawLineToY (mean_x-2, pos, box.Y1 () - 1);
					port->DrawLineToY (mean_x+2, pos, box.Y1 () - 1);
//					port->DrawLineToY (middle-2, box.Y1 () + 1, box.Y2 () - 1);
//					port->DrawLineToY (middle+2, box.Y1 () + 1, box.Y2 () - 1);
					port->SetForeColor (col1);
					port->DrawLineToY (mean_x-1, pos, box.Y1 () - 1);
					port->DrawLineToY (mean_x+1, pos, box.Y1 () - 1);
//					port->DrawLineToY (middle-1, box.Y1 () + 1, box.Y2 () - 1);
//					port->DrawLineToY (middle+1, box.Y1 () + 1, box.Y2 () - 1);
					port->SetForeColor (col0);
					port->DrawLineToY (mean_x, pos, box.Y1 () - 1);
//					port->DrawLineToY (middle, box.Y1 () + 1, box.Y2 () - 1);
					
					port->SetForeColor (grid);
					port->DrawLineDX (0, box.Y1 (), wdx);
					port->DrawLineDX (0, box.Y2 (), wdx);
					
					pos = box.Y2 () + 1;
					
				} else {
					
//					OPCoord middle = (box.Y1 () + box.Y2 ()) / 2;
					
					port->SetForeColor (col2);
					port->DrawLineToX (pos, mean_y-2, box.X1 () - 1);
					port->DrawLineToX (pos, mean_y+2, box.X1 () - 1);
//					port->DrawLineToX (box.X1 () + 1, middle-2, box.X2 () - 1);
//					port->DrawLineToX (box.X1 () + 1, middle+2, box.X2 () - 1);
					port->SetForeColor (col1);
					port->DrawLineToX (pos, mean_y-1, box.X1 () - 1);
					port->DrawLineToX (pos, mean_y+1, box.X1 () - 1);
//					port->DrawLineToX (box.X1 () + 1, middle-1, box.X2 () - 1);
//					port->DrawLineToX (box.X1 () + 1, middle+1, box.X2 () - 1);
					port->SetForeColor (col0);
					port->DrawLineToX (pos, mean_y, box.X1 () - 1);
//					port->DrawLineToX (box.X1 () + 1, middle, box.X2 () - 1);
					
					port->SetForeColor (grid);
					port->DrawLineDY (box.X1 (), 0, wdy);
					port->DrawLineDY (box.X2 (), 0, wdy);
					
					pos = box.X2 () + 1;
				}
			}
		}
		
		OPRect box = this->window->ReturnBox ();
		
		if (this->band->IsVertical ()) {
			
			port->SetForeColor (col2);
			port->DrawLineToY (mean_x-2, pos, box.DY ());
			port->DrawLineToY (mean_x+2, pos, box.DY ());
			port->SetForeColor (col1);
			port->DrawLineToY (mean_x-1, pos, box.DY ());
			port->DrawLineToY (mean_x+1, pos, box.DY ());
			port->SetForeColor (col0);
			port->DrawLineToY (mean_x, pos, box.DY ());
			
		} else {
			
			port->SetForeColor (col2);
			port->DrawLineToX (pos, mean_y-2, box.DX ());
			port->DrawLineToX (pos, mean_y+2, box.DX ());
			port->SetForeColor (col1);
			port->DrawLineToX (pos, mean_y-1, box.DX ());
			port->DrawLineToX (pos, mean_y+1, box.DX ());
			port->SetForeColor (col0);
			port->DrawLineToX (pos, mean_y, box.DX ());
			
			pos = box.X2 () + 1;
		}
	}
}
